"""Mqtt Api service."""
import logging
from queue import Queue

from paho.mqtt.client import Client

from system_configurator import settings
from system_configurator.core.api.api_mqtt.entities import (
    IMqttMessage,
    RespMsgDict,
)
from system_configurator.core.api.api_mqtt.main_loop import (
    IMainLoop,
    MainLoop,
)
from system_configurator.core.api.api_mqtt.mqtt_clients import (
    MqttMsgSenderViaQueue,
    MqttRPClient,
    MqttUserdataDict,
    PahoMqttMsgSerializer,
)
from system_configurator.core.api.api_mqtt.msg_handlers import (
    RouterMsgHandlerWithResponse,
)
from system_configurator.core.api.api_mqtt.publishers import (
    EventCallback,
    PublishRouterBuilder,
    get_publish_routes,
    run_startup_publisher,
)
from system_configurator.core.api.api_mqtt.router import (
    get_router,
    get_subscribe_topics,
)
from system_configurator.core.custom_eventloops.event_loops import (
    InternalEventLoopCreator,
    TimeIntervalEventLoopCreator,
)
from system_configurator.core.file_system.events import FsEventCreator
from system_configurator.core.db.events import DbEventCreator


logger = logging.getLogger(__name__)


receiver_queue: Queue[IMqttMessage] = Queue()
publisher_queue: Queue[RespMsgDict] = Queue()

msg_sender = MqttMsgSenderViaQueue(publisher_queue)


router = get_router()  # FIXME: заменить создание майн луп на паттерн билдер
_msg_handler = RouterMsgHandlerWithResponse(
    router,
    settings.SERVICE_TOPICS,
    msg_sender,
    )
_main_loop = MainLoop(msg_queue=receiver_queue, msg_handler=_msg_handler)

_receive_msg_serializer = PahoMqttMsgSerializer()  # FIXME: заменить создание клиента на паттерн билдер
_userdata: MqttUserdataDict = {
    'topics': get_subscribe_topics(router),  # Dependencies
    'msg_serializer': _receive_msg_serializer,
    'msg_queue': receiver_queue,
    'publisher_queue': publisher_queue,
    'publisher_loop_timeout': settings.PUBLISHER_LOOP_TIMEOUT,  # Dependencies
}
_mqtt_client: Client = MqttRPClient(client_id=settings.MQTT_CLIENT_NAME,  protocol=5)
_mqtt_client.user_data_set(userdata=_userdata)


publish_routes = get_publish_routes(settings.MODULES)

_publish_router = PublishRouterBuilder()(publish_routes)


def run(
    mqtt_client: Client = _mqtt_client,
    msg_handler_loop: IMainLoop = _main_loop,
) -> None:
    """Запускает сервисы mqtt router'a и publisher'a.

    Функция является блокирующей.

    Args:
        mqtt_client (Client, optional): Mqtt клиент. Defaults to _mqtt_client.
        msg_handler_loop (IMainLoop, optional): Обработчик входящих mqtt сообщений. Defaults to _main_loop.
    """
    mqtt_client.connect('localhost')
    mqtt_client.loop_start()
    logger.debug('Mqtt client started')
    event_callback = EventCallback(_publish_router)
    DbEventCreator().create(event_callback, set(_publish_router))  # Dependencies
    FsEventCreator().create(event_callback, set(_publish_router))
    TimeIntervalEventLoopCreator().create(event_callback, _publish_router.keys())
    InternalEventLoopCreator().create(event_callback, _publish_router.keys())
    logger.debug('Publisher events created.')
    run_startup_publisher(publish_routes)  # Dependencies
    logger.debug('All startup messages published.')
    msg_handler_loop.start()
