"""Msg handlers."""
import json
import logging
from typing import Any, Optional, Protocol, Type, TypedDict


from system_configurator.core.api.api_mqtt.entities import (
    IMqttMessage,
    RespMsgDict,
    ResponsePayloadDict,
)
from system_configurator.core.api.api_mqtt.mqtt_clients import IMqttMsgSender
from system_configurator.core.api.api_mqtt.router import TRouter
from system_configurator.core.exceptions import Error


logger = logging.getLogger(__name__)


class IMsgHandler(Protocol):
    """Interface for message handlers.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def handle_msg(self, msg: IMqttMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (IMqttMessage): Mqtt message
        """
        ...


class SimpleMsgHandler():
    """Простой обработчик Mqtt сообщений.

    По топику Mqtt сообщения запускает нужную команду роутера.

    Args:
        topic_fabric (TRouter): Фабрика содержащая
        название топиков и команды которые должны их обрабатывать.
    """

    def __init__(self, topic_fabric: TRouter) -> None:
        self.topic_fabric = topic_fabric

    def handle_msg(self, msg: IMqttMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (IMqttMessage): Mqtt message
        """
        if command := self.topic_fabric.get(msg.topic):
            command.execute(msg.payload)


class RouterMsgHandlerWithResponse():
    """message handlers with error topics and response.

    По mqtt топику получает команду обработки из фабрики команд.
    Запускает ее передав mqtt payload. Если команда вернула результат
    то он отправляется в очередь отправки для публикации в mqtt.

    Если возникло исключение Error то помещает сообщение об ошибке в
    очередь на отправку для публикации в mqtt error_topic.

    Args:
        topic_fabric (TRouter): Фабрика содержащая название топиков и команды которые должны их обрабатывать.
        mqtt_msg_sender (IMqttMsgSender): Публикатор mqtt сообщений.
        service_topics (dict[str, str]): Словарь содержащий сервисные mqtt топики.
    """

    def __init__(
        self,
        topic_fabric: TRouter,
        service_topics: dict[str, str],
        mqtt_msg_sender: IMqttMsgSender,
    ) -> None:
        self._topic_fabric = topic_fabric
        self._error_topic = service_topics['error']
        self._mqtt_msg_sender = mqtt_msg_sender

    def handle_msg(self, msg: IMqttMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (IMqttMessage): Mqtt message
        """
        logger.debug('Start handle msg %s', msg)
        resp_topic = msg.response_topic
        if command := self._topic_fabric.get(msg.topic):
            try:
                resp_payload: Optional[ResponsePayloadDict] = command.execute(msg.payload)
            except Error as err:
                logger.warning(err.message)
                resp_topic = self._error_topic
                resp_payload = {
                    'status': 'error',
                    'msg': err.message,
                    'data': err.data,
                }
            if resp_payload:
                resp_msg: RespMsgDict = {
                    'topic': resp_topic,
                    'cor_data': msg.cor_data,
                    'payload': resp_payload,
                    'retain': False,
                }
                self._mqtt_msg_sender.send_msg(**resp_msg)


class CommandInfoDict(TypedDict):
    """Command information.

    command: str    - command name
    data: Any       - data for command.
    """

    command: str
    data: Any


class ISetSystemParamCommand(Protocol):
    """Interface for set params to network interface.

    Args:
        data (Any): data

    Returns:
        None
    """

    def __init__(self, data: Any) -> None:
        ...

    def execute(self) -> None:
        """Run command."""


class CommandRunner():
    """Execute commands.

    Args:
        commands_fabric (dict[str, Type[ISetSystemParamCommand]]): fabric with executed command.

    Methods:
        run_command: execute property commands
    """

    def __init__(self, commands_fabric: dict[str, Type[ISetSystemParamCommand]]) -> None:
        self.commands_fabric = commands_fabric

    def execute(self, payload: bytes) -> None:
        """Execute property commands.

        Args:
            payload (CommandInfoDict): property commands
        """
        payload_serialized = json.loads(payload)
        command_runnable = self.commands_fabric[payload_serialized['command']]
        command_runnable(payload_serialized['data']).execute()
