"""Custom events."""

from threading import Thread
import time
from typing import Callable, Iterable, NoReturn, Union
from system_configurator.core.api.api_mqtt.publishers import ObserverObject
from system_configurator.core.msgbus import Event, IEventHandler, handle_event_from_msgbus, register_handler


def time_range_loop(
    timeout: Union[int, float], callback: Callable[[ObserverObject], None], obsrv_obj: ObserverObject,
) -> NoReturn:
    """Цикл событий запускающий callback через равные интервалы времени.

    Args:
        timeout (Union[int, float]): Интервал времени.
        callback (Callable[[ObserverObject], None]): Callback функция.
        obsrv_obj (ObserverObject): Аргумент для callback функции.

    Returns:
        NoReturn: _description_
    """
    while True:
        time.sleep(timeout)
        callback(obsrv_obj)


class TimeIntervalEventLoopCreator():
    """Создатель цикла обработки событий запускающихся через равные интервалы времени.

    По наступлению события происходит запуск обратного вызова по отправке mqtt сообщения.
    """

    def create(self, callback: Callable[[ObserverObject], None],  observing_objs: Iterable[ObserverObject]) -> None:
        """Создает обработчик события запускающийся через равные интервалы времени.

        По наступлению события происходит запуск обратного вызова по отправке mqtt сообщения.

        Args:
            callback (Callable[[list[ObserverObject]], None]): Обратный вызов по отправке mqtt сообщения.
            observing_objs (set[ObserverObject]): Наблюдаемый объект.
        """
        for observing_obj in observing_objs:
            if observing_obj.type == 'time_interval':
                thread: Thread = Thread(
                    target=time_range_loop,
                    args=(int(observing_obj.value), callback, observing_obj),
                )
                thread.start()


def closure_internal_event_handler(
    callback_func: Callable[[ObserverObject], None], observing_objs: ObserverObject,
) -> IEventHandler[Event]:
    """Замыкает обработчик события на запуск функции обратного вызова с предустановленным ObserverObject.

    Args:
        callback_func (Callable[[ObserverObject], None]): Функция обратного вызова.
        observing_objs (ObserverObject): Наблюдаемый объект.

    Returns:
        IEventHandler[Event]: Обработчик событий.
    """
    def event_handler(event: Event) -> None:
        callback_func(observing_objs)
    return event_handler


class InternalEventLoopCreator():
    """Создатель цикла обработки внутренних событий.

    По наступлению события происходит запуск обратного вызова по отправке mqtt сообщения.
    """

    def create(self, callback: Callable[[ObserverObject], None],  observing_objs: Iterable[ObserverObject]) -> None:
        """Регистрирует обработчики событий запускающееся по наступлению внутреннего события и запускает их обработку.

        По наступлению события происходит запуск обратного вызова по отправке mqtt сообщения.

        Args:
            callback (Callable[[list[ObserverObject]], None]): Обратный вызов по отправке mqtt сообщения.
            observing_objs (set[ObserverObject]): Наблюдаемый объект.
        """
        for observing_obj in observing_objs:
            if observing_obj.type == 'internal_event':
                register_handler(
                    closure_internal_event_handler(
                        callback, observing_obj,
                    ),
                    observing_obj.value,
                )
        thread: Thread = Thread(
            target=handle_event_from_msgbus,
        )
        thread.start()
