"""Содержит все необходимое для создания событий mqtt publisher'a базы данных."""
from typing import Any, Callable

from sqlalchemy import event
from sqlalchemy.orm import Session

from system_configurator.core.api.api_mqtt.publishers import ObserverObject
from system_configurator.core.db.database import SessionLocal


@event.listens_for(SessionLocal, 'before_commit')
def before_commit(session: Session, *args: Any, **kwargs: Any) -> None:
    """Hook executed before commit to the db.

    Add created, edited or deleted object to _edited_objects variable for future processing.

    Args:
        session (Session): Alchemy session.
    """
    session.tables_names_edited = set()
    for edited_obj in session.new | session.deleted | session.dirty:
        session.tables_names_edited.add(edited_obj.__table__.name)


class DbEventCallback:
    """Обратный вызов для базы данных."""

    def __init__(self, event_callback: Callable[[ObserverObject], None], observing_objs: set[ObserverObject]) -> None:
        self._event_callback = event_callback
        self._observing_objs = observing_objs

    def __call__(self, session: Session) -> None:
        """Hook executed after commit to the db.

        Запускает обратный вызов mqtt publisher'a

        Args:
            session (Session): Alchemy session.
        """
        for table_name in session.tables_names_edited:
            observed_obj = ObserverObject('table', table_name)
            if observed_obj in self._observing_objs:
                self._event_callback(observed_obj)


class DbEventCreator():
    """Создатель событий для базы данных.

    По наступлению события в бд происходит запуск обратного вызова по отправке mqtt сообщения.
    """

    def create(self, callback: Callable[[ObserverObject], None], observing_objs: set[ObserverObject]) -> None:
        """Создает событий для базы данных.

        По наступлению события в бд происходит запуск обратного вызова по отправке mqtt сообщения.

        Args:
            callback (Callable[[list[ObserverObject]], None]): Обратный вызов по отправке mqtt сообщения.
            observing_objs (set[ObserverObject]): Наблюдаемый объект.
        """
        db_event_callback = DbEventCallback(callback, observing_objs)
        event.listen(Session, 'after_commit', db_event_callback)
