"""Интерфейсы и классы для создания класса транзакция имплементирующего паттерн unit of work."""
from typing import Generic, Protocol

from sqlalchemy.orm import sessionmaker, Session

from system_configurator.core.db.database import SessionLocal
from system_configurator.core.db.repository import (
    IRepository,
    ISqlAlchemyRepositoryFabric,
    TSchema,
)


class ITransaction(Protocol):
    """Базовый интерфейс транзакции."""

    def commit(self) -> None:
        """Подтвердить транзакцию."""
        ...

    def rollback(self) -> None:
        """Откатить транзакцию."""
        ...


class ITransactionManager(Protocol, Generic[TSchema]):
    """Интерфейс для контекстного менеджера транзакции."""

    def __enter__(self) -> tuple[ITransaction, IRepository[TSchema]]:
        """Точка входа в контекстный менеджер. Возвращает транзакцию и репозиторий."""
        ...

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Выход из контекстного менеджера."""
        ...


class SqlAlchemyTransaction(ITransaction):
    """Транзакция SqlAlchemy.

    Args:
        session (Session): SqlAlchemy сессия.
    """

    def __init__(self, session: Session) -> None:
        self._session = session

    def commit(self) -> None:
        """Подтвердить транзакцию."""
        self._session.commit()

    def rollback(self) -> None:
        """Откатить транзакцию."""
        self._session.rollback()


class SqlAlchemyTransactionManager(ITransactionManager):
    """Контекстный менеджер для транзакции SqlAlchemy.

    Args:
        repository_fabric (ISqlAlchemyRepositoryFabric): фабрика по созданию репозиториев.
        session_fabric (sessionmaker, optional): Фабрика по созданию сессий SqlAlchemy. Defaults to SessionLocal.
    """

    transaction_class = SqlAlchemyTransaction  # Dependencies

    def __init__(
        self,
        repository_fabric: ISqlAlchemyRepositoryFabric,
        session_fabric: sessionmaker = SessionLocal,
    ) -> None:
        self._session_fabric = session_fabric
        self._repository_fabric = repository_fabric

    def __enter__(self) -> tuple[ITransaction, IRepository[TSchema]]:
        """Точка входа в контекстный менеджер. Возвращает транзакцию и репозиторий."""
        self._session = self._session_fabric()
        self._transaction = self.transaction_class(self._session)
        self._repository = self._repository_fabric(self._session)
        return (self._transaction, self._repository)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Выход из контекстного менеджера.

        Закрывает sqlalchemy сессию.

        Если менеджер закрылся с ошибкой то откатывает изменения.
        """
        if exc_type is not None:
            self._session.rollback()
        self._session.close()
