"""Шина сообщений."""

from abc import ABC
import logging
from queue import Queue
from typing import NoReturn, Protocol, Type, TypeVar, Union

from system_configurator.core.exceptions import Error


logger = logging.getLogger(__name__)


class Event(ABC):
    """Базовый класс события."""
    ...


T = TypeVar('T', bound=Event, contravariant=True)


class IEventHandler(Protocol[T]):
    """Интерфейс обработчика события."""

    def __call__(self, event: T) -> None:
        """Запускает обработку события.

        Args:
            event (T): Событие. Потомок класса Event.
        """
        ...


TEventHandlers = dict[Type[Event], list[IEventHandler[Event]]]


EVENTS: dict[str, Type[Event]] = {}

EVENT_HANDLERS: TEventHandlers = {}

MSGBUS: Queue[Event] = Queue()


def handle_event(
    event: Event,
    event_handlers: TEventHandlers = EVENT_HANDLERS,
) -> None:
    """Обрабатывает события.

    Args:
        event (Event): Событие.
        event_handlers (TEventHandlers, optional): Обработчик события. Defaults to EVENT_HANDLERS.
    """
    handlers = event_handlers.get(type(event), [])
    for handler in handlers:
        logger.debug("handling event %s with handler %s", event, handler)
        try:
            handler(event)
        except Error:
            logger.error("Error handling event %s with handler %s", event, handler)
            continue


def handle_event_from_msgbus(msg_bus: Queue[Event] = MSGBUS) -> NoReturn:
    """Обрабатывает события в цикле по мере их поступления.

    Args:
        msg_bus (Queue[Event], optional): Шина сообщений. Defaults to MSGBUS.
    """
    logger.debug('Starting event handling.')
    while True:
        event = msg_bus.get()
        handle_event(event)  # DEPENDENCY


def register_event(
    name: str, event_type: Type[Event], all_registered_events: dict[str, Type[Event]] = EVENTS,
) -> None:
    """Регистрирует новое событие в справочнике событий.

    Args:
        name (str): Имя события.
        event_type (Type[Event]): Имя класса события.
        all_registered_events (dict[str, Type[Event]], optional): Справочник всех событий. Defaults to EVENTS.
    """
    all_registered_events[name] = event_type
    logger.debug('Registered new event. %s', name)


def register_handler(
    handler: IEventHandler[Event],
    event_name: str,
    event_handlers: TEventHandlers = EVENT_HANDLERS,
) -> None:
    """Регистрирует новый обработчик события.

    Args:
        handler (IEventHandler[Event]): Обработчик события.
        event_name (str): Имя события.
        event_handlers (TEventHandlers, optional): Справочник событий и их обработчиков. Defaults to EVENT_HANDLERS.
    """
    event_type = EVENTS.get(event_name)
    if not event_type:
        logger.debug('Event handler %s was not registered for event because event not found. %s', handler, event_name)
        return
    if event_type not in event_handlers:
        event_handlers[event_type] = []
    event_handlers[event_type].append(handler)
    logger.debug('Registered new event handler %s for event %s', handler, event_name)
