"""Модуль взаимодействия с командной строкой."""
import logging
import subprocess
from typing import Optional

from system_configurator.core.exceptions import ShellCommandError


logger = logging.getLogger(__name__)


def run_shell_command(
    command: list[str], sudo: bool = False, check: bool = False, capture_output: bool = False,
) -> Optional[str]:
    """Запускает команду в терминале.

    Raises:
        ShellCommandError: raise if return code is not zero and check flag was set to true.

    Args:
        command (list[str]): Команда разбитая на список слов из которых она состоит.
        sudo (bool, optional): Запускать ли команду от имени sudo. Defaults to False.
        check (bool, optional): Проверять ли результат выполнения команды на ошибки. Defaults to False.
        capture_output (bool, optional): Захватывать ли вывод команды. Defaults to False.

    Returns:
        Optional[str]: Строковый вывод команды если параметр capture_output истинен иначе None.
    """
    def check_cmd_process(
        cmd_process: subprocess.CompletedProcess[bytes],
        logger: Optional[logging.Logger] = logger,
    ) -> None:
        """Check that cmd process return without problem.

        Check that return code is zero or logging stderr and raises error.

        Args:
            cmd_process (subprocess.CompletedProcess): Complete command process.

        Raises:
            subprocess.CalledProcessError: raise if return code is not zero.
        """
        if cmd_process.returncode != 0:
            if cmd_process.stderr and logger:
                for row in cmd_process.stderr.decode('utf-8').splitlines():
                    logger.error(row)
            raise ShellCommandError(
                f"Command '{cmd_process.args}' exit with none zeros status. Exit status {cmd_process.returncode}",
                data=cmd_process.stderr,
            )

    if sudo:
        command.insert(0, 'sudo')

    res = subprocess.run(command, capture_output=capture_output)

    if check:
        check_cmd_process(res)

    return res.stdout.decode('utf-8') if capture_output else None
