"""Publisher commands."""
from dataclasses import asdict
import logging

from system_configurator.core.api.api_mqtt.mqtt_clients import IMqttMsgSender
from system_configurator.core.api.api_mqtt import msg_sender
from system_configurator.core.api.api_mqtt.publishers import IPublisherCommand
from system_configurator.core.db.transaction import ITransactionManager

from system_configurator.external_access.entities import ICertificateInfo
from system_configurator.external_access.services.web_access_settings_services import (
    WebAccessSettingsDict,
    parse_nginx_config,
)
from system_configurator.external_access.db.uow import unit_of_work_manager


logger = logging.getLogger(__name__)


class PublishWebAccessSettingsCommand(IPublisherCommand):
    """Команда публикует настройки web доступа.

    Args:
        msg_sender (IMqttMsgSender, optional): Отправитель mqtt сообщений.

    Methods:
        __call__: Отправляет сообщение на публикацию.
    """

    def __init__(self, msg_sender: IMqttMsgSender = msg_sender) -> None:
        self._msg_sender = msg_sender

    def __call__(self, publisher_topic: str) -> None:
        """Отправляет сообщение с настройками web доступа на публикацию.

        Args:
            publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.
        """
        web_access_settings: WebAccessSettingsDict = parse_nginx_config()  # Dependencies
        self._msg_sender.send_msg(
            topic=publisher_topic,
            payload=web_access_settings,
            retain=True,
            cor_data=None,
        )
        logger.debug('Published Web Access settings')


class PublishAllCertificatesCommand(IPublisherCommand):
    """Команда для публикации параметров сертификатов.

    Args:
        msg_sender (IMqttMsgSender, optional): Отправитель mqtt сообщений.
        unit_of_work_cm (ITransactionManager[ICertificateInfo]): Контекстный менеджер транзакции.
        Открывает транзакцию к бд.

    Methods:
        __call__: Отправляет сообщение на публикацию.
    """

    def __init__(
        self,
        msg_sender: IMqttMsgSender = msg_sender,
        unit_of_work_cm: ITransactionManager[ICertificateInfo] = unit_of_work_manager,
    ) -> None:
        self._msg_sender = msg_sender
        self._uow_cm = unit_of_work_cm

    def __call__(self, publisher_topic: str) -> None:
        """Отправляет сообщение со списком всех сертификатов на публикацию.

        Args:
            publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.
        """
        with self._uow_cm as (_, repository):
            certificates = repository.get_list()
        self._msg_sender.send_msg(
            topic=publisher_topic,
            payload=[asdict(certificate) for certificate in certificates],
            retain=True,
            cor_data=None,
        )
        logger.debug('Published All certificates')
