"""Команды для обработки mqtt роутером."""
from abc import ABC
import logging
from typing import Optional, Type

from pydantic import ValidationError, BaseModel

from system_configurator.core.api.api_mqtt.entities import IRouterCommand, ResponsePayloadDict
from system_configurator.external_access.api.api_mqtt.schema import (
    CertificateRequestSchema,
    CertificateSchema,
    UploadCertificateSchema,
    UploadCorrespondingCsrInfo,
    WebAccessSettingsSchema,
)
from system_configurator.core.exceptions import ValidationException
from system_configurator.external_access.services.certificate_services import (
    DeleteCertificateService,
    GenerateCertService,
    UploadCertificateCorrespondingCsrService,
    UploadCertificateService,
)
from system_configurator.external_access.services.web_access_settings_services import (
    ChangeNginxConfigService,
)


logger = logging.getLogger(__name__)


class AbstractRouterCommand(ABC):
    """Абстрактный класс для обработки запроса от роутера.

    Class Attrs:
        service (Type[IRouterCommand]): Сервис бизнес логики для обработки payload.
        parser (Type[BaseModel]): Парсер для обработки чистых данных payload и проверки типов.
    """

    service: Type[IRouterCommand]
    parser: Type[BaseModel]

    def execute(self, data: bytes) -> Optional[ResponsePayloadDict]:
        """Запускает команду обработки mqtt payload.

        Args:
            data (bytes): тело запроса.
        """
        self._parse_input(data)
        self._run_service()
        return {'status': 'success', 'msg': 'OK', 'data': None}

    def _parse_input(self, data: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._data_parsed = self.parser.parse_raw(data)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())

        logger.debug('Input data validated successful.')

    def _run_service(self) -> None:
        self.service(self._data_parsed).execute()


class GenerateSelfSignedCertCommand(AbstractRouterCommand):
    """Генерирует самоподписанный сертификат."""

    service = GenerateCertService
    parser = CertificateRequestSchema


class GenerateCsrCommand(AbstractRouterCommand):
    """Генерирует Certificate Signing Request."""

    service = GenerateCertService
    parser = CertificateRequestSchema


class DeleteCertCommand(AbstractRouterCommand):
    """Удаляет сертификат."""

    service = DeleteCertificateService
    parser = CertificateSchema


class UploadCertCommand(AbstractRouterCommand):
    """Загружает сертификат и его ключ."""

    service = UploadCertificateService
    parser = UploadCertificateSchema


class UploadCertCorrespondingCsrCommand(AbstractRouterCommand):
    """Загружает сертификат относящийся к созданному ранее csr."""

    service = UploadCertificateCorrespondingCsrService
    parser = UploadCorrespondingCsrInfo


class ChangeNginxConfigCommand(AbstractRouterCommand):
    """Изменяет конфигурацию web доступа."""

    service = ChangeNginxConfigService
    parser = WebAccessSettingsSchema
