"""Pydantic Schemas."""
from typing import Literal, Union

from pydantic import BaseModel


class CertificateRequestSchema(BaseModel):
    """Схема для запроса генерации нового сертификата.

    Args:
        cert_name (str):                Читаемое имя сертификата.
        cert_type (TCertificateTypes):  Тип сертификата.
        Может принимать значения 'csr' или 'certificate'
        key_size (Literal[2048, 4096]): Размер ключа в байтах.
        Принимает значения 2048 иои 4096.
        subject (TSubjectString):       Строка в формате rfc4514.
        san (TSanString):               Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.
    """
    cert_name: str
    cert_type: Literal['certificate', 'csr']
    key_size: Literal[2048, 4096] = 2048
    subject: str
    san: str = ''


class CertParamsSchema(BaseModel):
    """Схема для параметров сертификата.

    Args:
        subject (TSubjectString):   Строка в формате rfc4514.
        san (TSanString):           Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую
        без пробелов с префиксами `IP=` или `DNS=`.
        issuer (str):               Строка в формате rfc4514.
        valid_from (float):         Дата с которой сертификат действителен. Формат Posix timestamp.
        valid_to (float):           Дата по которую сертификат действителен. Формат Posix timestamp.
    """
    subject: str
    san: str
    issuer: str
    valid_from: float
    valid_to: float


class CsrParamsSchema(BaseModel):
    """Схема для параметров csr.

    Args:
        subject (TSubjectString):   Строка в формате rfc4514.
        san (TSanString):           Стока представляющее расширение SubjectAltName.
    """
    subject: str
    san: str


class CertificateSchema(BaseModel):
    """Схема x509 сертификата.

    Args:
        id (Optional[str]): id сертификата.
        name (str): Имя сертификата
        cert_type (str): Тип сертификата. Может принимать значения 'csr' или 'certificate'
        public_bytes (str): Открытый ключ сертификата.
        params (Union[ICertParamsInfo, ICsrParamsInfo]): Словарь с параметрами сертификата.
        Набор параметров отличается в зависимости от типа сертификата.
    """

    id: str = None
    name: str
    cert_type: Literal['certificate', 'csr']
    public_bytes: str
    params: Union[CertParamsSchema, CsrParamsSchema]


class UploadCertificateSchema(BaseModel):
    """Схема загрузки нового сертификата.

    Args:
        cert_name (str):        - Читаемое имя сертификата.
        certificate (bytes)     - x.509 сертификат.
        key (bytes)             - Приватный ключ в pem формате.
        intermediate (bytes)    - (Опционально) промежуточный сертификат.
    """

    cert_name: str
    certificate: bytes
    key: bytes
    intermediate: bytes = b''


class UploadCorrespondingCsrInfo(BaseModel):
    """Схема загрузки сертификата относящегося к csr..

    Args:
        cert_name (str):        - Читаемое имя csr сертификата.
        certificate (bytes)         - x.509 сертификат.
    """

    cert_name: str
    certificate: bytes


class WebAccessSettingsSchema(BaseModel):
    """Web Access settings Info.

    http_port (int):             Http порт. По умолчанию 80.
    https_port (int):            Https порт. По умолчанию 443.
    is_https_enabled (bool):     Индикатор включен ли https. По умолчанию False.
    is_http_redirected (bool):   Индикатор включена ли переадресация http to https. По умолчанию False.
    cert_name (str):             Имя сертификата сервера. По умолчанию пустое.
    """

    http_port: int = 80
    https_port: int = 443
    is_https_enabled: bool = False
    is_http_redirected: bool = False
    cert_name: str = ''
