"""Репозитории для работы с бд."""
from __future__ import annotations
from typing import Any

import sqlalchemy
from sqlalchemy import select
from sqlalchemy.orm import Session

from system_configurator.core.db.repository import (
    IRepository,
    ISqlAlchemyRepositoryFabric,
)
from system_configurator.external_access.db.models import (
    CertParam,
    SpCertType,
    SpCertParam,
    SpParam,
    Certificate,
)
from system_configurator.core.exceptions import (
    NotFoundException,
)
from system_configurator.external_access.entities import (
    ICertificateInfo,
    CertParamsInfo,
    CertificateInfo,
    CsrParamsInfo,
)


class DbCCertificateRepositoryFabric(ISqlAlchemyRepositoryFabric[ICertificateInfo]):
    """фабрика по созданию репозитория модели сертификата."""

    def __call__(self, session: Session) -> DbCertificateRepository:
        """Создает репозиторий модели сертификата.

        Репозиторий скрывает взаимодействие с интерфейсами SqlAlchemy.

        Args:
            session (Session): SqlAlchemy сессия.

        Returns:
            IRepository[TSchema]: Репозиторий.
        """
        return DbCertificateRepository(session)


class DbCertificateRepository(IRepository[ICertificateInfo]):
    """Репозиторий модели сертификата.

    Репозиторий скрывает взаимодействие с интерфейсами SqlAlchemy.

    Args:
        session (Session): SqlAlchemy сессия.
    """

    def __init__(self, session: Session) -> None:
        self._session = session

    def create(self, schema: ICertificateInfo) -> CertificateInfo:
        """Добавляет объект модели сертификата в репозиторий.

        Args:
            schema (ICertificateInfo): Объект модели сертификата.

        Returns:
            TSchema: Объект модели сертификата.
        """
        cert_obj = create_certificate_db_obj(self._session, schema)
        return serialize_cert_obj_to_dataclass(cert_obj)

    def delete(self, schema: ICertificateInfo) -> None:
        """Удаляет объект модели сертификата из репозитория.

        Args:
            schema (ICertificateInfo): Объект модели сертификата.
        """
        delete_certificate_db_obj(self._session, schema)

    def update(self, schema: ICertificateInfo) -> CertificateInfo:
        """Обновляет объект модели сертификата в репозитории.

        Args:
            schema (ICertificateInfo): Объект модели сертификата.

        Returns:
            TSchema: Объект модели сертификата.
        """
        delete_certificate_db_obj(self._session, schema)
        cert_obj = create_certificate_db_obj(self._session, schema)
        return serialize_cert_obj_to_dataclass(cert_obj)

    def get(self, name: str) -> CertificateInfo:
        """Возвращает объект модели сертификата из репозитория по его имени.

        Args:
            name (ICertificateInfo): Имя объекта модели сертификата.

        Returns:
            ICertificateInfo: Объект модели сертификата.
        """
        cert_obj = get_certificate_obj_by_name(self._session, name)
        return serialize_cert_obj_to_dataclass(cert_obj)

    def get_list(self) -> list[CertificateInfo]:
        """Возвращает список всех объектов модели сертификата из репозитория.

        Returns:
            list[ICertificateInfo]: Список всех объектов модели сертификата репозитория.
        """
        cert_objs = get_all_certificate_objs(self._session)
        return serialize_certificates_to_dataclass(cert_objs)


def create_certificate_db_obj(session: Session,  cert_schema: ICertificateInfo) -> Certificate:
    """Создает запись в таблице Certificate.

    Args:
        session (Session):   Сессия к базе данных.
        cert_schema (ICertificateInfo): Объект с настройками сертификата.

    Raises:
        AlreadyExistError: Происходит если объект в бд с таким именем уже существует.

    Returns:
        Certificate: Объект ORM таблицы Certificate.
    """
    query = (
        select(SpCertType.id, SpCertParam.id, SpParam.name)
        .join(SpCertParam, SpCertType.id == SpCertParam.id_sp_cert_type)
        .join(SpParam, SpCertParam.id_sp_param == SpParam.id)
        .where(SpCertType.name == cert_schema.cert_type)
    )
    query_result = session.execute(query)
    cert_attrs: list[tuple[int, int, str]] = query_result.all()

    # Create new trigger
    id_sp_c_type: int = cert_attrs[0][0]
    new_cert_obj: Certificate = Certificate(
        name=cert_schema.name, id_sp_c_type=id_sp_c_type, public_bytes=cert_schema.public_bytes,
    )
    session.add(new_cert_obj)

    for _, sp_c_p_id, sp_param_name in cert_attrs:
        trigger_param = CertParam(
            id_sp_c_p=sp_c_p_id,
            value=getattr(cert_schema.params, sp_param_name),
        )
        new_cert_obj.params.append(trigger_param)
    return new_cert_obj


def delete_certificate_db_obj(session: Session,  cert_schema: ICertificateInfo) -> None:
    """Удаляет запись из таблицы Certificate.

    Args:
        session (Session):   Сессия к базе данных.
        cert_schema (ICertificateInfo): Объект с настройками сертификата.

    Raises:
        NotFoundError: Происходит если объект в бд не найден.
    """
    cert_ob = get_certificate_obj_by_name(session, cert_schema.name)
    session.delete(cert_ob)


def get_certificate_obj_by_name(session: Session,  name: str) -> Certificate:
    """Находит запись в таблице сертификатов по имени.

    Args:
        session (Session):      Сессия к базе данных.
        name (str):             Имя сертификата.

    Raises:
        NotFoundException: Происходит если объект в бд не найден.
    """
    query = select(Certificate).where(Certificate.name == name)
    query_result = session.execute(query)
    try:
        cert_obj = query_result.scalars().one()
    except sqlalchemy.exc.NoResultFound:
        raise NotFoundException(f'Certificate with name {name} not found into db.')
    return cert_obj


def serialize_cert_obj_to_dict(cert_obj: Certificate) -> dict[str, Any]:
    """Сериализует объект триггера в словарь.

    Args:
        cert_obj (Certificate): объект сертификата из орм.

    Returns:
        dict[str, Any]: Сериализованный сертификат.
    """
    serialized_data = {
        'name': cert_obj.name,
        'cert_type': cert_obj.cert_type.name,
        'params': {param.name: param.value for param in cert_obj.params},
    }
    return serialized_data


def serialize_cert_obj_to_dataclass(cert_obj: Certificate) -> CertificateInfo:
    """Сериализует объект триггера в словарь.

    Args:
        cert_obj (Certificate): объект сертификата из орм.

    Returns:
        dict[str, Any]: Сериализованный сертификат.
    """
    if cert_obj.cert_type.name == 'csr':
        params = CsrParamsInfo(**{param.name: param.value for param in cert_obj.params})
    elif cert_obj.cert_type.name == 'certificate':
        params = CertParamsInfo(**{param.name: param.value for param in cert_obj.params})
    serialized_data = CertificateInfo(
        name=cert_obj.name,
        cert_type=cert_obj.cert_type.name,
        params=params,
        public_bytes=cert_obj.public_bytes,
    )
    return serialized_data


def get_all_certificate_objs(session: Session) -> list[Certificate]:
    """Возвращает список всех записей таблицы Certificate.

    Args:
        session (Session): SqlAlchemy session.

    Returns:
        list[Certificate]: Cписок всех записей таблицы Certificate.
    """
    query = select(Certificate)
    query_result = session.execute(query)
    cert_objs = query_result.scalars().all()
    return cert_objs


def serialize_certificates_to_dict(cert_objs: list[Certificate]) -> list[dict[str, Any]]:
    """Формирует список с параметрами сертификатов.

    Получает список всех сертификатов из базы данных и сериализует их в список словарей.

    Args:
        session (Session):   Сессия к базе данных.

    Returns:
        list[dict[str, Any]]: список словарей с параметрами сертификатов.
    """
    certificates: list[dict[str, Any]] = []
    for cert_obj in cert_objs:
        cert_serialized = serialize_cert_obj_to_dict(cert_obj)
        certificates.append(cert_serialized)
    return certificates


def serialize_certificates_to_dataclass(cert_objs: list[Certificate]) -> list[CertificateInfo]:
    """Сериализует записи таблицы Certificate в список моделей бизнес логики.

    Args:
        cert_objs (list[Certificate]): записи таблицы Certificate.

    Returns:
        list[CertificateInfo]: список моделей CertificateInfo.
    """
    certificates: list[CertificateInfo] = []
    for cert_obj in cert_objs:
        cert_serialized = serialize_cert_obj_to_dataclass(cert_obj)
        certificates.append(cert_serialized)
    return certificates
