"""Модели бизнес логики."""
from dataclasses import dataclass
from typing import Literal, Protocol, TypedDict, Union


TCertificateTypes = Literal['certificate', 'csr']
TRawCertBytes = bytes
TRawKeyBytes = bytes
TSanString = str
TSubjectString = str


class ICertificateRequest(Protocol):
    """Интерфейс для параметров генерации нового сертификата.

    Args:
        cert_name (str):                Читаемое имя сертификата.
        cert_type (TCertificateTypes):  Тип сертификата.
        Может принимать значения 'csr' или 'certificate'
        key_size (Literal[2048, 4096]): Размер ключа в байтах.
        Принимает значения 2048 иои 4096.
        subject (TSubjectString):       Строка в формате rfc4514.
        san (TSanString):               Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.
    """
    cert_name: str
    cert_type: TCertificateTypes
    key_size: Literal[2048, 4096]
    subject: TSubjectString
    san: TSanString


class ICertParamsInfo(Protocol):
    """Интерфейс для параметров сертификата.

    Args:
        subject (TSubjectString):   Строка в формате rfc4514.
        san (TSanString):           Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую
        без пробелов с префиксами `IP=` или `DNS=`.
        issuer (str):               Строка в формате rfc4514.
        valid_from (float):         Дата с которой сертификат действителен. Формат Posix timestamp.
        valid_to (float):           Дата по которую сертификат действителен. Формат Posix timestamp.
    """

    subject: TSubjectString
    san: TSanString
    issuer: str
    valid_from: float
    valid_to: float


class ICsrParamsInfo(Protocol):
    """Интерфейс для параметров csr.

    Args:
        subject (TSubjectString):   Строка в формате rfc4514.
        san (TSanString):           Стока представляющее расширение SubjectAltName.
    """
    subject: TSubjectString
    san: TSanString


class ICertificateInfo(Protocol):
    """Интерфейс для модели сертификата x509.

    Args:
        name (str): Имя сертификата
        cert_type (str): Тип сертификата. Может принимать значения 'csr' или 'certificate'
        public_bytes (str): Открытый ключ сертификата.
        params (Union[ICertParamsInfo, ICsrParamsInfo]): Словарь с параметрами сертификата.
        Набор параметров отличается в зависимости от типа сертификата.
    """
    name: str
    cert_type: TCertificateTypes
    public_bytes: str

    @property
    def params(self) -> Union[ICertParamsInfo, ICsrParamsInfo]: ...


class IUploadCertificateInfo(Protocol):
    """Интерфейс для параметров загрузки нового сертификата.

    Args:
        cert_name (str):        - Читаемое имя сертификата.
        certificate (bytes)     - x.509 сертификат.
        key (bytes)             - Приватный ключ в pem формате.
        intermediate (bytes)    - (Опционально) промежуточный сертификат.
    """

    cert_name: str
    certificate: bytes
    key: bytes
    intermediate: bytes


class IUploadCorrespondingCsrInfo(Protocol):
    """Интерфейс для параметров загрузки сертификата относящегося к csr.

    Args:
        cert_name (str):        - Читаемое имя csr сертификата.
        certificate (bytes):    - x.509 сертификат.
    """

    cert_name: str
    certificate: bytes


class IWebAccessSettingsInfo(Protocol):
    """Interface for Web Access settings Info.

    http_port (int):             Http порт. По умолчанию 80.
    https_port (int):            Https порт. По умолчанию 443.
    is_https_enabled (bool):     Индикатор включен ли https. По умолчанию False.
    is_http_redirected (bool):   Индикатор включена ли переадресация http to https. По умолчанию False.
    cert_name (str):             Имя сертификата сервера. По умолчанию пустое.
    """

    http_port: int
    https_port: int
    is_https_enabled: bool
    is_http_redirected: bool
    cert_name: str


class WebAccessSettingsDict(TypedDict):
    """Web Access settings Info.

    http_port (int):             Http порт. По умолчанию 80.
    https_port (int):            Https порт. По умолчанию 443.
    is_https_enabled (bool):     Индикатор включен ли https. По умолчанию False.
    is_http_redirected (bool):   Индикатор включена ли переадресация http to https. По умолчанию False.
    cert_name (str):             Имя сертификата сервера. По умолчанию пустое.
    """

    http_port: int
    https_port: int
    is_https_enabled: bool
    is_http_redirected: bool
    cert_name: str


@dataclass(frozen=True)
class WebAccessSettingsInfo:
    """Web Access settings Info.

    http_port (int):             Http порт. По умолчанию 80.
    https_port (int):            Https порт. По умолчанию 443.
    is_https_enabled (bool):     Индикатор включен ли https. По умолчанию False.
    is_http_redirected (bool):   Индикатор включена ли переадресация http to https. По умолчанию False.
    cert_name (str):             Имя сертификата сервера. По умолчанию пустое.
    """

    http_port: int
    https_port: int
    is_https_enabled: bool
    is_http_redirected: bool
    cert_name: str


@dataclass
class CertParamsInfo:
    """Модель для параметров сертификата.

    Args:
        subject (TSubjectString):   Строка в формате rfc4514.
        san (TSanString):           Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую
        без пробелов с префиксами `IP=` или `DNS=`.
        issuer (str):               Строка в формате rfc4514.
        valid_from (float):         Дата с которой сертификат действителен. Формат Posix timestamp.
        valid_to (float):           Дата по которую сертификат действителен. Формат Posix timestamp.
    """

    subject: str
    issuer: str
    san: str
    valid_from: float
    valid_to: float


@dataclass
class CsrParamsInfo:
    """Модель для параметров csr.

    Args:
        subject (TSubjectString):   Строка в формате rfc4514.
        san (TSanString):           Стока представляющее расширение SubjectAltName.
    """

    subject: str
    san: str


@dataclass
class CertificateInfo:
    """Модель x509 сертификата.

    Args:
        name (str): Имя сертификата
        cert_type (str): Тип сертификата. Может принимать значения 'csr' или 'certificate'
        public_bytes (str): Открытый ключ сертификата.
        params (Union[ICertParamsInfo, ICsrParamsInfo]): Словарь с параметрами сертификата.
        Набор параметров отличается в зависимости от типа сертификата.
    """

    name: str
    cert_type: TCertificateTypes
    params: Union[ICertParamsInfo, ICsrParamsInfo]
    public_bytes: str
