"""Сервисы для настройки web доступа."""
import datetime
from ipaddress import IPv4Address
import logging
import pathlib
from typing import Callable, Union

from cryptography import x509
from cryptography.x509.oid import ExtensionOID
from cryptography.x509.extensions import SubjectAlternativeName
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa

from system_configurator import settings
from system_configurator.core.exceptions import (
    AlreadyExistError,
    NotFoundException,
    ValidationException,
)
from system_configurator.core.db.transaction import ITransactionManager
from system_configurator.external_access.entities import (
    CertParamsInfo,
    CertificateInfo,
    CsrParamsInfo,
    ICertParamsInfo,
    ICertificateInfo,
    ICertificateRequest,
    ICsrParamsInfo,
    TSubjectString,
    IUploadCertificateInfo,
    IUploadCorrespondingCsrInfo,
    TCertificateTypes,
    TRawCertBytes,
    TRawKeyBytes,
    TSanString,
)
from system_configurator.external_access.services.web_access_settings_services import (
    WebAccessSettingsDict,
    parse_nginx_config,
)
from system_configurator.external_access.db.uow import unit_of_work_manager


logger = logging.getLogger(__name__)


TCertParamsSerializer = Callable[[TRawCertBytes], Union[ICertParamsInfo, ICsrParamsInfo]]
TCertGenerator = Callable[[TRawKeyBytes, TSubjectString, TSanString], TRawCertBytes]


def generate_rsa_key(key_size: int = 2048) -> bytes:
    """Генерирует rsa ключ.

    Args:
        key_size (int, optional): Размер ключа в байтах. Defaults to 2048.

    Returns:
        bytes: rsa ключ.
    """
    key: rsa.RSAPrivateKey = rsa.generate_private_key(
        public_exponent=65537,
        key_size=key_size,
    )
    return key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )


def generate_self_signed_certificate(
    private_key: bytes,
    subject: str,
    san: str,
) -> bytes:
    """Генерирует самоподписанный сертификат.

    Args:
        private_key (bytes): Приватный ключ.
        subject (str): Строка в формате rfc4514.
        san (str): Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.

    Returns:
        bytes: Самоподписанный сертификат.
    """
    key = load_pem_private_key(private_key, None)

    subject_serialized = issuer = x509.Name.from_rfc4514_string(subject)
    cert_builder = x509.CertificateBuilder().subject_name(
        subject_serialized,
    ).issuer_name(
        issuer,
    ).public_key(
        key.public_key(),
    ).serial_number(
        x509.random_serial_number(),
    ).not_valid_before(
        datetime.datetime.utcnow(),
    ).not_valid_after(
        # Our certificate will be valid for 3 years
        datetime.datetime.utcnow() + datetime.timedelta(days=1095)
    )
    if san:
        cert_builder = cert_builder.add_extension(
            x509.SubjectAlternativeName(
                serialize_san_to_cryptography_name_objects(san),  # dependencies
            ),
            critical=False,
        )
    cert = cert_builder.sign(key, hashes.SHA256())
    return cert.public_bytes(serialization.Encoding.PEM)


def generate_csr(
    private_key: bytes,
    subject: str,
    san: str = '',
) -> bytes:
    """Генерирует csr.

    Args:
        private_key (bytes): Приватный ключ.
        subject (str): Строка в формате rfc4514.
        san (str): Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.

    Returns:
        bytes: csr.
    """
    key = load_pem_private_key(private_key, None)

    subject_serialized = x509.Name.from_rfc4514_string(subject)
    csr_builder = x509.CertificateSigningRequestBuilder(
    ).subject_name(subject_serialized)
    if san:
        csr_builder = csr_builder.add_extension(
            x509.SubjectAlternativeName(
                serialize_san_to_cryptography_name_objects(san),  # dependencies
            ),
            critical=False,
        )
    csr = csr_builder.sign(key, hashes.SHA256())
    return csr.public_bytes(serialization.Encoding.PEM)


CERT_GENERATOR_FABRIC: dict[TCertificateTypes, TCertGenerator] = {
    'certificate': generate_self_signed_certificate,
    'csr': generate_csr,
}


def generate_raw_certificate_fabric(
    cert_type: TCertificateTypes,
    subject: TSubjectString,
    san: TSanString,
    key_raw: TRawKeyBytes,
) -> TRawCertBytes:
    """Генерирует сертификат согласно его типу.

    Args:
        cert_type (TCertificateTypes):  Тип сертификата.
        subject (TSubjectString):       Строка в формате rfc4514.
        san (TSanString):               Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.
        key_raw (TRawKeyBytes):         Приватный ключ.

    Returns:
        TRawCertBytes: x509 сертификат.
    """
    cert_raw: TRawCertBytes = CERT_GENERATOR_FABRIC[cert_type](key_raw, subject, san)  # Dependencies
    return cert_raw


def serialize_subject_rfc4514_string_to_cryptography_name_object(subject: str) -> x509.Name:
    """Сериализует строку в формате rfc4514 в объект x509.Name библиотеки cryptography.

    Args:
        subject (str): Строка в формате rfc4514.

    Returns:
        x509.Name: объект x509.Name библиотеки cryptography.
    """
    return x509.Name.from_rfc4514_string(subject)


def serialize_san_to_cryptography_name_objects(san_string: str) -> list[x509.GeneralName]:
    """Сериализует строку san в объект x509.GeneralName библиотеки cryptography.

    Args:
        san_string (str): Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.

    Returns:
        list[x509.GeneralName]: объект x509.GeneralName библиотеки cryptography.
    """
    san_serialized: list[x509.GeneralName] = []
    for name in san_string.split(','):
        if name.startswith('DNS='):
            san_serialized.append(x509.DNSName(name[4:]))
        elif name.startswith('IP='):
            san_serialized.append(x509.IPAddress(IPv4Address(name[3:])))
    return san_serialized


def serialize_from_cryptography_san_to_str(san_obj: SubjectAlternativeName) -> str:
    """Сериализует объект x509.GeneralName библиотеки cryptography в строку san.

    Args:
        san_obj (SubjectAlternativeName): Объект x509.GeneralName библиотеки cryptography.

    Returns:
        str: Стока представляющее расширение SubjectAltName.
        Принимаются только ip адреса или dns имена идущие подряд через запятую без пробелов
        с префиксами `IP=` или `DNS=`.
    """
    ip_addresses: list[IPv4Address] = san_obj.get_values_for_type(x509.IPAddress)
    dns_names: list[str] = san_obj.get_values_for_type(x509.DNSName)

    ip_addresses_strings: list[str] = ['IP=' + str(ip) for ip in ip_addresses]
    dns_names_strings: list[str] = ['DNS=' + dns_name for dns_name in dns_names]

    return ','.join(ip_addresses_strings + dns_names_strings)


def serialize_csr_params_from_bytes(cert_raw: bytes) -> CsrParamsInfo:
    """Сериализует параметры csr из байтов.

    Args:
        cert_raw (bytes): Байтовое представление csr.

    Returns:
        CsrParamsInfo: Параметры csr.
    """
    cert = x509.load_pem_x509_csr(cert_raw)
    try:
        san_extension = cert.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
    except x509.ExtensionNotFound:
        san_string = ''
    else:
        san_string = serialize_from_cryptography_san_to_str(  # Dependencies
            san_extension.value,
        )

    params = CsrParamsInfo(  # Dependencies
        subject=cert.subject.rfc4514_string(),
        san=san_string,
    )
    return params


def serialize_pem_certificate_params_from_bytes(cert_raw: bytes) -> CertParamsInfo:
    """Сериализует параметры сертификата из байтов.

    Args:
        cert_raw (bytes): Байтовое представление сертификата в pem формате.

    Returns:
        CertParamsInfo: Параметры сертификата.
    """
    cert = x509.load_pem_x509_certificate(cert_raw)
    try:
        san_extension = cert.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
    except x509.ExtensionNotFound:
        san_string = ''
    else:
        san_string = serialize_from_cryptography_san_to_str(  # Dependencies
            san_extension.value,
        )
    params = CertParamsInfo(  # Dependencies
        subject=cert.subject.rfc4514_string(),
        issuer=cert.issuer.rfc4514_string(),
        san=san_string,
        valid_from=cert.not_valid_before.timestamp(),
        valid_to=cert.not_valid_after.timestamp(),
    )
    return params


CERT_PARAMS_SERIALIZERS_FABRIC: dict[TCertificateTypes, TCertParamsSerializer] = {
    'certificate': serialize_pem_certificate_params_from_bytes,
    'csr': serialize_csr_params_from_bytes,
}


def serialize_certificate_from_bytes(cert_name: str, cert_type: TCertificateTypes, cert_raw: bytes) -> CertificateInfo:
    """Сериализует модель сертификата из байтов согласно его типу.

    Args:
        cert_name (str): Имя сертификата.
        cert_type (TCertificateTypes): Тип сертификата.
        cert_raw (bytes): Байтовое представление сертификата.

    Returns:
        CertificateInfo: модель сертификата.
    """
    cert_params = CERT_PARAMS_SERIALIZERS_FABRIC[cert_type](cert_raw)  # Dependencies
    schema = CertificateInfo(  # Dependencies
        name=cert_name,
        cert_type=cert_type,
        params=cert_params,
        public_bytes=cert_raw.decode('utf-8'),
    )
    return schema


def checkSubjectIsRfc4514String(subject: TSubjectString) -> None:
    """Проверяет что subject строка является валидной rfc4514 строкой.

    Args:
        subject (TSubjectString): _description_

    Raises:
        ValidationException: Выбрасывается если subject строка не валидна.
    """
    logger.debug('Check Subject string.')
    try:
        x509.Name.from_rfc4514_string(subject)
    except ValueError as err:
        raise ValidationException(message=str(err))
    else:
        logger.debug('Subject checked.')


class GenerateCertService():
    """Генерирует самоподписанный сертификат.

    Args:
        data (bytes): тело запроса.
    """

    def __init__(
        self,
        data: ICertificateRequest,
        unit_of_work_cm: ITransactionManager[ICertificateInfo] = unit_of_work_manager,
    ) -> None:
        self._cert_request = data
        self._uow_cm = unit_of_work_cm
        # Dependencies
        self._ssl_cert_path = pathlib.Path(settings.SSL_CERT_PATH)
        self._ssl_key_path = pathlib.Path(settings.SSL_KEY_PATH)
        self._ssl_cert_ext = settings.SSL_CERT_EXTENSION
        self._ssl_key_ext = settings.SSL_KEY_EXTENSION

    def execute(self) -> None:
        """Run Command."""
        self._checks()
        self._generate_pk()
        self._save_pk_file()
        self._generate_certificate()
        self._save_certificate_file()
        self._save_cert_params_into_db()

    def _checks(self) -> None:
        with self._uow_cm as (_, repository):
            certificates_existing: list[ICertificateInfo] = repository.get_list()
        for cert in certificates_existing:
            if cert.name == self._cert_request.cert_name:
                raise AlreadyExistError(f'The certificate {self._cert_request.cert_name} already exist!')

        checkSubjectIsRfc4514String(self._cert_request.subject)  # Dependencies

    def _generate_pk(self) -> None:
        logger.debug('Generating private key')
        self._key_raw: bytes = generate_rsa_key(self._cert_request.key_size)
        logger.debug('Private key generated.')

    def _save_pk_file(self) -> None:
        logger.debug('Saving private key file to storage.')
        pk_filename: str = self._cert_request.cert_name + self._ssl_key_ext
        with open(self._ssl_key_path / pk_filename, 'wb') as pk_file:
            pk_file.write(self._key_raw)
        logger.debug('Private key file saved.')

    def _generate_certificate(self) -> None:
        logger.debug('Generating certificate.')
        self._cert_raw: TRawCertBytes = generate_raw_certificate_fabric(  # Dependencies
            cert_type=self._cert_request.cert_type,
            subject=self._cert_request.subject,
            san=self._cert_request.san,
            key_raw=self._key_raw,
        )
        logger.debug('Self signed certificate generated.')

    def _save_certificate_file(self) -> None:
        logger.debug('Saving certificate file to storage.')
        cert_filename: str = self._cert_request.cert_name + self._ssl_cert_ext
        with open(self._ssl_cert_path / cert_filename, 'wb') as cert_file:
            cert_file.write(self._cert_raw)
        logger.debug('Certificate file saved.')

    def _save_cert_params_into_db(self) -> None:
        logger.debug('Saving certificate obj into db')
        schema = serialize_certificate_from_bytes(  # Dependencies
            cert_name=self._cert_request.cert_name,
            cert_type=self._cert_request.cert_type,
            cert_raw=self._cert_raw,
        )
        with self._uow_cm as (transaction, repository):
            repository.create(schema)
            transaction.commit()
        logger.debug('Certificate obj was saved into db')


class UploadCertificateService():
    """Загружает сертификат и его ключ.

    Args:
        data (bytes): тело запроса.

    Returns:
        None
    """

    def __init__(
        self,
        data: IUploadCertificateInfo,
        unit_of_work_cm: ITransactionManager[ICertificateInfo] = unit_of_work_manager,
    ) -> None:
        self._data = data
        self._uow_cm = unit_of_work_cm
        # Dependencies
        self._ssl_cert_path = pathlib.Path(settings.SSL_CERT_PATH)
        self._ssl_key_path = pathlib.Path(settings.SSL_KEY_PATH)
        self._ssl_cert_ext = settings.SSL_CERT_EXTENSION
        self._ssl_key_ext = settings.SSL_KEY_EXTENSION

    def execute(self) -> None:
        """Run Command."""
        self._checks()
        self._save_certificate_file()
        self._save_pk_file()
        self._save_cert_params_into_db()

    def _checks(self) -> None:  # TODO:Разобраться с checks
        logger.debug('Checking if certificate is not already exist.')
        with self._uow_cm as (_, repository):
            try:
                repository.get(self._data.cert_name)
            except NotFoundException:
                pass
            else:
                raise AlreadyExistError(f'The certificate {self._data.cert_name} already exist!')

        logger.debug('Checking if the certificate file is valid.')
        try:
            x509.load_pem_x509_certificate(self._data.certificate)
        except ValueError:
            raise ValidationException(message='Payload validation error.', data=['The certificate file is not valid.'])

        if self._data.intermediate:
            logger.debug('Checking if the intermediate certificate file is valid.')
            try:
                x509.load_pem_x509_certificate(self._data.intermediate)
            except ValueError:
                raise ValidationException(
                    message='Payload validation error.',
                    data=['The intermediate certificate file is not valid.'],
                )

        logger.debug('Checking if the key file is valid.')
        try:
            load_pem_private_key(self._data.key, None)
        except ValueError:
            raise ValidationException(message='Payload validation error.', data=['The key file is not valid.'])

        logger.debug('Checks passed.')

    def _save_certificate_file(self) -> None:
        logger.debug('Saving certificate file')

        cert_filename = f'{self._data.cert_name}{self._ssl_cert_ext}'
        with open(self._ssl_cert_path / cert_filename, 'wb') as file_cert:
            file_cert.write(self._data.certificate)
            if self._data.intermediate:
                file_cert.write(self._data.intermediate)

        logger.debug('Certificate file saved.')

    def _save_pk_file(self) -> None:
        logger.debug('Saving private key file')

        key_filename = f'{self._data.cert_name}{self._ssl_key_ext}'
        with open(self._ssl_key_path / key_filename, 'wb') as file_key:
            file_key.write(self._data.key)

        logger.debug('Private key file saved.')

    def _save_cert_params_into_db(self) -> None:
        logger.debug('Saving certificate obj into db')
        schema = serialize_certificate_from_bytes(
            cert_name=self._data.cert_name,
            cert_type='certificate',
            cert_raw=self._data.certificate)
        with self._uow_cm as (transaction, repository):
            repository.create(schema)
            transaction.commit()
        logger.debug('Certificate obj was saved into db')


class UploadCertificateCorrespondingCsrService():
    """Загружает сертификат относящийся к csr.

    Args:
        data (bytes): тело запроса.

    Returns:
        None
    """

    def __init__(
        self,
        data: IUploadCorrespondingCsrInfo,
        unit_of_work_cm: ITransactionManager[ICertificateInfo] = unit_of_work_manager,
    ) -> None:
        self._data = data
        self._uow_cm = unit_of_work_cm
        # Dependencies
        self._ssl_cert_path = pathlib.Path(settings.SSL_CERT_PATH)
        self._ssl_cert_ext = settings.SSL_CERT_EXTENSION

    def execute(self) -> None:
        """Run Command."""
        self._get_cert_model()
        self._checks()
        self._save_certificate_file()
        self._save_cert_params_into_db()

    def _get_cert_model(self):
        with self._uow_cm as (_, repository):
            self._cert = repository.get(self._data.cert_name)

    def _checks(self) -> None:  # TODO:Разобраться с checks
        logger.debug('Checking if the csr is already exist.')
        if self._cert.cert_type != 'csr':
            raise ValidationException(f'The certificate {self._cert.name} has already exist.')

        logger.debug('Checking if the certificate file is valid.')
        try:
            x509.load_pem_x509_certificate(self._data.certificate)
        except ValueError:
            raise ValidationException(message='Payload validation error.', data=['The certificate file is not valid.'])

        logger.debug('Checking if the certificate correspond the csr.')
        cert_public_numbers = x509.load_pem_x509_certificate(self._data.certificate).public_key().public_numbers()
        csr_public_numbers = x509.load_pem_x509_csr(self._cert.public_bytes.encode()).public_key().public_numbers()
        if cert_public_numbers != csr_public_numbers:
            raise ValidationException(f'The certificate is not corresponding the {self._cert.name} csr')
        logger.debug('Checks passed.')

    def _save_certificate_file(self) -> None:
        logger.debug('Saving certificate file')

        cert_filename = f'{self._cert.name}{self._ssl_cert_ext}'
        with open(self._ssl_cert_path / cert_filename, 'wb') as file_cert:
            file_cert.write(self._data.certificate)

        logger.debug('Certificate file saved.')

    def _save_cert_params_into_db(self) -> None:
        logger.debug('Saving certificate obj into db')
        cert_params = serialize_pem_certificate_params_from_bytes(self._data.certificate)  # Dependencies
        self._cert.cert_type = 'certificate'
        self._cert.params = cert_params
        with self._uow_cm as (transaction, repository):
            repository.update(self._cert)
            transaction.commit()
        logger.debug('Certificate obj was saved into db')


class DeleteCertificateService():
    """Удаляет сертификат.

    Удаляет модель сертификата и все связанные с ней файлы.

    Args:
        data (bytes): тело запроса.

    Returns:
        None
    """

    def __init__(
        self,
        data: ICertificateInfo,
        unit_of_work_cm: ITransactionManager[ICertificateInfo] = unit_of_work_manager,
    ) -> None:
        self._data = data
        self._uow_cm = unit_of_work_cm
        # Dependencies
        self._ssl_cert_path = pathlib.Path(settings.SSL_CERT_PATH)
        self._ssl_key_path = pathlib.Path(settings.SSL_KEY_PATH)
        self._ssl_cert_ext = settings.SSL_CERT_EXTENSION
        self._ssl_key_ext = settings.SSL_KEY_EXTENSION
        self._config_path = settings.NGINX_CONFIG_PATH

    def execute(self) -> None:
        """Run Command."""
        self._checks()
        self._delete_files()
        self._delete_cert_params_from_db()

    def _checks(self) -> None:
        logger.debug('Checking if certificate is exist into db')
        with self._uow_cm as (_, repository):
            repository.get(name=self._data.name)
        logger.debug('Checking if the certificate can not used in configs.')
        current_settings: WebAccessSettingsDict = parse_nginx_config(self._config_path)  # Dependencies
        if self._data.name == current_settings.get('cert_name'):
            raise ValidationException(
                message='Payload validation error.',
                data=[f'The certificate {self._data.name} is used.'],
            )

        logger.debug('Checks passed.')

    def _delete_files(self) -> None:
        cert_filename: str = f'{self._data.name}{self._ssl_cert_ext}'
        cert_file: pathlib.Path = pathlib.Path(self._ssl_cert_path) / cert_filename
        cert_file.unlink()

        key_filename: str = f'{self._data.name}{self._ssl_key_ext}'
        key_file: pathlib.Path = pathlib.Path(self._ssl_key_path) / key_filename
        key_file.unlink()

    def _delete_cert_params_from_db(self) -> None:
        with self._uow_cm as (transaction, repository):
            repository.delete(self._data)
            transaction.commit()
