"""Сервисы для настройки web доступа."""
from dataclasses import asdict, dataclass
import logging
import subprocess
from typing import Any, Callable, Optional
import re

from system_configurator import settings
from system_configurator.core.file_system import save_config
from system_configurator.external_access.entities import IWebAccessSettingsInfo, WebAccessSettingsDict

logger = logging.getLogger(__name__)


@dataclass
class WebAccessSettingsInfo:
    """Web Access settings Info.

    http_port (int):             Http порт. По умолчанию 80.
    https_port (int):            Https порт. По умолчанию 443.
    is_https_enabled (bool):     Индикатор включен ли https. По умолчанию False.
    is_http_redirected (bool):   Индикатор включена ли переадресация http to https. По умолчанию False.
    cert_name (str):             Имя сертификата сервера. По умолчанию пустое.
    """

    http_port: int = 80
    https_port: int = 443
    is_https_enabled: bool = False
    is_http_redirected: bool = False
    cert_name: str = ''


def parse_http_port(config: str) -> Optional[int]:
    """Возвращает номер прослушиваемого http порта из конфигурации nginx.

    Args:
        config (str): Конфигурация nginx.

    Returns:
        Optional[int]: номер прослушиваемого http порта.
    """
    regex = r'^    listen\s+(?P<http_port>\d+);$'
    match: Optional[re.Match[Any]] = re.search(regex, config, flags=re.MULTILINE)
    return int(match.group('http_port')) if match else 80


def parse_https_port(config: str) -> Optional[int]:
    """Возвращает номер прослушиваемого https порта из конфигурации nginx.

    Args:
        config (str): Конфигурация nginx.

    Returns:
        Optional[int]: номер прослушиваемого https порта.
    """
    regex = r'^    listen\s+(?P<https_port>\d+) ssl;$'
    match: Optional[re.Match[Any]] = re.search(regex, config, flags=re.MULTILINE)
    return int(match.group('https_port')) if match else 443


def parse_is_https_enabled(config: str) -> bool:
    """Проверяет активирован ли https в конфигурации nginx.

    Args:
        config (str):  Конфигурация nginx.

    Returns:
        bool: Активирован ли https в конфигурации nginx.
    """
    regex = r'^    listen\s+(?P<https_port>\d+) ssl;$'
    match: Optional[re.Match[Any]] = re.search(regex, config, flags=re.MULTILINE)
    return bool(match)


def parse_is_http_redirect(config: str) -> bool:
    """Проверяет активирован ли redirect с http на https в конфигурации nginx.

    Args:
        config (str): Конфигурация nginx.

    Returns:
        bool: Активирован ли redirect с http на https в конфигурации nginx.
    """
    redirect_config = 'return 301 https://$host$request_uri;'
    return redirect_config in config


def parse_server_name(config: str) -> Optional[str]:
    """Возвращает имя сервера из конфигурации nginx.

    Args:
        config (str): Конфигурация nginx.

    Returns:
        Optional[str]: Имя сервера из конфигурации nginx.
    """
    regex = r'    server_name\s+(?P<server_name>[^_]\S+);$'
    match: Optional[re.Match[Any]] = re.search(regex, config, flags=re.MULTILINE)
    return match.group('server_name') if match else None


def parse_cert_name(config: str) -> str:
    """Возвращает имя модели сертификата из конфигурации nginx.

    Args:
        config (str): Конфигурация nginx.

    Returns:
        Optional[str]: Имя модели сертификата из конфигурации nginx.
    """
    regex = r'    ssl_certificate\s+/etc/ssl/certs/(?P<cert_name>\S+).crt;$'
    match: Optional[re.Match[Any]] = re.search(regex, config, flags=re.MULTILINE)
    return match.group('cert_name') if match else ''


def parse_cert_key(config: str) -> Optional[str]:
    """Возвращает имя ключа модели сертификата из конфигурации nginx.

    Args:
        config (str): Конфигурация nginx.

    Returns:
        Optional[str]: Имя ключа модели сертификата из конфигурации nginx.
    """
    regex = r'    ssl_certificate_key\s+(?P<cert_key>\S+);$'
    match: Optional[re.Match[Any]] = re.search(regex, config, flags=re.MULTILINE)
    return match.group('cert_key') if match else None


NGINX_PARSER_FABRIC: dict[str, Callable[[str], Any]] = {
    'http_port': parse_http_port,
    'https_port': parse_https_port,
    'is_https_enabled': parse_is_https_enabled,
    'is_http_redirected': parse_is_http_redirect,
    'cert_name': parse_cert_name,
}


def parse_nginx_config(
    config_path: str = settings.NGINX_CONFIG_PATH,
    parser_fabric: dict[str, Callable[[str], Any]] = NGINX_PARSER_FABRIC,
) -> WebAccessSettingsDict:
    """Извлекает параметры модели настроек web доступа из конфигурации nginx.

    Args:
        config_path (str, optional): Путь до файла конфигурации  Nginx. Defaults to settings.NGINX_CONFIG_PATH.
        parser_fabric (dict[str, Callable[[str], Any]], optional): Фабрика содержащая парсеры отдельных параметров.
        Defaults to NGINX_PARSER_FABRIC.

    Returns:
        WebAccessSettingsDict: Модель настроек web доступа.
    """
    parsed_data: dict[str, Any] = {}
    with open(config_path) as file:
        nginx_config: str = file.read()
    for key, parser in parser_fabric.items():
        value = parser(nginx_config)
        parsed_data[key] = value
    return asdict(WebAccessSettingsInfo(**parsed_data))


class SaveNginxConfigCommand():
    """Save nginx config.

    Configuration saves to the hostname and hosts files.

    Args:
        data (IWebAccessSettingsInfo): конфигурация nginx.

    Returns:
        None
    """

    def __init__(self, data: IWebAccessSettingsInfo) -> None:
        self._data = data

    def execute(self) -> None:
        """Run Command."""
        save_config(
            data=self._data,
            template='nginx.conf',
            config='lightmaster.conf',
            config_path='/home/lightmaster/lightmaster/conf/nginx/',
        )


class ReloadSystemdServiceCommand():
    """Перезапускает systemctl сервис.

    Args:
        service_name (str) - Имя сервиса.
        sudo (bool) - Использовать ли sudo
    """

    def __init__(self, service_name: str, sudo: bool) -> None:
        self._service_name = service_name
        self._sudo = sudo

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to reload service %s', self._service_name)
        command: list[str] = [
            'systemctl',
            'reload',
            self._service_name,
        ]
        if self._sudo:
            command.insert(0, 'sudo')
        subprocess.run(command)

        logger.debug('%s service has been reloaded', self._service_name)


class ChangeNginxConfigService():
    """Изменяет конфигурацию nginx если необходимо.

    Args:
        data (WebAccessSettingsDict): data for this command is not needded.

    Returns:
        None
    """

    def __init__(self, data: IWebAccessSettingsInfo) -> None:
        self._data = data
        # Dependencies
        self._config_path = settings.NGINX_CONFIG_PATH

    def execute(self) -> None:
        """Run Command."""
        if self._check_changes():
            SaveNginxConfigCommand(self._data).execute()  # Dependencies
            ReloadSystemdServiceCommand('nginx', True).execute()  # Dependencies

    def _check_changes(self) -> bool:
        current_settings: WebAccessSettingsDict = parse_nginx_config(self._config_path)
        return any([
            not self._data.cert_name == current_settings['cert_name'],
            not self._data.http_port == current_settings['http_port'],
            not self._data.https_port == current_settings['https_port'],
            not self._data.is_http_redirected == current_settings['is_http_redirected'],
            not self._data.is_https_enabled == current_settings['is_https_enabled'],
        ])
