"""Netplan repository."""
import pathlib
from typing import Any, Callable, Literal, Protocol, Union, overload

import yaml

from system_configurator.network.adapters import utils
from system_configurator.core.exceptions import Error
from system_configurator.network.adapters.netplan.utils import FABRIC_IFACE_PROPS_GETTERS_FROM_NETPLAN
from system_configurator.network.adapters.repository import IEthernetRepository

from system_configurator.network.domain.models import (
    IModemIface,
    IWiredIface,
    ModemIface,
    WiredIface,
)
from system_configurator.network.exceptions import ModemNotInstalled


class IWiredIfaceBuilder(Protocol):
    """Интерфейс строителя для проводного сетевого интерфейса."""

    def build(self) -> IWiredIface:
        """Конструирует объект проводного сетевого интерфейса.

        Returns:
            IWiredIface: Объект проводного сетевого интерфейса.
        """
        ...


class IModemIfaceBuilder(Protocol):
    """Интерфейс строителя для модемного сетевого интерфейса."""

    def build(self) -> IModemIface:
        """Конструирует объект модемного сетевого интерфейса.

        Returns:
            IModemIface: Объект модемного сетевого интерфейса.
        """
        ...


NETPLAN_CONFIGS_PATH = '/etc/netplan'
NETPLAN_CONFIGS_RUNTIME_PATH = '/run/netplan'
NETPLAN_WIRED_IFACE_CONFIG_NAME = '10-ethernets.yaml'
NETPLAN_MODEM_IFACE_CONFIG_NAME = '20-modems.yaml'


class WiredIfaceBuilder:
    """Строитель для проводного сетевого интерфейса."""

    def __init__(
        self,
        netplan_config_path: str = NETPLAN_CONFIGS_PATH,
        iface_config_name: str = NETPLAN_WIRED_IFACE_CONFIG_NAME,
        iface_fabric: dict[str, Callable[[dict[str, Any]], Any]] = FABRIC_IFACE_PROPS_GETTERS_FROM_NETPLAN,
    ) -> None:
        self._config_path = pathlib.Path(netplan_config_path)
        self._iface_config_name = iface_config_name
        self._iface_fabric = iface_fabric

    def build(self) -> WiredIface:
        """Конструирует объект проводного сетевого интерфейса.

        Returns:
            IWiredIface: Объект проводного сетевого интерфейса.
        """
        iface_settings = self._get_netplan_wired_iface_config()
        ip_assign_mode = self._iface_fabric['get_ip_assign_method'](iface_settings)
        ip, netmask = self._iface_fabric['get_static_ip_credentials'](iface_settings)
        gateway = self._iface_fabric['get_gateway'](iface_settings)
        dns_assign_mode = self._iface_fabric['get_dns_assign_mode'](iface_settings)
        dns_servers = self._iface_fabric['get_dns_servers'](iface_settings)

        wired_iface = WiredIface()
        wired_iface.set_ip_assignment_method(ip_assign_mode)
        wired_iface.set_static_ip_credential(ip, netmask, gateway)
        wired_iface.set_dns_assignment_method(dns_assign_mode)
        wired_iface.set_static_dns_servers(dns_servers)

        return wired_iface

    def _get_netplan_wired_iface_config(self) -> dict[str, Any]:
        with open(self._config_path / self._iface_config_name) as file:
            iface = yaml.safe_load(file)
        iface_settings: dict[str, Any] = iface['network']['ethernets']['eth0']
        return iface_settings


class ModemIfaceBuilder:
    """Интерфейс строителя для модемного сетевого интерфейса."""

    def __init__(
        self,
        netplan_config_path: str = NETPLAN_CONFIGS_PATH,
        iface_config_name: str = NETPLAN_MODEM_IFACE_CONFIG_NAME,
        iface_fabric: dict[str, Callable[[dict[str, Any]], Any]] = FABRIC_IFACE_PROPS_GETTERS_FROM_NETPLAN,
    ) -> None:
        self._config_path = pathlib.Path(netplan_config_path)
        self._iface_config_name = iface_config_name
        self._iface_fabric = iface_fabric

    def build(self) -> ModemIface:
        """Конструирует объект модемного сетевого интерфейса.

        Returns:
            IModemIface: Объект модемного сетевого интерфейса.
        """
        if not utils.is_modem_installed():
            raise ModemNotInstalled()

        iface_settings = self._get_netplan_modem_iface_config()
        ip_assign_mode = self._iface_fabric['get_ip_assign_method'](iface_settings)
        ip, netmask = self._iface_fabric['get_static_ip_credentials'](iface_settings)
        gateway = self._iface_fabric['get_gateway'](iface_settings)
        dns_assign_mode = self._iface_fabric['get_dns_assign_mode'](iface_settings)
        dns_servers = self._iface_fabric['get_dns_servers'](iface_settings)
        apn_settings = self._iface_fabric['get_apn_settings'](iface_settings)

        modem_iface = ModemIface()
        modem_iface.set_ip_assignment_method(ip_assign_mode)
        modem_iface.set_static_ip_credential(ip, netmask, gateway)
        modem_iface.set_dns_assignment_method(dns_assign_mode)
        modem_iface.set_static_dns_servers(dns_servers)
        modem_iface.set_apn_settings(**apn_settings)

        return modem_iface

    def _get_netplan_modem_iface_config(self) -> dict[str, Any]:
        with open(self._config_path / self._iface_config_name) as file:
            iface = yaml.safe_load(file)
        iface_settings: dict[str, Any] = iface['network']['modems']['cdc-wdm0']
        return iface_settings


class NetplanRepository(IEthernetRepository):
    """Репозиторий с ethernet интерфейсами."""

    def __init__(
        self,
        wired_iface_builder: IWiredIfaceBuilder = WiredIfaceBuilder(),
        modem_iface_builder: IModemIfaceBuilder = ModemIfaceBuilder(),
    ) -> None:
        self._wired_iface_builder = wired_iface_builder
        self._modem_iface_builder = modem_iface_builder

        self.dirty: set[Union[IWiredIface, IModemIface]] = set()
        self._loaded_ifaces: dict[str, Union[IWiredIface, IModemIface]] = {}

    @overload
    def get(self, iface_type: Literal['wired']) -> IWiredIface: ...
    @overload
    def get(self, iface_type: Literal['modem']) -> IModemIface: ...

    def get(self, iface_type: Literal['wired', 'modem']) -> Union[IWiredIface, IModemIface]:
        """Получить объект доменной модели интерфейса.

        Принимает тип интерфейса и исходя из него возвращает нужный объект.

        Args:
            iface_type (Literal['wired', 'modem']): Тип интерфейса.

        Returns:
            Union[IWiredIface, IModemIface]: Объект доменной модели интерфейса.
        """
        iface_loaded = self._loaded_ifaces.get(iface_type)
        if not iface_loaded:
            iface_loaded = self._load_iface(iface_type)
            self._loaded_ifaces[iface_type] = iface_loaded
        return iface_loaded

    def _load_iface(self, iface: Literal['wired', 'modem']) -> Union[IWiredIface, IModemIface]:
        if iface == 'wired':
            return self._wired_iface_builder.build()
        elif iface == 'modem':
            return self._modem_iface_builder.build()
        else:
            raise Error(f'Unsupported interface type {iface}')

    def update(self, iface: Union[IWiredIface, IModemIface]) -> None:
        """Обновить интерфейс.

        Args:
            iface (Union[IWiredIface, IModemIface]): Объект доменной модели интерфейса.
        """
        if isinstance(iface, IWiredIface):
            self._loaded_ifaces['wired'] = iface
        elif isinstance(iface, IModemIface):
            self._loaded_ifaces['modem'] = iface
        else:
            raise Error('Unsupported interface type!')
        self.dirty.add(iface)


if __name__ == '__main__':
    repo: IEthernetRepository = NetplanRepository()
