"""Uow netplan."""
import logging
import pathlib
from queue import Queue
import shutil
import subprocess
from typing import Any, Literal, Union
import ipaddress

import yaml

from system_configurator.core.db.transaction import ITransaction, ITransactionManager
from system_configurator.core.exceptions import Error, ShellCommandError
from system_configurator.core.msgbus import MSGBUS, Event, register_event
from system_configurator.core.shell import run_shell_command
from system_configurator.network.adapters.repository import IEthernetRepository
from system_configurator.network.adapters.netplan.repository import (
    NETPLAN_CONFIGS_PATH,
    NETPLAN_CONFIGS_RUNTIME_PATH,
    NETPLAN_MODEM_IFACE_CONFIG_NAME,
    NETPLAN_WIRED_IFACE_CONFIG_NAME,
    NetplanRepository,
)
from system_configurator.network.domain.models import IModemIface, IWiredIface


logger = logging.getLogger(__name__)


TIPLikeString = str


class WiredIfaceConfigWasApplied(Event):
    """Событие применения конфигурации проводного сетевого интерфейса."""
    ...


class ModemIfaceConfigWasApplied(Event):
    """Событие применения конфигурации модемного сетевого интерфейса."""
    ...


register_event('wired_iface_config_was_applied', WiredIfaceConfigWasApplied)
register_event('modem_iface_config_was_applied', ModemIfaceConfigWasApplied)


def make_netplan_dhcp_part(enable_dhcp: bool) -> dict[str, Any]:
    """Формирует dhcp часть netplan конфигурации.

    Args:
        enable_dhcp (bool): Включить или выключить dhcp на интерфейсе.

    Returns:
        dict[str, Any]: dhcp часть netplan конфигурации.
    """
    return {
        'dhcp4': enable_dhcp,
    }


def make_dhcp_overrides_part(enable_dns_assignment_by_dhcp: bool) -> dict[str, Any]:
    """Формирует часть настроек dhcp netplan конфигурации.

    Args:
        enable_dns_assignment_by_dhcp (bool): Включить или выключить назначение dns серверов по dhcp.

    Returns:
        dict[str, Any]: Часть настроек dhcp netplan конфигурации.
    """
    return {
        'dhcp4-overrides': {
            'use-dns': enable_dns_assignment_by_dhcp,
            'use-ntp': False,
        },
    }


def make_netplan_dns_static_part(dns_servers: list[str]) -> dict[str, Any]:
    """Формирует dns часть netplan конфигурации.

    Args:
        dns_servers (list[str]): Список dns серверов.

    Returns:
        dict[str, Any]: dns часть настроек netplan конфигурации.
    """
    return {
        'nameservers': {
            'addresses': dns_servers,
        },
    }


def make_netplan_ip_addresses_part(ip: str, netmask: str) -> dict[str, Any]:
    """Формирует часть статического ip адреса netplan конфигурации.

    Args:
        ip (str): ip адрес интерфейса.
        netmask (str): Сетевая маска интерфейса.

    Returns:
        dict[str, Any]: Часть статического ip адреса netplan конфигурации.
    """
    return {
        'addresses': [str(ipaddress.IPv4Interface(f'{ip}/{netmask}'))],
    }


def make_netplan_gateway_part(gateway: str) -> dict[str, Any]:
    """Формирует часть статического маршрута по умолчанию netplan конфигурации.

    Args:
        gateway (str): ip адрес шлюза для маршрута по умолчанию.

    Returns:
        dict[str, Any]: Часть статического маршрута по умолчанию netplan конфигурации.
    """
    return {
        'routes': [
            {'to': '0.0.0.0/0', 'via': gateway},
        ],
    }


def make_netplan_apn_part(apn: str, username: str, password: str) -> dict[str, Any]:
    """Формирует часть apn настроек netplan конфигурации.

    Args:
        apn (str): Apn сервер.
        username (str): Имя пользователя.
        password (str): Пароль.

    Returns:
        dict[str, Any]: Часть apn настроек netplan конфигурации.
    """
    return {
        'apn': apn,
        'username': username,
        'password': password,
    }


def make_netplan_wired_config(config_part: dict[str, Any]) -> dict[str, Any]:
    """Формирует netplan конфигурацию проводного интерфейса eth0.

    Args:
        config_part (dict[str, Any]): Конфигурация отдельных параметров интерфейса.

    Returns:
        dict[str, Any]: netplan конфигурация проводного интерфейса eth0.
    """
    return {
        'network': {
            'version': 2,
            'ethernets': {
                'eth0': config_part,
            },
        },
    }


def make_netplan_modem_config(config_part: dict[str, Any]) -> dict[str, Any]:
    """Формирует netplan конфигурацию модемного интерфейса cdc-wdm0.

    Args:
        config_part (dict[str, Any]): Конфигурация отдельных параметров интерфейса.

    Returns:
        dict[str, Any]: netplan конфигурация  модемного интерфейса cdc-wdm0.
    """
    return {
        'network': {
            'version': 2,
            'modems': {
                'renderer': 'NetworkManager',
                'cdc-wdm0': config_part,
            },
        },
    }


def generate_netplan_config(iface: Union[IWiredIface, IModemIface]) -> dict[str, Any]:
    """Конструирует netplan конфигурацию.

    Args:
        iface (Union[IWiredIface, IModemIface]): Объект доменной модели сетевого интерфейса.

    Returns:
        dict[str, Any]: Netplan конфигурация.
    """
    config_extends = {}
    if iface.ip_assign_method == 'dhcp':
        config_extends.update(make_netplan_dhcp_part(True))
        config_extends.update(make_dhcp_overrides_part(iface.dns_assign_method == 'dhcp'))
    else:
        config_extends.update(make_netplan_ip_addresses_part(iface.static_ip, iface.static_netmask))
        config_extends.update(make_netplan_gateway_part(iface.static_gateway))
    config_extends.update(make_netplan_dns_static_part(iface.static_dns_servers))
    if isinstance(iface, IModemIface):
        config_extends.update(make_netplan_apn_part(iface.apn, iface.username, iface.password))
        config = make_netplan_modem_config(config_extends)
    else:
        config = make_netplan_wired_config(config_extends)
    return config


def get_iface_type(iface: Union[IWiredIface, IModemIface]) -> Literal['wired', 'modem']:
    """Десериализует объект интерфейса в имя его типа.

    Args:
        iface (Union[IWiredIface, IModemIface]): Объект доменной модели сетевого интерфейса.

    Returns:
        Literal['wired', 'modem']: Название типа сетевого интерфейса.
    """
    if isinstance(iface, IWiredIface):
        return 'wired'
    elif isinstance(iface, IModemIface):
        return 'modem'


class NetplanUow(ITransaction):
    """Netplan Unit of work."""

    def __init__(
        self,
        repo: IEthernetRepository,
        netplan_config_runtime_path: str = NETPLAN_CONFIGS_RUNTIME_PATH,
        netplan_config_permanent_path: str = NETPLAN_CONFIGS_PATH,
        wired_config_filename: str = NETPLAN_WIRED_IFACE_CONFIG_NAME,
        modem_config_filename: str = NETPLAN_MODEM_IFACE_CONFIG_NAME,
        event_msg_bus: Queue[Event] = MSGBUS,
    ) -> None:
        self._netplan_config_runtime_path = pathlib.Path(netplan_config_runtime_path)
        self._netplan_config_permanent_path = pathlib.Path(netplan_config_permanent_path)
        self._wired_config_filename = wired_config_filename
        self._modem_config_filename = modem_config_filename
        self._repo = repo
        self._event_msg_bus = event_msg_bus

    def commit(self) -> None:
        """Применяет изменения."""
        self._save_changes_to_runtime_dir()
        self._apply_runtime_config()
        self._move_config_to_startup_dir()
        logger.info('Committed network changes.')

    def rollback(self) -> None:
        """Откатывает изменения если они еще не приняты."""
        if self._netplan_config_runtime_path.exists():
            for file in self._netplan_config_runtime_path.iterdir():
                file.unlink(missing_ok=True)
        logger.info('Rollback network changes.')

    def _save_changes_to_runtime_dir(self) -> None:
        self._netplan_config_runtime_path.mkdir(exist_ok=True)
        while self._repo.dirty:
            iface = self._repo.dirty.pop()
            self._save_iface_config_to_run(iface)

    def _save_iface_config_to_run(self, iface: Union[IWiredIface, IModemIface]) -> None:
        iface_config = generate_netplan_config(iface)  # DEPENDENCY
        if isinstance(iface, IModemIface):
            config_path = self._netplan_config_runtime_path / self._modem_config_filename
        else:
            config_path = self._netplan_config_runtime_path / self._wired_config_filename
        with open(config_path, 'w') as file:
            yaml.dump(iface_config, file, sort_keys=False)

    def _apply_runtime_config(self) -> None:
        command: list[str] = [
            'netplan',
            'apply',
        ]
        try:
            run_shell_command(command, sudo=True, check=True)  # DEPENDENCY
        except ShellCommandError as err:
            raise Error('An error occurred while applying the network configuration', err.data)

    def _register_apply_events(self) -> None:
        self._event_msg_bus.put(ModemIfaceConfigWasApplied())
        self._event_msg_bus.put(WiredIfaceConfigWasApplied())

    def _move_config_to_startup_dir(self) -> None:
        for file in self._netplan_config_runtime_path.iterdir():
            shutil.move(file, self._netplan_config_permanent_path / file.name)
        self._register_apply_events()


class NetplanUowManager(ITransactionManager):
    """Интерфейс для контекстного менеджера транзакции."""

    def __init__(self) -> None:
        self.repo = NetplanRepository()
        self.uow = NetplanUow(self.repo)

    def __enter__(self) -> tuple[ITransaction, IEthernetRepository]:
        """Точка входа в контекстный менеджер. Возвращает транзакцию и репозиторий."""
        return (self.uow, self.repo)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Выход из контекстного менеджера."""
        if exc_type is not None:
            self.uow.rollback()
