import ipaddress
from typing import Any, Callable, Literal, TypedDict

import yaml


class ApnSettingsDict(TypedDict):
    apn: str
    username: str
    password: str


def get_netplan_iface_config(netplan_config_path: str, iface_config_name: str,) -> dict[str, Any]:
    with open(f'{netplan_config_path}/{iface_config_name}') as file:
        iface = yaml.safe_load(file)
    try:
        iface_settings: dict[str, Any] = iface['network']['ethernets']['eth0']
    except KeyError:
        iface_settings = iface['network']['modems']['cdc-wdm0']
    return iface_settings


def get_ip_assign_method(iface_settings: dict[str, Any]) -> Literal['dhcp', 'manual']:
    return 'dhcp' if iface_settings.get('dhcp4') else 'manual'


def get_static_ip_credentials(iface_settings: dict[str, Any]) -> list[str]:
    try:
        ip_with_netmask = ipaddress.IPv4Interface(iface_settings['addresses'][0]).with_netmask
    except (IndexError, KeyError):
        ip_with_netmask = '/'
    return ip_with_netmask.split('/')


def get_gateway(iface_settings: dict[str, Any]) -> str:
    gateway: str = ''
    if routes := iface_settings.get('routes'):
        for route in routes:
            if route['to'] == '0.0.0.0/0' or route['to'] == 'default':
                gateway = route['via']
                break
    return gateway


def get_dns_assign_mode(iface_settings: dict[str, Any]) -> Literal['dhcp', 'manual']:
    if not iface_settings.get('dhcp4'):
        return 'manual'
    try:
        use_dns_prop = iface_settings['dhcp4-overrides']['use-dns']
    except KeyError:
        use_dns_prop = True
    return 'dhcp' if use_dns_prop else 'manual'


def get_dns_servers(iface_settings: dict[str, Any]) -> list[str]:
    try:
        dns_servers: list[str] = iface_settings['nameservers']['addresses']
    except KeyError:
        dns_servers = []
    return dns_servers


def get_apn_settings(iface_settings: dict[str, Any]) -> ApnSettingsDict:
    return {
        'apn': iface_settings.get('apn') or '',
        'username': iface_settings.get('username') or '',
        'password': iface_settings.get('password') or '',
    }


FABRIC_IFACE_PROPS_GETTERS_FROM_NETPLAN: dict[str, Callable[[dict[str, Any]], Any]] = {
    'get_ip_assign_method': get_ip_assign_method,
    'get_static_ip_credentials': get_static_ip_credentials,
    'get_gateway': get_gateway,
    'get_dns_assign_mode': get_dns_assign_mode,
    'get_dns_servers': get_dns_servers,
    'get_apn_settings': get_apn_settings,
}
