"""Network configuration utilities."""
from abc import ABC, abstractmethod
import array
import fcntl
import ipaddress
import json
import pathlib
import socket
import struct
import sys
from typing import Any, Generic, Iterable, Sequence, Type, TypeVar, TypedDict, Union, cast

from system_configurator import settings
from system_configurator.core.exceptions import Error
from system_configurator.core.file_system import save_config
from system_configurator.core.shell import run_shell_command
from system_configurator.settings.settings import MODEM_DEVICE_PATH


# Interface flags
IFF_BROADCAST = 0x0002
IFF_UP = 0x0001
IFF_LOOPBACK = 0x0008
IFF_POINTOPOINT = 0x0010
IFF_NOARP = 0x0040
IFF_AUTOUP = 0x0080
IFF_PROMISC = 0x0100
IFF_ALLMULTI = 0x0200
IFF_SIMPLEX = 0x0800
IFF_LINK = 0x1000
IFF_AUTO_CONFIGURED = 0x2000
IFF_CONFIGURING = 0x4000
IFF_MULTICAST = 0x8000
IFF_VIRTUAL = 0x2000000000
IFF_DYNAMIC = 0x8000

# ioctl commands
SIOCGIFCONF = 0x8912
SIOCGIFFLAGS = 0x8913
SIOCGIFNETMASK = 0x891b
SIOCSIFNETMASK = 0x891C
SIOCGIFADDR = 0x8915
SIOCSIFADDR = 0x8916
SIOCGIFHWADDR = 0x8927
SIOCADDRT = 0x890B
SIOCDELRT = 0x890C
RTF_UP = 0x0001
RTF_GATEWAY = 0x0002


T = TypeVar('T')


class Command(ABC, Generic[T]):
    """Abstract class for command pattern."""

    @abstractmethod
    def execute(self) -> T:
        """Contain all logic.

        The only one public method.
        """
        pass


class NetworkInterfaceNameInfoDict(TypedDict):
    """Network interface name info.

    ifname: str - Network interface name
    """

    ifname: str


class ChangeIPInfoDict(TypedDict):
    """Info for change ip address to the property interface.

    ifname: str - Network interface name
    ip: str     - ip address
    """

    ifname: str
    ip: str


class ChangeNetmaskInfoDict(TypedDict):
    """Info for change network mask to the property interface.

    ifname: str - Network interface name
    prefix: str - Network mask.
    """

    ifname: str
    prefix: str


class NetworkRouteInfoDict(TypedDict):
    """Network route info.

    dst_address: str    - destination network address.
    dst_netmask: str    - destination network mask
    gateway: str        - gateway.
    """

    dst_address: str
    dst_netmask: str
    gateway: str


class HostnameInfoDict(TypedDict):
    """Network route info.

    hostname: str    - system hostname.

    """
    hostname: str


class DnsServersInfoDict(TypedDict):
    """Dns server addresses info.

    dns_servers: list[str] - list of dns server ip address.
    """

    dns_servers: list[str]


class GetIntFieldValueCommandInterface(Command[T]):
    """Abstract class for generating report field value classes."""

    @abstractmethod
    def __init__(self, ifname: str) -> None:
        """Initiate variables.

        Args:
            ifname (str): Network interface name.
        """

    @abstractmethod
    def execute(self) -> T:
        """Run command.

        Returns:
            str: report_value
        """


class GetInterfaceParamViaIoctl(Command[bytes]):
    """Get interface param via ioctl."""

    def __init__(self, ifname: str, command_flag: int) -> None:
        """Init variables."""
        self.ifname = ifname
        self.command_flag = command_flag

        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.arg = struct.pack('256s', bytes(self.ifname, 'utf-8'))
        super().__init__()

    def execute(self) -> bytes:
        """Run command.

        Returns:
            bytes: raw ioctl output
        """
        try:
            output_hex: bytes = fcntl.ioctl(self.s.fileno(), self.command_flag, self.arg)
        except OSError:
            output_hex = b''
        return output_hex


class GetInterfaceIpAddressViaIoctlCommand(GetIntFieldValueCommandInterface[str]):
    """Get interface ip address."""

    command_flag = SIOCGIFADDR
    ioctl_command = GetInterfaceParamViaIoctl

    def __init__(self, ifname: str) -> None:
        """Init variables."""
        self.ifname = ifname

    def execute(self) -> str:
        """Run command."""
        ip_raw: bytes = self.ioctl_command(self.ifname, self.command_flag).execute()[20:24]
        return socket.inet_ntoa(ip_raw) if ip_raw else ''


class GetInterfacePrefixViaIoctlCommand(GetIntFieldValueCommandInterface[str]):
    """Get interface network prefix."""

    command_flag = SIOCGIFNETMASK
    ioctl_command = GetInterfaceParamViaIoctl

    def __init__(self, ifname: str) -> None:
        """Init variables."""
        self.ifname = ifname

    def execute(self) -> str:
        """Run command."""
        mask_hex = self.ioctl_command(self.ifname, self.command_flag).execute()[20:24]
        return socket.inet_ntoa(mask_hex) if mask_hex else ''


class GetInterfacePrefixLenViaIoctlCommand(GetIntFieldValueCommandInterface[str]):
    """Get interface network prefix length."""

    command_flag = SIOCGIFNETMASK
    ioctl_command = GetInterfaceParamViaIoctl

    def __init__(self, ifname: str) -> None:
        """Init variables."""
        self.ifname = ifname

    def execute(self) -> str:
        """Run command."""
        mask_hex: bytes = self.ioctl_command(self.ifname, self.command_flag).execute()[20:24]
        prefix_len: int = sum(bin(octet).count('1') for octet in mask_hex) if mask_hex else 0
        return str(prefix_len)


class SetIPAddressViaIoctl():
    """Set ip address.

    Args:
        data (ChangeIPInfoDict): dict with new ip address and interface name.

    Returns:
        None
    """

    def __init__(self, data: ChangeIPInfoDict) -> None:
        self.ifname: str = data['ifname']
        self.new_ip: str = data['ip']
        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def execute(self) -> None:
        """Run Command."""
        bin_ip = socket.inet_aton(self.new_ip)
        iface_name: bytes = bytes(self.ifname.encode('utf-8'))
        arg = struct.pack('16sH2s4s8s', iface_name, socket.AF_INET, b'\x00' * 2, bin_ip, b'\x00' * 8)
        fcntl.ioctl(self.s, SIOCSIFADDR, arg)


class SetNetmaskViaIoctl():
    """Set network interface mask.

    Args:
        data (ChangeNetmaskInfoDict): dict with new network interface mask and interface name.

    Returns:
        None
    """

    def __init__(self, data: ChangeNetmaskInfoDict) -> None:
        self.ifname: str = data['ifname']
        self.new_netmask: str = data['prefix']
        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def execute(self) -> None:
        """Run Command."""
        bin_prefix = socket.inet_aton(self.new_netmask)
        iface_name: bytes = bytes(self.ifname.encode('utf-8'))
        arg = struct.pack('16sH2s4s8s', iface_name, socket.AF_INET, b'\x00' * 2, bin_prefix, b'\x00' * 8)
        fcntl.ioctl(self.s, SIOCSIFNETMASK, arg)


def struct_rtentry(dest: str, mask: str, gw: str) -> bytes:
    """Struct network route to c array for kernel command.

    Args:
        dest (str): destination network address.
        mask (str): destination network mask
        gw (str):   gateway

    Returns:
        bytes: c array for kernel command
    """
    inet_aton = socket.inet_aton
    sockaddr_in_fmt = 'hH4s8s'
    rtentry_fmt = 'L16s16s16sH38s'
    dst = struct.pack(sockaddr_in_fmt, socket.AF_INET, 0, inet_aton(dest), b'\x00' * 8)
    next_gw = struct.pack(sockaddr_in_fmt, socket.AF_INET, 0, inet_aton(gw), b'\x00' * 8)
    netmask = struct.pack(sockaddr_in_fmt, socket.AF_INET, 0, inet_aton(mask), b'\x00' * 8)
    rt_flags = RTF_UP | RTF_GATEWAY
    rtentry = struct.pack(rtentry_fmt, 0, dst, next_gw, netmask, rt_flags, b'\x00' * 38)
    return rtentry


class AddNetworkRouteViaIoctlCommand():
    """Add runtime network route.

    Args:
        data (NetworkRouteInfoDict): dict with information about route.

    Returns:
        None
    """

    def __init__(self, data: NetworkRouteInfoDict) -> None:
        self.dst_address = data['dst_address']
        self.dst_netmask = data['dst_netmask']
        self.gateway = data['gateway']

    def execute(self) -> None:
        """Run command."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        rtentry: bytes = struct_rtentry(self.dst_address, self.dst_netmask, self.gateway)
        fcntl.ioctl(sock.fileno(), SIOCADDRT, rtentry)


class DelNetworkRouteViaIoctlCommand():
    """Delete runtime network route.

    Args:
        data (NetworkRouteInfoDict): dict with information about route.

    Returns:
        None
    """

    def __init__(self, data: NetworkRouteInfoDict) -> None:
        self.dst_address = data['dst_address']
        self.dst_netmask = data['dst_netmask']
        self.gateway = data['gateway']

    def execute(self) -> None:
        """Run Command."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        rtentry: bytes = struct_rtentry(self.dst_address, self.dst_netmask, self.gateway)
        fcntl.ioctl(sock.fileno(), SIOCDELRT, rtentry)


class SetHostnameCommand():
    """Set runtime hostname.

    Args:
        data (HostnameInfoDict): dict with information about hostname.

    Returns:
        None
    """

    def __init__(self, data: HostnameInfoDict) -> None:
        self.new_hostname = data['hostname']

    def execute(self) -> None:
        """Run Command."""
        socket.sethostname(self.new_hostname)


class SetDnsServersCommand():
    """Set dns servers.

    Changes permanently save in the system.

    Args:
        data (DnsServersInfoDict): dict with information about dns server addresses.

    Returns:
        None
    """

    def __init__(self, data: DnsServersInfoDict) -> None:
        self.dns_servers = data['dns_servers']

    def execute(self) -> None:
        """Run Command."""
        data = {'dns_servers': self.dns_servers}
        save_config(data=data, template='resolv.conf', config='resolv.conf', config_path='/etc')


class SaveNetworkInterfaceConfig():
    """Save network interface configuration for systemd-networkd service.

    Configuration saves in format <prefix>-<interface_name>.network to systemd-networkd configuration catalog.

    Args:
        data (NetworkInterfaceNameInfoDict): dict with information about network interface name.

    Returns:
        None
    """

    config_prefix = settings.INTERFACE_CONFIG_PREFIX
    get_ip_address_command = GetInterfaceIpAddressViaIoctlCommand
    get_prefix_len_command = GetInterfacePrefixLenViaIoctlCommand

    def __init__(self, data: NetworkInterfaceNameInfoDict) -> None:
        self.ifname = data['ifname']

    def execute(self) -> None:
        """Run Command."""
        self._prepare_interface_config_data()
        config_name: str = f"{self.config_prefix}-{self.ifname}.network"
        save_config(
            data=self.data,
            template='interface.template',
            config=config_name,
            config_path='/etc/systemd/network',
        )

    def _prepare_interface_config_data(self) -> None:
        """Prepare data for saving in interface configuration."""
        ip_address: str = self.get_ip_address_command(self.ifname).execute()
        prefix_len: str = self.get_prefix_len_command(self.ifname).execute()
        self.data = {
            'interface': {
                'name': self.ifname,
                'ip': ip_address,
                'prefix_len': prefix_len,
            },
        }


class SaveNetworkRoutesConfigCommand():
    """Save network routes config.

    Configuration saves in file routes.sh to network-routes service configuration catalog.

    Args:
        data (None): data for this command is not needed.

    Returns:
        None
    """

    def __init__(self, data: None) -> None:
        pass

    def execute(self) -> None:
        """Run Command."""
        routes = get_routes()
        data = {'routes': routes}
        save_config(data=data, template='routes.sh', config='routes.sh', config_path='/usr/local/bin')


class SaveHostnameConfigCommand():
    """Save system hostname config.

    Configuration saves to the hostname and hosts files.

    Args:
        data (None): data for this command is not needed.

    Returns:
        None
    """

    def __init__(self, data: None) -> None:
        pass

    def execute(self) -> None:
        """Run Command."""
        hostname: str = socket.gethostname()
        data: dict[str, str] = {'hostname': hostname}
        save_config(data=data, template='hostname', config='hostname', config_path='/etc')
        save_config(data=data, template='hosts', config='hosts', config_path='/etc')


def get_routes() -> list[dict[str, Union[str, int]]]:
    """Получает список всех сетевых маршрутов.

    Returns:
        list[dict[str, Union[str, int]]]: Список всех сетевых маршрутов.
    """
    routes: list[dict[str, Union[str, int]]] = []
    command = ['ip', '--json', '-4', 'route', 'show', 'scope', 'global']
    resp = run_shell_command(command, capture_output=True)
    if resp is not None:
        routes = json.loads(resp)
    return routes


def get_gateway(iface_name: str) -> str:
    """Получает адрес шлюза для маршрута по умолчанию определенного интерфейса.

    Args:
        iface_name (str): имя интерфейса.

    Returns:
        str: IP адрес шлюза.
    """
    command = ['ip', '--json', '-4', 'route', 'show', 'dev', iface_name, 'exac', '0/0', 'scope', 'global']
    resp = run_shell_command(command, capture_output=True, check=True)
    if resp is not None:
        routes = json.loads(resp)
    gateway: str = routes[0]['gateway'] if routes else ''
    return gateway


def get_all_active_interfaces_names() -> set[str]:
    """Generate set of all network interfaces names."""
    is_64bits = sys.maxsize > 2**32
    struct_size = 40 if is_64bits else 32
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    max_possible = 8  # initial value
    while True:
        _bytes = max_possible * struct_size
        names = array.array('B', [0] * _bytes)
        arg = struct.pack('iL', _bytes, names.buffer_info()[0])
        ifconf_hex = fcntl.ioctl(s.fileno(), SIOCGIFCONF, arg)
        out_bytes = struct.unpack('iL', ifconf_hex)[0]
        if out_bytes == _bytes:
            max_possible *= 2
        else:
            break
    name_str = names.tobytes()
    ifaces_names: set[str] = set()
    for i in range(0, out_bytes, struct_size):
        iface_name = bytes.decode(name_str[i:i+16]).split('\0', 1)[0]
        ifaces_names.add(iface_name)
    return ifaces_names


def get_mmcli_modem_statistic() -> dict[str, Any]:
    """Получает статистику модема.

    Returns:
        dict[str, Any]: Словарь со статистикой модема.
    """
    command = ['mmcli', '--output-json', '-m', '0']
    resp = run_shell_command(command, capture_output=True, check=True)
    if resp:
        return cast(dict[str, Any], json.loads(resp))
    raise Error('Command not returned output')


def get_mmcli_bearer_statistic() -> dict[str, Any]:
    """Получает статистику bearer.

    Returns:
        dict[str, Any]: Словарь со статистикой модема.
    """
    command = ['mmcli', '--output-json', '-b', '0']
    resp = run_shell_command(command, capture_output=True, check=False)

    # Респ равен b"" если shell команда завершилась с ошибкой т.к. мы отключили проверку.
    # Сделано это для того что бы не засорять логи сообщениями о недоступном bearer
    # каждые 10 секунд если модем не настроен.
    if not resp:
        raise Error('Command not returned output')

    return cast(dict[str, Any], json.loads(resp))


class GetGatewayCommand(GetIntFieldValueCommandInterface[str]):
    """Command for getting interface gateway.

    Args:
        ifname (str): Имя интерфейса.
    """

    def __init__(self, ifname: str) -> None:
        """Init variables."""
        self.ifname = ifname

    def execute(self) -> str:
        """Get interface gateway.

        Returns:
            str: Gateway.
        """
        return get_gateway(self.ifname)


class GetInterfaceDnsServersCommand(GetIntFieldValueCommandInterface[list[str]]):
    """Command for getting interface dns servers.

    Args:
        ifname (str): Имя интерфейса.
    """

    def __init__(self, ifname: str) -> None:
        self.ifname = ifname

    def execute(self) -> list[str]:
        """Get interface dns servers.

        Returns:
            list[str]: Список dns серверов.
        """
        command = ['resolvectl', 'dns', self.ifname]
        resp = run_shell_command(command, capture_output=True, check=True)
        if resp:
            parsed_string = resp.split(': ')
            if len(parsed_string) == 2:
                servers_string = parsed_string[1]
            else:
                return []
            servers_string = resp.split(': ')[1] if len(resp.split(': ')) == 2 else ''
            servers = (ipaddress.ip_address(srv) for srv in servers_string.strip().split(' '))
        return [srv.exploded for srv in servers if isinstance(srv, ipaddress.IPv4Address)]


class CheckInterfaceFlagCommand(Command[bool]):
    """Command for getting interface flag status.

    Args:
        ifname (str): Имя интерфейса.
        checked_flag (int): Флаг.
    """

    command_flag = SIOCGIFFLAGS
    ioctl_command = GetInterfaceParamViaIoctl

    def __init__(self, ifname: str, checked_flag: int) -> None:
        """Init variables."""
        self.ifname = ifname
        self.checked_flag = checked_flag

    def execute(self) -> bool:
        """Get interface flag status.

        Returns:
            bool: interface flag status.
        """
        flags_hex = self.ioctl_command(self.ifname, self.command_flag).execute()[16:18]
        flags_int, = struct.unpack('H', flags_hex)
        is_flag_set = bool(flags_int & self.checked_flag)
        return is_flag_set


class GetInterfaceStatusViaIoctlCommand(GetIntFieldValueCommandInterface[str]):
    """Command for getting interface status.

    Args:
        ifname (str): Имя интерфейса.
    """

    flag = IFF_UP
    flag_checker_command = CheckInterfaceFlagCommand

    def __init__(self, ifname: str) -> None:
        self.ifname = ifname

    def execute(self) -> str:
        """Get interface status.

        Returns:
            str: interface status.
        """
        is_interface_up: bool = self.flag_checker_command(self.ifname, self.flag).execute()
        return 'up' if is_interface_up else 'down'


class GetInterfaceMacAddressViaIoctlCommand(GetIntFieldValueCommandInterface[str]):
    """Command for getting interface mac address.

    Args:
        ifname (str): Имя интерфейса.
    """

    command_flag = SIOCGIFHWADDR
    ioctl_command = GetInterfaceParamViaIoctl

    def __init__(self, ifname: str) -> None:
        self.ifname = ifname

    def execute(self) -> str:
        """Get mac-address in xx.xx.xx.xx.xx.xx format.

        Returns:
            str: Mac-address in xx.xx.xx.xx.xx.xx format.
        """
        maс_hex = self.ioctl_command(self.ifname, self.command_flag).execute()[18:24]
        mac = '.'.join(['%02x' % char for char in bytearray(maс_hex)])
        return mac


class GetInterfaceNameCommand(GetIntFieldValueCommandInterface[str]):
    """Команда заглушка.

    Возвращает то же имя интерфейса которое получает.

    Args:
        ifname (str): Имя интерфейса.
    """

    def __init__(self, ifname: str) -> None:
        """Init variables."""
        self.ifname = ifname

    def execute(self) -> str:
        """Get connection mode.

        Returns:
            str: Имя интерфейса.
        """
        return self.ifname


class GetModemConnectionModeCommand(GetIntFieldValueCommandInterface[str]):
    """Command for getting modem connection mode.

    Args:
        ifname (str): Имя модемного интерфейса.
    """

    def __init__(self, ifname: str) -> None:
        self.ifname = ifname

    def execute(self) -> str:
        """Get connection mode.

        Returns:
            str: modem connection mode.
        """
        modem_statistic = get_mmcli_modem_statistic()
        mode: list[str] = modem_statistic["modem"]['generic']['access-technologies']
        return mode[0] if mode else ''


class GetModemConnectionStatusCommand(GetIntFieldValueCommandInterface[str]):
    """Command for getting modem connection status.

    Args:
        ifname (str): Имя модемного интерфейса.
    """

    def __init__(self, ifname: str) -> None:
        self.ifname = ifname

    def execute(self) -> str:
        """Get connection status.

        Returns:
            str: modem connection status.
        """
        bearer_statistic = get_mmcli_bearer_statistic()
        connection_status: str = bearer_statistic['bearer']['status']['connected']
        return 'disconnected' if connection_status == 'no' else 'connected'


class GetModemSignalQualityCommand(GetIntFieldValueCommandInterface[str]):
    """Command for getting modem signal quality.

    Args:
        ifname (str): Имя модемного интерфейса.
    """

    def __init__(self, ifname: str) -> None:
        """Init variables."""
        self.ifname = ifname

    def execute(self) -> str:
        """Get signal quality.

        Returns:
            str: signal quality in percent.
        """
        modem_statistic = get_mmcli_modem_statistic()
        signal_quality: str = modem_statistic["modem"]['generic']['signal-quality']['value']
        return signal_quality


def get_modem_interface_name() -> str:
    """Находит имя l3 интерфейса модема.

    Returns:
        str: Имя l3 интерфейса модема.
    """
    modem_statistic = get_mmcli_modem_statistic()
    for iface in modem_statistic["modem"]['generic']['ports']:
        if iface.endswith('(net)'):
            iface_name: str = iface.split()[0]
            return iface_name
    raise Error('L3 Modem interface not found.')


INTERFACE_REPORT_FIELDS_FABRIC: dict[str, Type[GetIntFieldValueCommandInterface[Any]]] = {
    'name': GetInterfaceNameCommand,
    'status': GetInterfaceStatusViaIoctlCommand,
    'ip': GetInterfaceIpAddressViaIoctlCommand,
    'netmask': GetInterfacePrefixViaIoctlCommand,
    'mac': GetInterfaceMacAddressViaIoctlCommand,
    'gateway': GetGatewayCommand,
    'dns_servers': GetInterfaceDnsServersCommand,
    'connection_mode': GetModemConnectionModeCommand,
    'connection_status': GetModemConnectionStatusCommand,
    'signal_quality': GetModemSignalQualityCommand,
}


class MakeInterfaceReportCommand(Command[dict[str, Any]]):
    """Interface summary report maker.

    Args:
        ifname (str): Имя интерфейса.
        report_fields (Iterable[str]): Список полей интерфейса.
        field_fabric (dict[str, Type[GetIntFieldValueCommandInterface[Any]]], optional): Фабрика по
        высчитыванию значений полей интерфейса. Defaults to INTERFACE_REPORT_FIELDS_FABRIC.
    """

    def __init__(
        self, ifname: str, report_fields: Iterable[str],
        field_fabric: dict[str, Type[GetIntFieldValueCommandInterface[Any]]] = INTERFACE_REPORT_FIELDS_FABRIC,
    ) -> None:
        self.ifname = ifname
        self.report_fields = report_fields
        self.field_fabric = field_fabric

        self.report: dict[str, Any] = {}

    def execute(self) -> dict[str, Any]:
        """Make interface summary report."""
        for field in self.report_fields:
            try:
                value = self.field_fabric[field](self.ifname).execute()
            except BaseException:
                value = ''
            self.report[field] = value
        return self.report


def get_interfaces_summary_report(
    ifnames: Sequence[str], iface_report_fields: list[str], include_loopback: bool = False,
) -> list[dict[str, str]]:
    """Generate interfaces params report.

    Args:
        ifnames (Sequence[str]):                list of interface names.
        iface_report_fields (list[str]):        list of the report fields.
        include_loopback (bool, optional):      Include or not loopback interface in report.
                                                Defaults to False.

    Returns:
        List[Dict[str, str]]: Report in json serializable format.
    """
    interfaces_summary_report: list[dict[str, str]] = []
    for ifname in ifnames:
        is_loopback = CheckInterfaceFlagCommand(ifname, IFF_LOOPBACK).execute()
        if not include_loopback and is_loopback:
            continue
        interface_report = MakeInterfaceReportCommand(
                ifname,
                report_fields=iface_report_fields,
                field_fabric=INTERFACE_REPORT_FIELDS_FABRIC,
            ).execute()
        interfaces_summary_report.append(interface_report)
    return interfaces_summary_report


def is_modem_installed() -> bool:
    """Проверяет установлен ли модем в системе.

    Returns:
        bool: Установлен ли модем в системе.
    """
    return pathlib.Path(MODEM_DEVICE_PATH).exists()
