"""Publisher commands."""
import logging

from system_configurator.core.api.api_mqtt.mqtt_clients import IMqttMsgSender
from system_configurator.core.api.api_mqtt import msg_sender
from system_configurator.core.api.api_mqtt.publishers import IPublisherCommand
from system_configurator.network.exceptions import ModemNotInstalled

from system_configurator.network.services.network_interface_services import (
    WiredIfaceSettingsDict,
    ModemIfaceSettingsDict,
    get_modem_iface_current_settings,
    get_wired_iface_current_settings,
)


logger = logging.getLogger(__name__)


class PublishWiredInterfacesCredentialsCommand(IPublisherCommand):
    """Команда для публикации статистики проводного сетевого интерфейса."""

    def __init__(self, msg_sender: IMqttMsgSender = msg_sender) -> None:
        self._msg_sender = msg_sender
        self._last_result: WiredIfaceSettingsDict

    def __call__(self, publisher_topic: str) -> None:
        """Отправляет сообщение со статистикой проводного сетевого интерфейса на публикацию.

        Args:
            publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.
        """
        wired_credential = get_wired_iface_current_settings()
        if hasattr(self, '_last_result') and wired_credential == self._last_result:
            return
        else:
            self._last_result = wired_credential
        self._msg_sender.send_msg(
            topic=publisher_topic,
            payload=wired_credential,
            retain=True,
            cor_data=None,
        )
        logger.debug('Published Wired interface settings.')


class PublishModemInterfacesCredentialsCommand(IPublisherCommand):
    """Команда для публикации статистики модемного сетевого интерфейса."""

    def __init__(self, msg_sender: IMqttMsgSender = msg_sender) -> None:
        self._msg_sender = msg_sender
        self._last_result: ModemIfaceSettingsDict

    def __call__(self, publisher_topic: str) -> None:
        """Отправляет сообщение со статистикой модемного сетевого интерфейса на публикацию.

        Args:
            publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.
        """
        try:
            modem_credential = get_modem_iface_current_settings()
        except ModemNotInstalled:
            return

        if hasattr(self, '_last_result') and modem_credential == self._last_result:
            return
        else:
            self._last_result = modem_credential

        self._msg_sender.send_msg(
            topic=publisher_topic,
            payload=modem_credential,
            retain=True,
            cor_data=None,
        )
        logger.debug('Published Modem interface settings.')
