"""Команды для обработки mqtt роутером."""
from abc import ABC
import json
import logging
from typing import Any, Optional, Type, Union, TypeVar

from pydantic import BaseModel, ValidationError, parse_raw_as

from system_configurator.core.api.api_mqtt.entities import ResponsePayloadDict
from system_configurator.core.exceptions import ValidationException

from system_configurator.network.api.api_mqtt.schema import (
    ApnCredentialSchema,
    DhcpDnsCredentialSchema,
    DhcpIpCredentialSchema,
    StaticDnsCredentialSchema,
    StaticIPCredentialSchema,
)
from system_configurator.network.services.network_interface_services import (
    SetApnCredentialService,
    SetDnsCredentialService,
    SetIPCredentialService,
)


logger = logging.getLogger(__name__)


T = TypeVar('T', bound=BaseModel)


class AbstractRouterCommand(ABC):
    """Абстрактный класс для обработки запроса от роутера.

    Class Attrs:
        service (Type[IRouterCommand]): Сервис бизнес логики для обработки payload.
        parser (Type[BaseModel]): Парсер для обработки чистых данных payload и проверки типов.
    """

    service: Any
    parser: Type[BaseModel]

    def execute(self, payload: bytes) -> Optional[ResponsePayloadDict]:
        """Запускает команду обработки mqtt payload.

        Args:
            payload (bytes): тело запроса.
        """
        self._parse_input(payload)
        self._run_service()
        return {'status': 'success', 'msg': 'OK', 'data': None}

    def _parse_input(self, data: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._data_parsed = parse_raw_as(self.parser, data)
        except ValidationError as err:
            print(err.errors())
            raise ValidationException(message='Payload validation error.', data=err.errors())
        except json.decoder.JSONDecodeError:
            raise ValidationException(message='Payload must contain json data.')

        logger.debug('Input data validated successful.')

    def _run_service(self) -> None:
        self.service(self._data_parsed)


class SetIpCredentialToWiredIfaceCommand(AbstractRouterCommand):
    """Команда установки новых настроек ip адресации для проводного сетевого интерфейса."""

    service = SetIPCredentialService()
    parser = Union[StaticIPCredentialSchema, DhcpIpCredentialSchema]  # type: ignore

    def _run_service(self) -> None:
        self.service('wired', self._data_parsed)


class SetDnsCredentialToWiredIfaceCommand(AbstractRouterCommand):
    """Команда установки новых настроек dns для проводного сетевого интерфейса."""

    service = SetDnsCredentialService()
    parser = Union[StaticDnsCredentialSchema, DhcpDnsCredentialSchema]  # type: ignore

    def _run_service(self) -> None:
        self.service('wired', self._data_parsed)


class SetIpCredentialToModemIfaceCommand(AbstractRouterCommand):
    """Команда установки новых настроек ip адресации для модемного сетевого интерфейса."""

    service = SetIPCredentialService()
    parser = Union[StaticIPCredentialSchema, DhcpIpCredentialSchema]  # type: ignore

    def _run_service(self) -> None:
        self.service('modem', self._data_parsed)


class SetDnsCredentialToModemIfaceCommand(AbstractRouterCommand):
    """Команда установки новых настроек dns для модемного сетевого интерфейса."""

    service = SetDnsCredentialService()
    parser = Union[StaticDnsCredentialSchema, DhcpDnsCredentialSchema]  # type: ignore

    def _run_service(self) -> None:
        self.service('modem', self._data_parsed)


class SetApnCredentialToModemIfaceCommand(AbstractRouterCommand):
    """Команда установки новых настроек apn для модемного сетевого интерфейса."""

    service = SetApnCredentialService()
    parser = ApnCredentialSchema

    def _run_service(self) -> None:
        self.service('modem', self._data_parsed)
