"""Доменные модели для network settings."""
from abc import ABC
from typing import Literal, Protocol, runtime_checkable
import ipaddress


TAssignMethod = Literal['dhcp', 'manual']


class IEthernetIface(Protocol):
    """Интерфейс для сетевого интерфейса типа ethernet."""

    @property
    def ip_assign_method(self) -> TAssignMethod:
        """Способ назначения ip адреса.

        Может принимать два значения 'dhcp' и 'manual'.
        """
        ...

    @property
    def static_ip(self) -> str:
        """Статически заданный ip адрес."""
    ...

    @property
    def static_netmask(self) -> str:
        """Статически заданная маска сети."""
        ...

    @property
    def static_gateway(self) -> str:
        """Статически заданный шлюз по умолчанию."""
        ...

    @property
    def dns_assign_method(self) -> TAssignMethod:
        """Способ назначения dns серверов.

        Может принимать два значения 'dhcp' и 'manual'.
        """
        ...

    @property
    def static_dns_servers(self) -> list[str]:
        """Список статически заданных dns серверов."""
        ...

    def set_ip_assignment_method(self, method: TAssignMethod) -> None:
        """Устанавливает способ задания настроек ip.

        Args:
            method (TAssignMethod): Способ задания настроек ip.
        """
        ...

    def set_static_ip_credential(self, ip: str, netmask: str, gateway: str = '') -> None:
        """Устанавливает ip адрес.

        Args:
            ip (str):                   ip адрес.
            netmask (str):              Сетевая маска.
            gateway (str, optional):    Шлюз по умолчанию. Defaults to ''.
        """
        ...

    def set_dns_assignment_method(self, method: TAssignMethod) -> None:
        """Устанавливает способ назначения dns серверов.

        Args:
            method (TAssignMethod): Способ назначения dns серверов.
        """
        ...

    def set_static_dns_servers(self, servers: list[str]) -> None:
        """Устанавливает dns адреса.

        Args:
            servers (list[str]): Список dns адресов.
        """
        ...


@runtime_checkable
class IWiredIface(IEthernetIface, Protocol):
    """Интерфейс для проводного сетевого интерфейса."""
    ...


@runtime_checkable
class IModemIface(IEthernetIface, Protocol):
    """Интерфейс для модемного сетевого интерфейса."""

    @property
    def apn(self) -> str:
        """APN сервер."""
        ...

    @property
    def username(self) -> str:
        """Имя пользователя для apn подключения."""
        ...

    @property
    def password(self) -> str:
        """Пароль для APN подключения."""
        ...

    def set_apn_settings(self, apn: str, username: str = '', password: str = '') -> None:
        """Задает настройки APN подключения.

        Args:
            apn (str):                  APN сервер.
            username (str, optional):   Имя пользователя для apn подключения. Defaults to ''.
            password (str, optional):   Пароль для APN подключения. Defaults to ''.
        """
        ...


class EthernetIface(IEthernetIface, ABC):
    """Базовый класс сетевого интерфейса."""

    def __init__(self) -> None:
        self._ip_assign_method: TAssignMethod = 'dhcp'
        self._static_ip = ''
        self._static_netmask = ''
        self._static_gateway = ''
        self._dns_assign_method: TAssignMethod = 'dhcp'
        self._static_dns_servers: list[str] = []

    @property
    def ip_assign_method(self) -> TAssignMethod:
        """Способ назначения ip адреса.

        Может принимать два значения 'dhcp' и 'manual'.
        """
        return self._ip_assign_method

    @property
    def static_ip(self) -> str:
        """Статически заданный ip адрес."""
        return self._static_ip

    @property
    def static_netmask(self) -> str:
        """Статически заданная маска сети."""
        return self._static_netmask

    @property
    def static_gateway(self) -> str:
        """Статически заданный шлюз по умолчанию."""
        return self._static_gateway

    @property
    def dns_assign_method(self) -> TAssignMethod:
        """Способ назначения dns серверов.

        Может принимать два значения 'dhcp' и 'manual'.
        """
        return self._dns_assign_method

    @property
    def static_dns_servers(self) -> list[str]:
        """Список статически заданных dns серверов."""
        return self._static_dns_servers

    def set_ip_assignment_method(self, method: TAssignMethod) -> None:
        """Устанавливает способ задания настроек ip.

        Args:
            method (TAssignMethod): Способ задания настроек ip.
        """
        if method == self._ip_assign_method:
            return
        self._ip_assign_method = method
        self._erase_static_ip_credential()

        if self._ip_assign_method == 'manual':
            self.set_dns_assignment_method(self._ip_assign_method)

    def set_static_ip_credential(self, ip: str, netmask: str, gateway: str = '') -> None:
        """Устанавливает ip адрес.

        Args:
            ip (str):                   ip адрес.
            netmask (str):              Сетевая маска.
            gateway (str, optional):    Шлюз по умолчанию. Defaults to ''.
        """
        if self._ip_assign_method == 'dhcp':
            return
        self._validate_ip_address_with_mask(ip, netmask)
        self._validate_gateway_address(gateway)
        self._static_ip = ip
        self._static_netmask = netmask
        self._static_gateway = gateway

    def set_dns_assignment_method(self, method: TAssignMethod) -> None:
        """Устанавливает способ назначения dns серверов.

        Args:
            method (TAssignMethod): Способ назначения dns серверов.
        """
        if method == self._dns_assign_method:
            return
        if method == 'dhcp' and self._ip_assign_method == 'manual':
            return
        self._dns_assign_method = method
        self._erase_static_dns_credential()

    def set_static_dns_servers(self, servers: list[str]) -> None:
        """Устанавливает dns адреса.

        Args:
            servers (list[str]): Список dns адресов.
        """
        if self._dns_assign_method == 'dhcp':
            return
        self._validate_dns_servers(servers)
        self._static_dns_servers = servers

    def _erase_static_ip_credential(self) -> None:
        self._static_ip = ''
        self._static_netmask = ''
        self._static_gateway = ''

    def _erase_static_dns_credential(self) -> None:
        self._static_dns_servers = []

    def _validate_dns_servers(self, servers: list[str]) -> None:
        for server in servers:
            try:
                ipaddress.IPv4Address(server)
            except ipaddress.AddressValueError:
                ValueError('Wrong dns server address.')

    def _validate_gateway_address(self, gateway: str) -> None:
        if not gateway:
            return
        try:
            ipaddress.IPv4Address(gateway)
        except ipaddress.AddressValueError:
            raise ValueError('Wrong gateway address.')

    def _validate_ip_address_with_mask(self, ip: str, netmask: str) -> None:
        try:
            ipaddress.IPv4Interface(f'{ip}/{netmask}')
        except (ipaddress.AddressValueError, ipaddress.NetmaskValueError):
            raise ValueError(f'Wrong ip address or network mask {ip}/{netmask}.')


class WiredIface(EthernetIface):
    """Проводной сетевой интерфейс."""
    ...


class ModemIface(EthernetIface, IModemIface):
    """Модемный сетевой интерфейс."""

    def __init__(self) -> None:
        super().__init__()
        self._apn = ''
        self._username = ''
        self._password = ''

    @property
    def apn(self) -> str:
        """APN сервер."""
        return self._apn

    @property
    def username(self) -> str:
        """Имя пользователя для apn подключения."""
        return self._username

    @property
    def password(self) -> str:
        """Пароль для APN подключения."""
        return self._password

    def set_apn_settings(self, apn: str, username: str = '', password: str = '') -> None:
        """Задает настройки APN подключения.

        Args:
            apn (str):                  APN сервер.
            username (str, optional):   Имя пользователя для apn подключения. Defaults to ''.
            password (str, optional):   Пароль для APN подключения. Defaults to ''.
        """
        self._apn = apn
        self._username = username
        self._password = password


if __name__ == '__name__':
    iface_wired: IWiredIface = WiredIface()
    isinstance(iface_wired, IWiredIface)

    iface_modem: IModemIface = ModemIface()
    isinstance(iface_modem, IWiredIface)
