"""Сервисы для управления сетевыми интерфейсами."""
from ipaddress import IPv4Address
from typing import Literal, Protocol, TypedDict, Union, runtime_checkable

from system_configurator.network.adapters.netplan.uow import NetplanUowManager
from system_configurator.network.domain.models import TAssignMethod
from system_configurator.network.adapters import utils
from system_configurator.network.exceptions import ModemNotInstalled


TInterfaceTypes = Literal['wired', 'modem']


@runtime_checkable
class IStaticIPCredential(Protocol):
    """Интерфейс для статических настроек ip адресации.

    ip_assign_method (Literal['manual']):   Способ назначения ip адреса. Должно быть manual.
    static_ip (IPv4Address):                IPv4 адрес интерфейса.
    static_netmask (IPv4Address):           Сетевая маска интерфейса.
    static_gateway (IPv4Address):           Шлюз по умолчанию.
    """

    ip_assign_method: Literal['manual']
    static_ip: IPv4Address
    static_netmask: IPv4Address
    static_gateway: IPv4Address


@runtime_checkable
class IDhcpIpCredential(Protocol):
    """Интерфейс для dhcp настроек ip адресации.

    Args:
        ip_assign_method (Literal['dhcp']): Способ назначения ip адреса. Должно быть dhcp.
    """
    ip_assign_method: Literal['dhcp']


@runtime_checkable
class IStaticDnsCredential(Protocol):
    """Интерфейс для статических настроек dns серверов.

    Args:
        dns_assign_method (Literal['manual']):  Способ назначения dns серверов. Должно быть manual.
        static_dns_servers (list[IPv4Address]): Список DNS серверов.
    """
    dns_assign_method: Literal['manual']
    static_dns_servers: list[IPv4Address]


@runtime_checkable
class IDhcpDnsCredential(Protocol):
    """Интерфейс для dhcp настроек dns серверов.

    Args:
        dns_assign_method (Literal['dhcp']):  Способ назначения dns серверов. Должно быть dhcp.
    """
    dns_assign_method: Literal['dhcp']


@runtime_checkable
class IApnCredential(Protocol):
    """Интерфейс для настройки APN подключения.

    Args:
        apn (str):          APN сервер.
        username (str):     Имя пользователя если есть либо пустая строка.
        password (str):     Пароль если есть либо пустая строка.
    """
    apn: str
    username: str
    password: str


class WiredIfaceSettingsDict(TypedDict):
    """Словарь с настройками проводного интерфейса.

    Args:
        status (str): Статус интерфейса. Может быть up или down.
        ip_assign_method (Literal['manual', 'dhcp']): Способ назначения ip адреса. Может быть manual или dhcp.
        ip (str): IP адрес интерфейса.
        netmask (str): Маска интерфейса.
        gateway (str): Шлюз по умолчанию.
        dns_assign_method (Literal['manual', 'dhcp']): Способ назначения dns серверов. Может быть manual или dhcp.
        dns_servers (list[str]): Список dns серверов.
        mac_address (str): MAC адрес интерфейса.
    """
    status: str
    ip_assign_method: TAssignMethod
    ip: str
    netmask: str
    gateway: str
    dns_assign_method: TAssignMethod
    dns_servers: list[str]
    mac_address: str


class ModemIfaceSettingsDict(WiredIfaceSettingsDict):
    """Словарь с настройками модемного интерфейса.

    Args:
        status (str): Статус интерфейса. Может быть up или down.
        ip_assign_method (Literal['manual', 'dhcp']): Способ назначения ip адрecа. Может быть manual или dhc.
        ip (str): IP адрес интерфейса.
        netmask (str): Маска интерфейса.
        gateway (str): Шлюз по умолчанию.
        dns_assign_method (Literal['manual', 'dhcp']): Способ назначения dns серверов. Может быть manual или dhcp.
        dns_servers (list[str]): Список dns серверов.
        mac_address (str): MAC адрес интерфейса.
        apn (str): APN сервер.
        username (str): Имя пользователя для apn сервера.
        password (str): Пароль для apn сервера.
        connection_mode (str): Режим подключения (LTE, UMTS и т.д.).
        connection_status (str): Статус подключения.
        signal_quality (str): Качество сигнала в процентах.

    """

    apn: str
    username: str
    password: str
    connection_mode: str
    connection_status: str
    signal_quality: str


class SetIPCredentialService:
    """Сервис для настройки ip адресации."""

    def __call__(
        self,
        ifname: TInterfaceTypes,
        credential: Union[IStaticIPCredential, IDhcpIpCredential],
    ) -> None:
        """Устанавливает ip адресацию для интерфейса.

        Args:
            ifname (Literal['wired', 'modem']): Тип интерфеса.
            credential (Union[IStaticIPCredential, IDhcpIpCredential]): Новые настройки ip адресации.
        """
        with NetplanUowManager() as (uow, repository):
            iface = repository.get(ifname)
            iface.set_ip_assignment_method(credential.ip_assign_method)
            if isinstance(credential, IStaticIPCredential):
                iface.set_static_ip_credential(
                    str(credential.static_ip),
                    str(credential.static_netmask),
                    str(credential.static_gateway),
                )
            repository.update(iface)
            uow.commit()


class SetDnsCredentialService:
    """Сервис для настройки dns."""

    def __call__(
        self,
        ifname: TInterfaceTypes,
        credential: Union[IStaticDnsCredential, IDhcpDnsCredential],
    ) -> None:
        """Устанавливает dns сервера для интерфейса.

        Args:
            ifname (Literal['wired', 'modem']): Тип интерфеса.
            credential (Union[IStaticDnsCredential, IDhcpDnsCredential]): Новые настройки dns серверов.
        """
        with NetplanUowManager() as (uow, repository):
            iface = repository.get(ifname)
            iface.set_dns_assignment_method(credential.dns_assign_method)
            if isinstance(credential, IStaticDnsCredential):
                iface.set_static_dns_servers([str(dns_server) for dns_server in credential.static_dns_servers])
            repository.update(iface)
            uow.commit()


class SetApnCredentialService:
    """Сервис для настройки apn подключения."""

    def __call__(
        self,
        ifname: Literal['modem'],
        credential: IApnCredential,
    ) -> None:
        """Устанавливает настройки apn подключения.

        Args:
            ifname (Literal['wired', 'modem']): Тип интерфеса.
            credential (Union[IApnCredential]): Новые настройки apn подключения.
        """
        with NetplanUowManager() as (uow, repository):
            iface = repository.get(ifname)
            iface.set_apn_settings(
                credential.apn,
                credential.username,
                credential.password,
            )
            repository.update(iface)
            uow.commit()


def get_wired_iface_current_settings() -> WiredIfaceSettingsDict:
    """Собирает настройки проводного сетевого интерфейса.

    Returns:
        WiredIfaceSettingsDict: Настройки проводного сетевого интерфейса
    """
    runtime_report_fields = ['ip', 'netmask', 'gateway', 'mac', 'status', 'dns_servers']
    ifconfig_runtime = utils.MakeInterfaceReportCommand('eth0', runtime_report_fields).execute()
    with NetplanUowManager() as (_, repository):
        ifconfig_permanent = repository.get('wired')
    return WiredIfaceSettingsDict(
        status=ifconfig_runtime.get('status') or 'down',
        ip_assign_method=ifconfig_permanent.ip_assign_method,
        ip=ifconfig_permanent.static_ip or ifconfig_runtime['ip'],
        netmask=ifconfig_permanent.static_netmask or ifconfig_runtime['netmask'],
        gateway=ifconfig_permanent.static_gateway or ifconfig_runtime['gateway'],
        dns_assign_method=ifconfig_permanent.dns_assign_method,
        dns_servers=ifconfig_permanent.static_dns_servers or ifconfig_runtime['dns_servers'],
        mac_address=ifconfig_runtime.get('mac') or '',
    )


def get_modem_iface_current_settings() -> ModemIfaceSettingsDict:
    """Собирает настройки модемного сетевого интерфейса.

    Returns:
        ModemIfaceSettingsDict: Настройки модемного сетевого интерфейса
    """
    if not utils.is_modem_installed():
        raise ModemNotInstalled()

    runtime_report_fields = [
        'ip', 'netmask', 'gateway', 'mac', 'status', 'dns_servers',
        'connection_mode', 'connection_status', 'signal_quality',
    ]
    ifname = utils.get_modem_interface_name()
    ifconfig_runtime = utils.MakeInterfaceReportCommand(ifname, runtime_report_fields).execute()
    with NetplanUowManager() as (_, repository):
        ifconfig_permanent = repository.get('modem')

    return ModemIfaceSettingsDict(
        status=ifconfig_runtime.get('status') or 'down',
        ip_assign_method=ifconfig_permanent.ip_assign_method,
        ip=ifconfig_runtime.get('ip') or ifconfig_permanent.static_ip,
        netmask=ifconfig_runtime.get('netmask') or ifconfig_permanent.static_netmask,
        gateway=ifconfig_runtime.get('gateway') or ifconfig_permanent.static_gateway,
        dns_assign_method=ifconfig_permanent.dns_assign_method,
        dns_servers=ifconfig_runtime.get('dns_servers') or ifconfig_permanent.static_dns_servers,
        mac_address=ifconfig_runtime.get('mac') or '',
        apn=ifconfig_permanent.apn,
        username=ifconfig_permanent.username,
        password=ifconfig_permanent.password,
        connection_mode=ifconfig_runtime.get('connection_mode') or '',
        connection_status=ifconfig_runtime.get('connection_status') or '',
        signal_quality=ifconfig_runtime.get('signal_quality') or '',
    )
