"""Time and date utilities."""
import logging
from typing import List, TypedDict
import subprocess

from system_configurator.core.file_system import save_config


class TimezoneInfoDict(TypedDict):
    """Timezone Info.

    timezone: str - timezone name
    """

    timezone: str


class DateInfoDict(TypedDict):
    """Date Info.

    date: str - date in format 'Y:M:D'
    """

    date: str


class TimeInfoDict(TypedDict):
    """Time Info.

    time: str - time in format 'HH:mm:ss'
    """

    time: str


class DateTimeInfoDict(TypedDict):
    """Date and Time Info.

    datetime: str - time in format 'Y:M:D HH:mm:ss'
    """

    datetime: str


class NtpSyncStatusInfoDict(TypedDict):
    """NTP synchronization status info.

    ntp: bool - is ntp sync enable
    """

    ntp: bool


class NtpServerAddressesInfoDict(TypedDict):
    """NTP server addresses info.

    ntp_servers: list[str] - list of servers ip addresses or dns names
    """

    ntp_servers: list[str]


class SetSystemTimezoneCommand():
    """Set system timezone.

    Args:
        data (TimezoneInfoDict): dict with timezone name.

    Returns:
        None
    """

    def __init__(self, data: TimezoneInfoDict) -> None:
        self.timezone = data['timezone']

    def execute(self) -> None:
        """Run command."""
        cmd: List[str] = ['sudo', 'timedatectl', 'set-timezone', self.timezone]
        subprocess.run(cmd, capture_output=True)


class SetSystemDateCommand():
    """Set system date.

    Args:
        data (DateInfoDict): dict with new date in 'Y:M:D' format.

    Returns:
        None
    """

    def __init__(self, data: DateInfoDict) -> None:
        self.date = data['date']

    def execute(self) -> None:
        """Run command."""
        cmd: List[str] = ['sudo', 'timedatectl', 'set-time', self.date]
        subprocess.run(cmd)


class SetSystemTimeCommand():
    """Set system time.

    Args:
        data (TimeInfoDict): dict with new system time in 'HH:mm:ss' format.

    Returns:
        None
    """

    def __init__(self, data: TimeInfoDict) -> None:
        self.time = data['time']

    def execute(self) -> None:
        """Run command."""
        cmd: List[str] = ['sudo', 'timedatectl', 'set-time', self.time]
        subprocess.run(cmd)


class SetSystemDatetimeCommand():
    """Set system date and time.

    Args:
        data (DateTimeInfoDict): dict with new system date and time in 'Y:M:D HH:mm:ss' format.

    Returns:
        None
    """

    def __init__(self, data: DateTimeInfoDict) -> None:
        self.datetime = data['datetime']

    def execute(self) -> None:
        """Run command."""
        cmd: List[str] = ['sudo', 'timedatectl', 'set-time', self.datetime]
        subprocess.run(cmd)
        logging.info('System time was changed to %s', self.datetime)


class ChangeNtpSynchronizationCommand():
    """Enable or disable ntp synchronization.

    Args:
        data (NtpSyncStatusInfoDict): dict with new ntp sync status.

    Returns:
        None
    """

    def __init__(self, data: NtpSyncStatusInfoDict) -> None:
        self.is_ntp_enable = data['ntp']

    def execute(self) -> None:
        """Run command."""
        formated_sync_status: str = str(self.is_ntp_enable).lower()
        cmd: List[str] = ['sudo', 'timedatectl', 'set-ntp', formated_sync_status]
        subprocess.run(cmd)


class SetNtpServersCommand():
    """Set ntp servers.

    Generate ntp config, replace it then restart systemd-timesyncd.service

    Accepts list of ip addresses or domain names

    Args:
        data (NtpServerAddressesInfoDict): dict with list of ip addresses or domain names

    Returns:
        None
    """

    def __init__(self, data: NtpServerAddressesInfoDict) -> None:
        self.ntp_servers = data.get('ntp_servers')

    def execute(self) -> None:
        """Run command."""
        data = {'ntp_servers': self.ntp_servers}
        save_config(
            data=data,
            template='timesyncd.conf',
            config='timesyncd.conf',
            config_path='/etc/systemd',
        )
        cmd: List[str] = ['sudo', 'systemctl', 'restart', 'systemd-timesyncd.service']
        subprocess.run(cmd)
        logging.info('New ntp servers was configured. %s', self.ntp_servers)
