"""Async Actions."""
import asyncio
import json
import logging
import queue
from typing import (
    Any,
    Optional,
    Protocol,
    Sequence,
)

from triggers.actions.base import IAction, IMqttMsgParams, TPossibleActionTypes
from triggers.core.msg_handlers import RespMsgDict
from triggers.db.database import AsyncSessionLocal
from triggers.db.schema import ActionSchema
from triggers.db.utils import get_serialized_actions_related_trigger

from triggers.listeners.base import ITrigger


logger = logging.getLogger(__name__)


class IRunActionCommand(Protocol):
    """Интерфейс для обработчика action.

    Args:
        action (IAction[IMqttMsgParams]): объект с параметрами действия.
        trigger (ITrigger[Any]): объект с параметрами триггера.
    """

    def __call__(self, action: IAction[IMqttMsgParams], trigger: ITrigger[Any]) -> None:
        """Выполняет действие.

        Args:
            action (IAction[IMqttMsgParams]): объект с параметрами действия.
            trigger (ITrigger[Any]): объект с параметрами триггера.
        """
        ...


class IRunMqttSendMsgActionCommand(Protocol):
    """Интерфейс для объекта при вызов которого помещает сообщение mqtt в очередь для отправки."""

    _msg_queue: queue.Queue[RespMsgDict]

    def __call__(self, action: IAction[IMqttMsgParams], trigger: ITrigger[Any]) -> None:
        """Выполняет действие.

        Args:
            action (IAction[IMqttMsgParams]): объект с параметрами действия.
            trigger (ITrigger[Any]): объект с параметрами триггера.
        """
        ...


class RunMqttSendRawMsgActionCommand(IRunMqttSendMsgActionCommand):
    """При вызове помещает сообщение mqtt с произвольным payload в очередь для отправки.

    Args:
        msg_queue (queue.Queue[RespMsgDict]): очередь для отправки сообщений.
    """

    def __init__(self, msg_queue: queue.Queue[RespMsgDict]) -> None:
        self._msg_queue = msg_queue

    def __call__(self, action: IAction[IMqttMsgParams], trigger: ITrigger[Any]) -> None:
        """Помещает сообщение mqtt с произвольным payload в очередь для отправки.

        Args:
            action (IAction[IMqttMsgParams]): объект с параметрами действия.
            trigger (ITrigger[Any]): объект с параметрами триггера.
        """
        payload = json.loads(action.params.payload)
        msg: RespMsgDict = {
            'topic': action.params.topic,
            'cor_data': None,
            'payload': payload,
            'retain': action.params.retain,
        }
        logger.info(
            'Action %s generated message to send over mqtt. Topic: %s. Payload: %s',
            action.name,
            action.params.topic,
            payload,
        )
        self._msg_queue.put(msg)


class RunMqttSendTriggerMsgActionCommand(IRunMqttSendMsgActionCommand):
    """При вызове помещает сообщение mqtt с триггером в payload в очередь для отправки.

    Формат payload:
        '{trigger: ITrigger[Any]}'

    Args:
        msg_queue (queue.Queue[RespMsgDict]): очередь для отправки сообщений.
    """

    def __init__(self, msg_queue: queue.Queue[RespMsgDict]) -> None:
        self._msg_queue = msg_queue

    def __call__(self, action: IAction[IMqttMsgParams], trigger: ITrigger[Any]) -> None:
        """Помещает сообщение mqtt с триггером в payload в очередь для отправки.

        Args:
            action (IAction[IMqttMsgParams]): объект с параметрами действия.
            trigger (ITrigger[Any]): объект с параметрами триггера.
        """
        serialized_trigger = trigger.__dict__.copy()
        serialized_trigger['params'] = trigger.params.__dict__.copy()
        payload = {'trigger': serialized_trigger}
        msg: RespMsgDict = {
            'topic': action.params.topic,
            'cor_data': None,
            'payload': payload,
            'retain': False,
        }
        self._msg_queue.put(msg)


send_queue: queue.Queue[RespMsgDict] = queue.Queue()


ACTION_COMMAND_FABRIC: dict[TPossibleActionTypes, IRunActionCommand] = {
    'send_mqtt_msg_raw': RunMqttSendRawMsgActionCommand(send_queue),
    'send_trigger_to_mqtt': RunMqttSendTriggerMsgActionCommand(send_queue),
}


async def handle_activated_trigger(
    triggers_active_queue: asyncio.Queue[ITrigger[Any]],
    action_command_fabric: dict[TPossibleActionTypes, IRunActionCommand]
) -> None:
    """Запускает action при получении сработавших триггеров.

    Args:
        triggers_active_queue (asyncio.Queue[ITrigger[Any]]): очередь для получения сработавших триггеров
        action_command_fabric (dict[TPossibleActionTypes, IRunActionCommand]): Фабрика actions.
    """
    while True:
        trigger = await triggers_active_queue.get()
        actions: Sequence[IAction[Any]] = await get_actions_by_trigger(trigger)
        if actions:
            for action in actions:
                command: Optional[IRunActionCommand] = action_command_fabric.get(action.action_type)
                if command:
                    command(action, trigger)


async def get_actions_by_trigger(trigger: ITrigger[Any]) -> list[ActionSchema]:
    """Получает actions из базы по триггеру.

    Args:
        trigger (ITrigger[Any]): Сработавший триггер

    Returns:
        list[ActionSchema]: Список actions
    """
    async with AsyncSessionLocal() as async_session:
        actions = await get_serialized_actions_related_trigger(async_session, trigger)
    return actions
