"""Mqtt API."""
import asyncio
import json
import logging
from queue import Empty, Queue
import threading
from typing import Any, Optional, TypedDict

from paho.mqtt.client import Client, MQTTMessage
from paho.mqtt.properties import Properties
from paho.mqtt.packettypes import PacketTypes

from triggers.core.msg_handlers import RespMsgDict


logger = logging.getLogger(__name__)


class MqttUserdataDict(TypedDict):
    """Additional data for paho mqtt client.

    topics: list[str]                   - list of topics name for subscribe.
    msg_queue: Queue[MQTTMessage]       - queue for received msg.
    publisher_queue: Queue[Any]         - queue for publishing msg.
    publisher_loop_timeout: float       - timeout for publishing loop.
    """

    topics: list[str]
    msg_queue: Queue[MQTTMessage]
    publisher_queue: Queue[Any]
    publisher_loop_timeout: float


class MqttUserdataWithAsyncQueueDict(TypedDict):
    """Additional data for paho mqtt client.

    topics:                  - list of topics name for subscribe.
    msg_queue:               - async queue for received msg.
    publisher_queue:         - queue for publishing msg.
    publisher_loop_timeout:  - timeout for publishing loop.
    """

    topics: list[str]
    msg_queue: asyncio.Queue[MQTTMessage]
    publisher_queue: Queue[Any]
    publisher_loop_timeout: float


class MQTTReceiver(Client):
    """Mqtt receiver client."""

    def on_connect(
        self, client: Client, userdata: dict[str, Any], flags: Any, rc: int, properties: Optional[Properties] = None,
    ) -> None:
        """Define the connect callback implementation.

        Subscribe to all sub topics listed in userdata.

        Args:
            client (Client):                - the client instance for this callback
            userdata (dict[str, Any]):      - the private user data as set in Client() or userdata_set()
            flags (Any):                    - gives the severity of the message
            rc (int):                       - return code
            properties (Properties):        - MQTTv5 properties
        """
        logger.debug("Connected to mqtt server with result code %s", rc)
        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        if topics := userdata.get('topics'):
            for topic in topics:
                self.subscribe(topic)

    def on_disconnect(
        self, client: Client, userdata: dict[str, Any], rc: int, properties: Optional[Properties] = None,
    ) -> None:
        """Define the disconnect callback implementation.

        Logged reason code.

        Args:
            client (Client):                - the client instance for this callback
            userdata (dict[str, Any]):      - the private user data as set in Client() or userdata_set()
            rc (int):                       - return code
            properties (Properties):        - MQTTv5 properties
        """
        logger.error('Mqtt client has been disconnected with code %s', rc)

    def on_message(self, mqttc: Client, userdata: Any, msg: MQTTMessage) -> None:
        """Define the message received callback implementation.

        Serialize payload and put msg to queue for further processing.
        Queue is taken from userdata.

        Args:
            mqttc (Client):                 - the client instance for this callback
            userdata (Any):                 - the private user data as set in Client() or userdata_set()
            msg (MQTTMessage):              - Mqtt message
        """
        logger.debug('Mqtt message has been received: %s %s', msg.topic, str(msg.payload))
        if msg_queue := userdata.get('msg_queue'):
            if isinstance(msg_queue, asyncio.Queue):  # FIXME: Грязь. Переделать.
                asyncio.run_coroutine_threadsafe(msg_queue.put(msg), msg_queue._loop)
            else:
                msg_queue.put(msg)


class MqttRPClient(MQTTReceiver):
    """Mqtt receiver and publisher client.

    Realize publishing in loop from queues.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Add params to standard paho mqtt client init method.

        Added params:
            self._publisher_thread_terminate: bool
            self._publisher_thread: Optional[threading.Thread]
        """
        self._publisher_thread_terminate: bool = False
        self._publisher_thread: Optional[threading.Thread] = None
        super().__init__(*args, **kwargs)

    def publisher_loop_forever(self) -> None:
        """Read publishing queues and publish their messages to mqtt topics in blocking infinity loop."""
        # def publish(self, topic, payload=None, qos=0, retain=False, properties=None):
        resp_queue = self._userdata['publisher_queue']
        while not self._publisher_thread_terminate:
            try:
                resp_msg: RespMsgDict = resp_queue.get(timeout=self._userdata.get('publisher_loop_timeout') or 0.5)
            except Empty:
                continue

            topic = resp_msg.get('topic')
            cor_data = resp_msg.get('cor_data')
            retain = bool(resp_msg.get('retain'))
            try:
                json.loads(resp_msg['payload'])
            except TypeError:
                payload = json.dumps(resp_msg['payload'])
            else:
                payload = resp_msg['payload']

            if cor_data:
                properties = Properties(PacketTypes.PUBLISH)
                properties.CorrelationData = cor_data
            else:
                properties = None

            self.publish(topic, payload=payload, properties=properties, retain=retain)

            logger.debug('Msg has been published to topic "%s". Cor data: %s', topic, cor_data)

    def publisher_loop_start(self) -> None:
        """This is part of the threaded client interface for publisher.

        Call this once to start a new thread for publishing message.
        """
        if self._publisher_thread is not None:
            return

        self._publisher_thread_terminate = False
        self._publisher_thread = threading.Thread(target=self.publisher_loop_forever)
        self._publisher_thread.daemon = True
        self._publisher_thread.start()
        logger.debug('Mqtt publisher thread has been started.')

    def publisher_loop_stop(self) -> None:
        """This is part of the threaded client interface for publisher.

        Call this once to stop the network thread previously created with publisher_loop_start().
        This call will block until the network thread finishes.
        """
        if self._publisher_thread is None:
            return

        self._publisher_thread_terminate = True
        if threading.current_thread() != self._publisher_thread:
            self._publisher_thread.join()
            self._publisher_thread = None
            logger.debug('Mqtt publisher thread has been stopped.')


def send_to_mqtt(topic: str, payload: Any, retain: bool, publisher_queue: Queue[RespMsgDict]) -> None:
    """Помещает сообщение mqtt в очередь для отправки.

    Args:
        topic (str):    Топик для отправки
        payload (Any):  Полезная нагрузка
        retain (bool):  Оставлять ли сообщения в топике постоянно.
        publisher_queue (Queue[RespMsgDict]): Очередь для отправки сообщения.
    """
    pub_msg: RespMsgDict = {
        'topic': topic,
        'cor_data': None,
        'retain': retain,
        'payload': payload,
    }
    publisher_queue.put(pub_msg)
