"""Module contain trigger checkers."""
import asyncio
import logging
import socket
import time
from typing import Any, Protocol

from triggers.listeners.base import (
    ArtNetFrame,
    IArtNetParams,
    IRawUdpParams,
    ISocketParams,
    ITrigger,
    TPossibleTriggerTypes,
    IListener,
)


logger = logging.getLogger(__name__)


TCheckingQueue = asyncio.Queue[tuple[tuple[Any, ...], IListener[Any]]]


class ITriggerChecker(Protocol):
    """Интерфейс для вызываемого объекта который проверяет триггер на сработку.

    Принимает объект триггера и дополнительные параметры в зависимости от типа триггера.

    Возвращает True если триггер сработал и False если не сработал.
    """

    def __call__(self, trigger: ITrigger[Any], *args: Any, **kwds: Any) -> bool:
        """Проверяет триггер на сработку.

        Args:
            trigger (ITrigger[Any]): Объект с параметрами триггера.
        """
        ...


def _get_artnet_activated_triggers(trigger: ITrigger[IArtNetParams], soc: socket.socket, data: bytes) -> bool:
    """Проверяет соответствие сокета и полученных данных ArtNet триггеру.

    Args:
        soc (socket.socket): сокет
        data (bytes): data
        trigger (ITrigger[ISocketParams]): Проверяемый триггер.

    Returns:
        bool: True если триггер сработал и False если нет.
    """
    try:
        artnet_frame = ArtNetFrame(raw_bytes=data)
    except BaseException:
        return False

    checks = [
        artnet_frame.subuni == trigger.params.universe,
        trigger.params.min_level <= artnet_frame.data[trigger.params.channel - 1] <= trigger.params.max_level,
    ]
    return True if all(checks) else False


def _get_rawudp_activated_triggers(trigger: ITrigger[IRawUdpParams], soc: socket.socket, data: bytes) -> bool:
    """Проверяет соответствие сокета и полученных данных триггеру.

    Args:
        soc (socket.socket): сокет
        data (bytes): data
        trigger (ITrigger[ISocketParams]): Проверяемый триггер.

    Returns:
        bool: True если триггер сработал и False если нет.
    """
    checks = [
        trigger.params.data.encode().decode('unicode_escape').encode() == data,
    ]
    return True if all(checks) else False


TRIGGERS_CHECKS_FABRIC: dict[TPossibleTriggerTypes, ITriggerChecker] = {
    # Фабрика для сопоставления типа триггера с конкретной функцией проверки
    'ArtNet': _get_artnet_activated_triggers,
    'RawUDP': _get_rawudp_activated_triggers,
}


class TriggerCheckerLoop:
    """Сопоставляет принимаемые данные с триггерами в бесконечном цикле.

    Если триггер сопоставлен с полученными данными то он считается сработавшим и
    помещается в очередь сработавших триггеров для дальнейшей обработки.

    Args:
        checking_queue: Очередь для приема новых данных.
        activated_queue: Очередь в которую складываются сработавшие триггеры.
        listener_to_triggers_dict : Словарь оперативно заполняющийся данными сопоставления
        объектов триггеров с объектом listener.
        triggers_check_fabric: Фабрика для сопоставления типа триггера с конкретной функцией проверки.
    """
    def __init__(
        self,
        checking_queue: TCheckingQueue,
        activated_queue: asyncio.Queue[ITrigger[Any]],
        listener_to_triggers_dict: dict[IListener[Any], set[ITrigger[Any]]],
        triggers_check_fabric: dict[TPossibleTriggerTypes, ITriggerChecker]
    ) -> None:
        self._checking_queue = checking_queue
        self._activated_queue = activated_queue
        self._listener_to_triggers_dict = listener_to_triggers_dict
        self._checks_fabric = triggers_check_fabric

    async def start(self) -> None:
        """Асинхронно сопоставляет принимаемые данные с триггерами в бесконечном цикле.

        Если триггер сопоставлен с полученными данными то он считается сработавшим и
        помещается в очередь сработавших триггеров для дальнейшей обработки.
        """
        while True:
            data, listener = await self._checking_queue.get()

            triggers = self._listener_to_triggers_dict[listener]

            for trigger in triggers:
                check_command = self._checks_fabric[trigger.tr_type]
                is_activated: bool = check_command(trigger, *data)
                if is_activated:
                    logger.info('Trigger %s activated.', trigger.name)
                    await self._activated_queue.put(trigger)


class FlowTriggerCheckerLoop(TriggerCheckerLoop):
    """Сопоставляет принимаемые данные с триггерами в бесконечном цикле.

    Если триггер сопоставлен с полученными данными то он считается сработавшим и
    помещается в очередь сработавших триггеров для дальнейшей обработки.

    Если предыдущая сработка триггера была меньше секунды назад то он игнорируется.

    Args:
        checking_queue: Очередь для приема новых данных.
        activated_queue: Очередь в которую складываются сработавшие триггеры.
        listener_to_triggers_dict : Словарь оперативно заполняющийся данными сопоставления
        объектов триггеров с объектом listener.
        triggers_check_fabric: Фабрика для сопоставления типа триггера с конкретной функцией проверки.
    """

    async def start(self) -> None:
        """Асинхронно сопоставляет принимаемые данные с триггерами в бесконечном цикле.

        Если триггер сопоставлен с полученными данными то он считается сработавшим и
        помещается в очередь сработавших триггеров для дальнейшей обработки.

        Дополнительно проверяется что сработавший триггер не принадлежит потоку данных
        на котором уже была сработка до этого.
        """
        self._triggers_activated_time: dict[ITrigger[ISocketParams], float] = {}

        while True:
            data, listener = await self._checking_queue.get()
            triggers = self._listener_to_triggers_dict[listener]
            for trigger in triggers:
                check_command = self._checks_fabric[trigger.tr_type]
                is_activated: bool = check_command(trigger, *data)
                if is_activated:
                    if self._check_flow(trigger):
                        logger.info('Trigger %s activated.', trigger.name)
                        await self._activated_queue.put(trigger)


    def _check_flow(self, trigger: ITrigger[Any]) -> bool:
        activated_time: float = time.perf_counter()
        last_activated_time = self._triggers_activated_time.get(trigger)
        self._triggers_activated_time[trigger] = activated_time

        if last_activated_time and activated_time - last_activated_time < 0.7:  # FIXME: вынести 0.5 в settings обозвав flow_time_interval
            return False
        return True


class CashedFlowTriggerCheckerLoop(FlowTriggerCheckerLoop):
    """Cash."""

    async def start(self) -> None:
        """Асинхронно сопоставляет принимаемые данные с триггерами в бесконечном цикле.

        Если триггер сопоставлен с полученными данными то он считается сработавшим и
        помещается в очередь сработавших триггеров для дальнейшей обработки.

        Проверяется что сработавший триггер не принадлежит потоку данных
        на котором уже была сработка до этого.

        Дополнительно кешируется результат.
        BUG: На данный момент не работает из за того что арт нет пакеты всегда уникальны. Нужно разобраться почему они уникальны.
        """
        self._triggers_activated_time: dict[ITrigger[ISocketParams], float] = {}
        _cash: dict[int, ITrigger[Any]] = {}

        while True:
            data, listener = await self._checking_queue.get()

            input_hash: int = hash((data, listener))

            trigger_from_cashed_result = _cash.get(input_hash)

            if trigger_from_cashed_result:
                if self._check_flow(trigger_from_cashed_result):
                    await self._activated_queue.put(trigger_from_cashed_result)
            else:
                triggers = self._listener_to_triggers_dict[listener]

                for trigger in triggers:
                    check_command = self._checks_fabric[trigger.tr_type]
                    is_activated: bool = check_command(trigger, *data)
                    if is_activated:
                        if self._check_flow(trigger):
                            _cash[input_hash] = trigger
                            logger.info('Trigger %s activated.', trigger.name)
                            await self._activated_queue.put(trigger)

