"""Base module for system setting command."""
from abc import ABC, abstractmethod
from asyncio.log import logger
from typing import Any, Protocol, Type, TypedDict


class CommandInfoDict(TypedDict):
    """Command information.

    command: str    - command name
    data: Any       - data for command.
    """

    command: str
    data: Any


class Command(ABC):
    """Abstract class for command pattern."""

    @abstractmethod
    def execute(self) -> Any:
        """Contain all logic.

        The only one public method.
        """
        pass


class ICommandToLogic(Protocol):
    """Interface for command to logic.

    Args:
        payload (Any): msg payload

    Methods:
        execute: Run command. Return Any.
    """

    payload: Any

    def execute(self) -> Any:
        """Run command."""


class IAsyncCommandToLogic(Protocol):
    """Async Interface for command to logic.

    Args:
        payload (Any): msg payload

    Methods:
        execute: Run command. Return Any.
    """

    payload: Any

    async def execute(self) -> Any:
        """Async run command."""


class IAsyncRouterCommand(Protocol):
    """Async Interface for command to logic.

    Methods:
        execute: Run command. Return Any.
    """

    async def execute(self, payload: bytes) -> Any:
        """Async run command."""


class CommandRunner():
    """Execute commands.

    Args:
        commands_fabric (dict[str, Type[ICommandToLogic]]): fabric with executed command.

    Methods:
        run_command: execute property commands
    """

    def __init__(self, commands_fabric: dict[str, Type[ICommandToLogic]]) -> None:
        self.commands_fabric = commands_fabric

    def run_command(self, command: CommandInfoDict) -> None:
        """Execute property commands.

        Args:
            command (CommandInfoDict): property commands
        """
        logger.debug('CommandRunner start command "%s"', command['command'])
        command_runnable = self.commands_fabric[command['command']]
        command_runnable(command['data']).execute()
