
"""Publisher commands."""
import asyncio
import queue
from typing import Optional, Protocol
from triggers import settings

from triggers.db.database import AsyncSessionLocal
from triggers.db.utils import (
    get_all_serialized_actions_to_dict,
    get_all_serialized_relations_to_dict,
    get_all_serialized_triggers_to_dict,
)
from triggers.core.msg_handlers import RespMsgDict
from triggers.api.mqtt import send_to_mqtt


class IAsyncPublisherCommand(Protocol):
    """Интерфейс для асинхронной команды публикации mqtt сообщения.

    Args:
        _publisher_queue (queue.Queue[RespMsgDict]): Очередь для публикации mqtt сообщения.
        _publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.

    Methods:
        execute: Отправляет сообщение на публикацию.
    """

    _publisher_queue: queue.Queue[RespMsgDict]
    _publisher_topic: str

    async def execute(self) -> None:
        """Отправляет сообщение на публикацию."""
        ...


class PublishTriggersList(IAsyncPublisherCommand):
    """Команда асинхронно публикующая список всех триггеров.

    Args:
        publisher_queue (queue.Queue[RespMsgDict]): Очередь для публикации mqtt сообщения.
        publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.

    Methods:
        execute: Отправляет сообщение на публикацию.
    """

    def __init__(self, publisher_queue: queue.Queue[RespMsgDict], publisher_topic: str) -> None:
        self._publisher_queue = publisher_queue
        self._publisher_topic = publisher_topic

    async def execute(self) -> None:
        """Отправляет сообщение со списком всех триггеров на публикацию."""
        async with AsyncSessionLocal() as async_session:
            triggers = await get_all_serialized_triggers_to_dict(async_session)
        send_to_mqtt(
            topic=self._publisher_topic,
            payload=triggers,
            retain=True,
            publisher_queue=self._publisher_queue,
        )


class PublishActionList(IAsyncPublisherCommand):
    """Команда асинхронно публикующая список всех actions.

    Args:
        publisher_queue (queue.Queue[RespMsgDict]): Очередь для публикации mqtt сообщения.
        publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.

    Methods:
        execute: Отправляет сообщение на публикацию.
    """

    def __init__(self, publisher_queue: queue.Queue[RespMsgDict], publisher_topic: str) -> None:
        self._publisher_queue = publisher_queue
        self._publisher_topic = publisher_topic

    async def execute(self) -> None:
        """Отправляет сообщение со списком всех actions на публикацию."""
        async with AsyncSessionLocal() as async_session:
            actions = await get_all_serialized_actions_to_dict(async_session)
        send_to_mqtt(
            topic=self._publisher_topic,
            payload=actions,
            retain=True,
            publisher_queue=self._publisher_queue,
        )


class PublishRelationList(IAsyncPublisherCommand):
    """Команда асинхронно публикующая список всех связей между триггерами и actions.

    В списке не отображаются связи с default action.

    Args:
        publisher_queue (queue.Queue[RespMsgDict]): Очередь для публикации mqtt сообщения.
        publisher_topic (str): Mqtt топик в который сообщение будет опубликовано.

    Methods:
        execute: Отправляет сообщение на публикацию.
    """

    def __init__(self, publisher_queue: queue.Queue[RespMsgDict], publisher_topic: str) -> None:
        self._publisher_queue = publisher_queue
        self._publisher_topic = publisher_topic

    async def execute(self) -> None:
        """Отправляет сообщение со списком всех связей между триггерами и actions.."""
        async with AsyncSessionLocal() as async_session:
            all_relations = await get_all_serialized_relations_to_dict(async_session)
            relations_without_default_action = [
                relation for relation in all_relations if relation['action']['name'] != 'default'
            ]
        send_to_mqtt(
            topic=self._publisher_topic,
            payload=relations_without_default_action,
            retain=True,
            publisher_queue=self._publisher_queue,
        )


def send_trigger_list() -> None:
    """Создает асинхронную задачу по отправке сообщения со списком всех триггеров на публикацию.

    Функция зависит от ServiceLocator. Конкретно от заранее созданного
    и помещенного в ServiceLocator класса PublishTriggersList.
    """
    async_publisher_command: Optional[PublishTriggersList] = settings.ServiceLocator.get('publish_trigger_list')
    if async_publisher_command:
        loop = asyncio.get_event_loop_policy().get_event_loop()
        loop.create_task(async_publisher_command.execute(), name='publish_triggers_task')


def send_action_list() -> None:
    """Создает асинхронную задачу по отправке сообщения со списком всех actions на публикацию.

    Функция зависит от ServiceLocator. Конкретно от заранее созданного
    и помещенного в ServiceLocator класса PublishActionList.
    """
    async_publisher_command: Optional[PublishActionList] = settings.ServiceLocator.get('publish_action_list')
    if async_publisher_command:
        loop = asyncio.get_event_loop_policy().get_event_loop()
        loop.create_task(async_publisher_command.execute(), name='publish_action_list_task')


def send_relations_list() -> None:
    """Создает асинхронную задачу по отправке сообщения со списком всех relations на публикацию.

    Функция зависит от ServiceLocator. Конкретно от заранее созданного
    и помещенного в ServiceLocator класса PublishRelationList.
    """
    async_publisher_command: Optional[PublishActionList] = settings.ServiceLocator.get('publish_relation_list')
    if async_publisher_command:
        loop = asyncio.get_event_loop_policy().get_event_loop()
        loop.create_task(async_publisher_command.execute(), name='publish_relation_list_task')
