"""Router command."""
import asyncio
import logging
from typing import Any

from pydantic import ValidationError

from triggers.commands.base import IAsyncRouterCommand
from triggers.core.exceptions import ValidationException
from triggers.db.database import AsyncSessionLocal
from triggers.db.schema import ActionSchema, SetTriggerToActionSchema, TriggerSchema
from triggers.db.utils import (
    create_action_db_obj,
    create_trigger_db_obj,
    del_trigger_to_action_relation,
    delete_action_db_obj,
    delete_trigger_db_obj,
    get_serialized_action_by_name,
    set_trigger_to_action_relation,
)
from triggers.listeners.base import ITrigger


logger = logging.getLogger(__name__)


class AddNewTriggerCommand(IAsyncRouterCommand):
    """Команда добавляющая триггер по запросу от mqtt.

    Args:
        listener_add_queue (asyncio.Queue[ITrigger[Any]]): очередь для создания listeners из триггеров.

    Methods:
        execute: Запускает команду.
    """

    def __init__(self, listener_add_queue: asyncio.Queue[ITrigger[Any]]) -> None:
        self._listener_add_queue = listener_add_queue

    async def execute(self, payload: bytes) -> None:
        """Запускает команду.

        Args:
            payload (bytes): mqtt payload
        """
        self._parse_input(payload)
        await self._create_db_object()
        await self._set_relation_to_default_action()
        await self._create_listener()

    def _parse_input(self, payload: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._trigger = TriggerSchema.parse_raw(payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

    async def _create_db_object(self) -> None:
        logger.debug('Saving new trigger %s to db', self._trigger.name)
        async with AsyncSessionLocal() as async_session:
            await create_trigger_db_obj(async_session, self._trigger)
        logger.debug('The new trigger %s was written into db.', self._trigger.name)

    async def _set_relation_to_default_action(self) -> None:
        logger.debug('Adding the relation between the new trigger %s and default action', self._trigger.name)

        async with AsyncSessionLocal() as async_session:
            default_action = await get_serialized_action_by_name(async_session, 'default')
            await set_trigger_to_action_relation(
                async_session,
                self._trigger,
                default_action,
            )
        logger.debug('The relation was added.')

    async def _create_listener(self) -> None:
        await self._listener_add_queue.put(self._trigger)


class DeleteTriggerCommand(IAsyncRouterCommand):
    """Команда удаляющая триггер по запросу от mqtt.

    Args:
        listener_del_queue (asyncio.Queue[ITrigger[Any]]): очередь для удаления listeners.

    Methods:
        execute: Запускает команду.
    """

    def __init__(self, listener_del_queue: asyncio.Queue[ITrigger[Any]]) -> None:
        self._listener_del_queue = listener_del_queue

    async def execute(self, payload: bytes) -> None:
        """Запускает команду.

        Args:
            payload (bytes): mqtt payload
        """
        self._parse_input(payload)
        await self._delete_db_object()
        await self._delete_listener()

    def _parse_input(self, payload: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._trigger = TriggerSchema.parse_raw(payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

    async def _delete_db_object(self) -> None:
        logger.debug('Deleting the trigger %s from the db', self._trigger.name)
        async with AsyncSessionLocal() as async_session:
            await delete_trigger_db_obj(async_session, self._trigger)
        logger.debug('The trigger %s deleted from the db.', self._trigger.name)

    async def _delete_listener(self) -> None:
        await self._listener_del_queue.put(self._trigger)


class AddNewActionCommand(IAsyncRouterCommand):
    """Команда добавляющая action по запросу от mqtt.

    Methods:
        execute: Запускает команду.
    """

    async def execute(self, payload: bytes) -> None:
        """Запускает команду.

        Args:
            payload (bytes): mqtt payload
        """
        self._parse_input(payload)
        await self._create_db_object()

    def _parse_input(self, payload: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._action = ActionSchema.parse_raw(payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

    async def _create_db_object(self) -> None:
        logger.debug('Saving the new action %s to db', self._action.name)
        async with AsyncSessionLocal() as async_session:
            await create_action_db_obj(async_session, self._action)
        logger.debug('The new action %s was written into db.', self._action.name)


class DeleteActionCommand(IAsyncRouterCommand):
    """Команда удаляющая action по запросу от mqtt.

    Methods:
        execute: Запускает команду.
    """

    async def execute(self, payload: bytes) -> None:
        """Запускает команду.

        Args:
            payload (bytes): mqtt payload
        """
        self._parse_input(payload)
        await self._delete_db_object()

    def _parse_input(self, payload: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._action = ActionSchema.parse_raw(payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

    async def _delete_db_object(self) -> None:
        logger.debug('Deleting the action %s from db', self._action.name)
        async with AsyncSessionLocal() as async_session:
            await delete_action_db_obj(async_session, self._action)
        logger.debug('The action %s was deleted db.', self._action.name)


class CreateTriggerToActionRelationCommand(IAsyncRouterCommand):
    """Команда создающая связь между триггером и action по запросу от mqtt.

    Methods:
        execute: Запускает команду.
    """

    async def execute(self, payload: bytes) -> None:
        """Запускает команду.

        Args:
            payload (bytes): mqtt payload
        """
        self._parse_input(payload)
        await self._set_trigger_to_action_relation()

    def _parse_input(self, payload: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._trigger_to_action = SetTriggerToActionSchema.parse_raw(payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

    async def _set_trigger_to_action_relation(self) -> None:
        logger.debug(
            'Setting trigger %s to action %s',
            self._trigger_to_action.trigger.name,
            self._trigger_to_action.action.name,
        )
        async with AsyncSessionLocal() as async_session:
            await set_trigger_to_action_relation(
                async_session,
                self._trigger_to_action.trigger,
                self._trigger_to_action.action,
            )
        logger.debug(
            'The Trigger %s was set to action %s',
            self._trigger_to_action.trigger.name,
            self._trigger_to_action.action.name,
        )


class DeleteTriggerToActionRelationCommand(IAsyncRouterCommand):
    """Команда удаляющая связь между триггером и action по запросу от mqtt.

    Methods:
        execute: Запускает команду.
    """

    async def execute(self, payload: bytes) -> None:
        """Запускает команду.

        Args:
            payload (bytes): mqtt payload
        """
        self._parse_input(payload)
        await self._del_trigger_to_action_relation()

    def _parse_input(self, payload: bytes) -> None:
        logger.debug('Validating input data.')
        try:
            self._trigger_to_action = SetTriggerToActionSchema.parse_raw(payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

    async def _del_trigger_to_action_relation(self) -> None:
        logger.debug(
            'Setting trigger %s to action %s',
            self._trigger_to_action.trigger.name,
            self._trigger_to_action.action.name,
        )

        async with AsyncSessionLocal() as async_session:
            await del_trigger_to_action_relation(
                async_session, self._trigger_to_action.trigger, self._trigger_to_action.action,
            )

        logger.debug(
            'The Trigger %s was set to action %s',
            self._trigger_to_action.trigger.name,
            self._trigger_to_action.action.name,
        )
