"""Модуль содержит обработчики mqtt команд."""
import asyncio
from typing import Protocol

from triggers.commands.base import IAsyncCommandToLogic, IAsyncRouterCommand, ICommandToLogic


class IAsyncCommandHandler(Protocol):
    """Интерфейс обработчика команд полученных из очереди в бесконечном цикле.

    Args:
        command_queue (asyncio.Queue[ICommandToLogic]): Очередь для получения команд.

    Methods:
        start: Запускает цикл обработки.
    """

    _command_queue: asyncio.Queue[IAsyncCommandToLogic]

    async def start(self) -> None:
        """Start async handle command in infinity loop."""
        ...


class IAsyncRouterCommandHandler(Protocol):
    """Обработчик запускающий команды полученные из очереди в бесконечном цикле.

    Args:
        command_queue (asyncio.Queue[ICommandToLogic]): Очередь для получения команд.

    Methods:
        start: Запускает цикл обработки.
    """

    _command_queue: asyncio.Queue[tuple[IAsyncRouterCommand, bytes]]

    async def start(self) -> None:
        """Start async handle command in infinity loop."""
        ...


class AsyncNaiveCommandHandler(IAsyncCommandHandler):
    """Обработчик запускающий команды полученные из очереди в бесконечном цикле.

    Args:
        command_queue (asyncio.Queue[ICommandToLogic]): Очередь для получения команд.

    Methods:
        start: Запускает цикл обработки.
    """

    def __init__(self, command_queue: asyncio.Queue[ICommandToLogic]) -> None:
        self._command_queue = command_queue

    async def start(self) -> None:
        """Start async handle command in infinity loop."""
        while True:
            command = await self._command_queue.get()
            await command.execute()


class AsyncRouterNaiveCommandHandler(IAsyncRouterCommandHandler):
    """Обработчик запускающий команды полученные из очереди в бесконечном цикле.

    Args:
        command_queue (asyncio.Queue[tuple[IAsyncRouterCommand, bytes]]): Очередь для получения команд.

    Methods:
        start: Запускает цикл обработки.
    """

    def __init__(self, command_queue: asyncio.Queue[tuple[IAsyncRouterCommand, bytes]]) -> None:
        self._command_queue = command_queue

    async def start(self) -> None:
        """Start async handle command in infinity loop."""
        while True:
            command, payload = await self._command_queue.get()
            await command.execute(payload)
