"""Main loop."""
import asyncio
import logging
from queue import Queue
from typing import Protocol

from paho.mqtt.client import MQTTMessage

from triggers.core.msg_handlers import IAsyncMsgHandler, IAsyncRouterMsgHandler, IMsgHandler


logger = logging.getLogger(__name__)


class IMainLoop(Protocol):
    """Interface for Main loop.

    Args:
        msg_queue (Queue[MQTTMessage]): que for receive message.
        msg_handler (IMsgHandler): Class for proceed message.
    """

    msg_queue: Queue[MQTTMessage]
    msg_handler: IMsgHandler

    def start(self) -> None:
        """Start main loop."""


class IAsyncMainLoop(Protocol):
    """Async Interface for Main loop.

    Args:
        msg_queue (asyncio.Queue[MQTTMessage]): que for receive message.
        msg_handler (IAsyncMsgHandler): Class for proceed message.
    """

    msg_queue: asyncio.Queue[MQTTMessage]
    msg_handler: IAsyncMsgHandler

    async def start(self) -> None:
        """Start main loop."""


class MainLoop():
    """Main loop.

    Receive messages from mqtt queue and sends for further processing.

    Args:
        msg_queue (Queue[MQTTMessage]): que for receive message.
        msg_handler (IMsgHandler): Class for proceed message.
    """

    def __init__(self, msg_queue: Queue[MQTTMessage], msg_handler: IMsgHandler) -> None:
        self.msg_queue = msg_queue
        self._msg_handler = msg_handler

    def start(self) -> None:
        """Start main loop."""
        logger.debug('Logic Main Loop has been started.')
        while True:
            msg = self.msg_queue.get()
            self._msg_handler.handle_msg(msg=msg)


class AsyncMainLoop():
    """Async Main loop.

    Receive messages from mqtt queue and sends for further processing.

    Args:
        msg_queue (asyncio.Queue[MQTTMessage]): queue for receive message.
        msg_handler (IAsyncMsgHandler): Class for proceed message.
    """

    def __init__(
        self,
        msg_queue: asyncio.Queue[MQTTMessage],
        msg_handler: IAsyncRouterMsgHandler,
    ) -> None:
        self.msg_queue = msg_queue
        self.msg_handler = msg_handler

    async def start(self) -> None:
        """Start async main loop."""
        logger.debug('Logic Main Loop has been started.')
        while True:
            msg = await self.msg_queue.get()
            await self.msg_handler.handle_msg(msg=msg)
