"""Msg handlers."""
import asyncio
import logging
from queue import Queue
from typing import Any, Callable, Optional, Protocol, Type, TypedDict

from paho.mqtt.client import MQTTMessage

from triggers.commands.base import CommandInfoDict, IAsyncCommandToLogic, IAsyncRouterCommand, ICommandToLogic
from triggers.core.exceptions import Error


logger = logging.getLogger(__name__)


class RespMsgDict(TypedDict):
    """Response msg dict.

    topic: str                  - response topic.
    cor_data: Optional[str]     - uniq request id.
    payload: Any                - json serializable payload.
    """

    topic: str
    cor_data: Optional[str]
    payload: Any
    retain: bool


class IMsgHandler(Protocol):
    """Interface for message handlers.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    topic_fabric: dict[str, Type[ICommandToLogic]]

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        ...


class IAsyncMsgHandler(Protocol):
    """Interface for async message handlers.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    _topic_fabric: dict[str, IAsyncCommandToLogic]

    async def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        ...


class IAsyncRouterMsgHandler(Protocol):
    """Интерфейс для асинхронного обработчика сообщений от mqtt.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    _topic_fabric: dict[str, IAsyncRouterCommand]

    async def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        ...


class NaiveMsgHandler():
    """Naive message handlers.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(self, topic_fabric: dict[str, Type[ICommandToLogic]]) -> None:
        self.topic_fabric = topic_fabric

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        if command := self.topic_fabric.get(msg['topic']):
            command(msg['payload']).execute()


class AsyncQueueMsgHandler():
    """Naive async message handlers.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(
        self,
        topic_fabric: dict[str, Type[ICommandToLogic]],
        command_queue: asyncio.Queue[ICommandToLogic],
    ) -> None:
        self._topic_fabric = topic_fabric
        self._command_queue = command_queue

    async def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        if command_class := self._topic_fabric.get(msg.topic):
            command = command_class(msg.get('payload'))
            await self._command_queue.put(command)


class AsyncRouterQueueMsgHandler():
    """Асинхронный обработчик сообщений от mqtt.

    По mqtt топику получает команду обработки из фабрики команд.
    Затем складывает ее и mqtt payload в очередь для дальнейшей обработки.

    Args:
        topic_fabric (dict[str, IAsyncRouterCommand]): _description_
        command_queue (asyncio.Queue[tuple[IAsyncRouterCommand, bytes]]): _description_
    """

    def __init__(
        self,
        topic_fabric: dict[str, IAsyncRouterCommand],
        command_queue: asyncio.Queue[tuple[IAsyncRouterCommand, bytes]],
    ) -> None:
        self._topic_fabric = topic_fabric
        self._command_queue = command_queue

    async def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        if command := self._topic_fabric.get(msg.topic):
            await self._command_queue.put((command, msg.payload))


class NaiveMsgHandlerWithErrorTopic():
    """Naive message handlers with error topics.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.
    Handle error. if error occurred then error message puts to the error topic.

    Handles only custom Error exception.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(self, topic_fabric: dict[str, Callable[[CommandInfoDict], None]], error_queue: Queue[Any]) -> None:
        self.topic_fabric = topic_fabric
        self.error_queue = error_queue

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        logger.debug('Start handle msg %s', msg)
        if command := self.topic_fabric.get(msg['topic']):
            try:
                command(msg.get('payload'))
            except Error as err:
                logger.warning(err.message)

                payload: dict[str, Any] = {
                    'topic': 'error',
                    'msg': err.message,
                    'data': msg.get('payload'),
                }
                self.error_queue.put(payload)


class MsgHandlerWithResponse():
    """Naive message handlers with error topics.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.
    Handle error. if error occurred then error message puts to the error topic.

    Handles only custom Error exception.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(
        self, topic_fabric: dict[str, Type[ICommandToLogic]],
        resp_queue: Queue[RespMsgDict], service_topics: dict[str, str],
    ) -> None:
        self.topic_fabric = topic_fabric
        self.resp_queue = resp_queue
        self.error_topic = service_topics['error']

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        logger.debug('Start handle msg %s', msg)
        cor_data: Optional[str] = getattr(msg.properties, 'CorrelationData', None)
        resp_topic: str = getattr(msg.properties, 'ResponseTopic', None) or f"{msg.topic}/response"
        if command := self.topic_fabric.get(msg.topic):
            try:
                resp_payload: Any = command(msg.payload).execute()
            except Error as err:
                logger.warning(err.message)
                resp_topic = self.error_topic
                resp_payload = {
                    'msg': err.message,
                    'data': err.data,
                }

            if resp_payload:
                resp_msg: RespMsgDict = {
                    'topic': resp_topic,
                    'cor_data': cor_data,
                    'payload': resp_payload,
                    'retain': False,
                }
                self.resp_queue.put(resp_msg)


class AsyncRouterMsgHandlerWithResponse():
    """Async message handlers with error topics and response.

    По mqtt топику получает команду обработки из фабрики команд.
    Запускает ее передав mqtt payload. Если команда вернула результат
    то он отправляется в очередь отправки для публикации в mqtt.

    Если возникло исключение Error то помещает сообщение об ошибке в
    очередь на отправку для публикации в mqtt error_topic.

    Args:
        topic_fabric (dict[str, IAsyncRouterCommand]): Фабрика содержащая
        название топиков и команды которые должны их обрабатывать.
        resp_queue (Queue[RespMsgDict]): очередь для отправки сообщений
        на публикацию в mqtt.
        service_topics (dict[str, str]): Словарь содержащий сервисные mqtt топики.
    """

    def __init__(
        self,
        topic_fabric: dict[str, IAsyncRouterCommand],
        resp_queue: Queue[RespMsgDict],
        service_topics: dict[str, str],
    ) -> None:
        self._topic_fabric = topic_fabric
        self._resp_queue = resp_queue
        self._error_topic = service_topics['error']

    async def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        logger.debug('Start handle msg %s', msg)
        cor_data: Optional[str] = getattr(msg.properties, 'CorrelationData', None)
        resp_topic: str = getattr(msg.properties, 'ResponseTopic', None) or f"{msg.topic}/response"
        if command := self._topic_fabric.get(msg.topic):
            try:
                resp_payload: Any = await command.execute(msg.payload)
            except Error as err:
                logger.warning(err.message)
                resp_topic = self._error_topic
                resp_payload = {
                    'msg': err.message,
                    'data': err.data,
                }

            if resp_payload:
                resp_msg: RespMsgDict = {
                    'topic': resp_topic,
                    'cor_data': cor_data,
                    'payload': resp_payload,
                    'retain': False,
                }
                self._resp_queue.put(resp_msg)
