"""Models."""
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from triggers.db.database import Base


class SpTriggerType(Base):
    """Справочник содержащий все доступные типы триггеров."""

    __tablename__ = 'sp_trigger_type'

    id = Column(Integer, primary_key=True, unique=True)
    name = Column(String(50), nullable=False, unique=True)

    sp_params = relationship('SpParam', secondary='sp_trigger_param', lazy='subquery')


class SpActionType(Base):
    """Справочник содержащий все доступные типы действий."""

    __tablename__ = 'sp_action_type'

    id = Column(Integer, primary_key=True, unique=True)
    name = Column(String(50), nullable=False, unique=True)

    params = relationship('SpParam', secondary='sp_action_param', lazy='subquery')


class SpParam(Base):
    """Справочник содержащий все доступные параметры действий и триггеров."""

    __tablename__ = 'sp_param'

    id = Column(Integer, primary_key=True, unique=True)
    name = Column(String(50), nullable=False, unique=True)
    type = Column(String, nullable=False, unique=False)


class SpTriggerParam(Base):
    """Справочник содержащий сопоставление типа триггера и его параметра."""

    __tablename__ = 'sp_trigger_param'

    id = Column(Integer, primary_key=True, unique=True)
    id_sp_trigger_type = Column(Integer, ForeignKey('sp_trigger_type.id'))
    id_sp_param = Column(Integer, ForeignKey('sp_param.id'))


class SpActionParam(Base):
    """Справочник содержащий сопоставление типа действия и его параметра."""

    __tablename__ = 'sp_action_param'

    id = Column(Integer, primary_key=True, unique=True)
    id_sp_action_type = Column(Integer, ForeignKey('sp_action_type.id'))
    id_sp_param = Column(Integer, ForeignKey('sp_param.id'))


class Trigger(Base):
    """Содержит список всех триггеров."""

    __tablename__ = 'trigger'

    id = Column(Integer, primary_key=True, unique=True)
    name = Column(String(50), nullable=False, unique=True)
    id_sp_tr_type = Column(Integer, ForeignKey('sp_trigger_type.id'))

    tr_type = relationship('SpTriggerType', lazy='subquery',  viewonly=True)
    params = relationship('TriggerParam', lazy='subquery', uselist=True)
    actions = relationship('Action', secondary='trigger_action', back_populates="triggers", lazy='subquery')


class Action(Base):
    """Содержит список всех действий."""

    __tablename__ = 'action'

    id = Column(Integer, primary_key=True, unique=True)
    name = Column(String(50), nullable=False, unique=True)
    id_sp_a_type = Column(Integer, ForeignKey('sp_action_type.id'))

    ac_type = relationship('SpActionType', lazy='subquery',  viewonly=True)
    params = relationship('ActionParam', lazy='subquery', uselist=True)
    triggers = relationship('Trigger', secondary='trigger_action', back_populates="actions")


class TriggerParam(Base):
    """Содержит список всех параметров триггеров."""

    __tablename__ = 'trigger_param'

    id = Column(Integer, primary_key=True, unique=True)
    id_trigger = Column(Integer, ForeignKey('trigger.id', ondelete='CASCADE'))
    id_sp_tr_param = Column(Integer, ForeignKey('sp_trigger_param.id', ondelete='CASCADE'))
    value = Column(String, nullable=False)

    sp_param = relationship('SpParam', secondary='sp_trigger_param', uselist=False, lazy='subquery',  viewonly=True)
    name = association_proxy('sp_param', 'name')


class ActionParam(Base):
    """Содержит список всех параметров действий."""

    __tablename__ = 'action_param'

    id = Column(Integer, primary_key=True, unique=True)
    id_action = Column(Integer, ForeignKey('action.id', ondelete='CASCADE'))
    id_sp_a_param = Column(Integer, ForeignKey('sp_action_param.id', ondelete='CASCADE'))
    value = Column(String, nullable=False)

    sp_param = relationship('SpParam', secondary='sp_action_param', uselist=False, lazy='subquery',  viewonly=True)
    name = association_proxy('sp_param', 'name')


class TriggerAction(Base):
    """Содержит связи между конкретным триггером и конкретным действием."""

    __tablename__ = 'trigger_action'

    id = Column(Integer, primary_key=True, unique=True)
    id_trigger = Column(Integer, ForeignKey('trigger.id', ondelete='CASCADE'))
    id_action = Column(Integer, ForeignKey('action.id', ondelete='CASCADE'))

    trigger = relationship('Trigger', lazy='subquery',  viewonly=True)
    action = relationship('Action', lazy='subquery',  viewonly=True)
