"""Pydantic Schemas."""
import ipaddress
import json
from typing import Literal, TypeVar, Union
from pydantic import BaseModel, validator


T = TypeVar("T")

TYPE_MAPPER: dict[str, type] = {
    # FIXME: TYPE_MAPPER не используется как и столбец Type в таблице sp_param.
    'str': str,
    'bool': bool,
    'int': int,
    'float': float,
    'bytes': bytes,
    'dict': dict,
}


def validate_ipv4_address(value: str) -> str:
    """Валидирует IPv4 адрес.

    Выбрасывает исключение если строка не является валидным IPv4 адресом.

    Args:
        value (str): строка с предполагаемым IPv4 адресом.

    Returns:
        str: строка без изменений.
    """
    ipaddress.IPv4Address(value)
    return value


class RawUdpTriggerParamsSchema(BaseModel):
    """Raw udp trigger settings.

    Дополнительно валидирует ip адрес.

    network_type (Literal['udp']):  Тип сети. Всегда udp.
    listen_ip (str):                Прослушиваемы ip.
    listen_port (int):              Прослушиваемый порт.
    data (str):                     Полезная нагрузка.
    """

    network_type: Literal['udp'] = 'udp'
    listen_ip: str
    listen_port: int
    data: str

    class Config:
        """Schema config."""
        orm_mode = True

    _ensure_listen_ip_is_valid_ipv4 = validator('listen_ip', allow_reuse=True)(validate_ipv4_address)


class ArtNetTriggerParamsSchema(BaseModel):
    """ArtNet trigger settings.

    Дополнительно валидирует ip адрес.

    network_type (Literal['tcp', 'udp']):  Тип сети.
    listen_ip (str):                Прослушиваемы ip.
    listen_port (int):              Прослушиваемый порт.
    channel (int):                  Номер DMX каналла.
    min_level (int):                Максисальное значение в канале.
    max_level (int):                Минимальное значение в канале.
    """

    network_type: Literal['tcp', 'udp']
    listen_ip: str
    listen_port: int
    universe: int
    channel: int
    min_level: int
    max_level: int

    class Config:
        """Schema config."""
        orm_mode = True

    _ensure_listen_ip_is_valid_ipv4 = validator('listen_ip', allow_reuse=True)(validate_ipv4_address)


class TriggerSchema(BaseModel):
    """Настройки триггера.

    Args:
        name:       Имя триггера.
        tr_type:    Тип триггера.
        params:     Параметры триггера.
    """

    name: str
    tr_type: Literal['ArtNet', 'RawUDP']
    params: Union[ArtNetTriggerParamsSchema, RawUdpTriggerParamsSchema]

    class Config:
        """Schema config."""
        orm_mode = True

    def __hash__(self) -> int:
        """Хеширует имя триггера."""
        return hash(self.name)


class SendMqttActionParamsSchema(BaseModel):
    """Action params for sending mqtt msg.

    topic (str): Mqtt topic.
    payload (str): Mqtt payload.
    retain (bool): Mqtt retain param.
    """

    topic: str
    payload: str
    retain: bool

    @validator('payload', pre=True)
    def convert_payload_to_str(cls, value: str) -> str:
        """Конвертирует payload в str.

        Args:
            value (str): payload value.
        """
        if not isinstance(value, str):
            value = json.dumps(value)
        return value


class ActionSchema(BaseModel):
    """Action settings.

    name:           Имя action.
    action_type:    Тип action.
    params:         Параметры action.
    """

    name: str
    action_type: Literal['send_mqtt_msg_raw', 'send_trigger_to_mqtt']
    params: SendMqttActionParamsSchema


class SetTriggerToActionSchema(BaseModel):
    """Настройки отношений триггера и action.

    trigger (TriggerSchema):    Триггер.
    action (ActionSchema):      Action.
    """

    trigger: TriggerSchema
    action: ActionSchema
