"""Utilities for working with db."""
from typing import Optional, Any, Type, TypeVar

from pydantic import BaseModel
import sqlalchemy
from sqlalchemy import select, and_
from sqlalchemy.orm import Session
from sqlalchemy.ext.asyncio import AsyncSession
from triggers.actions.base import IAction

from triggers.core.exceptions import (
    AlreadyExistError,
    NotFoundException,
)
from triggers.db.models import (
    Action,
    ActionParam,
    SpActionParam,
    SpActionType,
    SpParam,
    SpTriggerParam,
    SpTriggerType,
    Trigger,
    TriggerAction,
    TriggerParam,
)
from triggers.db.schema import (
    ActionSchema,
    SetTriggerToActionSchema,
    TriggerSchema,
)
from triggers.listeners.base import ITrigger


T = TypeVar("T")


def get_or_create(session: Session, model: Type[T], **kwargs: Any) -> T:
    """Get or create db instance.

    Args:
        session (Session): Alchemy session.
        model (Base): declarative base model.

    Raises:
        sqlalchemy.orm.exc.MultipleResultsFound: raises if multiple object identities are returned,
            or if multiple rows are returned for a query that returns only scalar values as opposed
            to full identity-mapped entities.

    Returns:
        Base: model instance.
    """
    instance: Optional[T] = session.query(model).filter_by(**kwargs).one_or_none()
    if not instance:
        instance = model(**kwargs)
        session.add(instance)
        session.commit()
    return instance


def get_from_db(session: Session, model: Type[T], **kwargs: Any) -> Optional[T]:
    """Get db instance.

    Args:
        session (Session): Alchemy session.
        model (Base): declarative base model.

    Raises:
        sqlalchemy.orm.exc.MultipleResultsFound: raises if multiple object identities are returned,
            or if multiple rows are returned for a query that returns only scalar values as opposed
            to full identity-mapped entities.

    Returns:
        Base: model instance.
    """
    instance: Optional[T] = session.query(model).filter_by(**kwargs).one_or_none()
    return instance


def update_db_obj(session: Session, model: Type[T], update_item: BaseModel) -> bool:
    """Update object into db.

    Args:
        session (Session):          Alchemy session.
        model (Type[T]):            Alchemy model class.
        update_item (BaseModel):    Pydantic schema instance.

    Returns:
        bool: _description_
    """
    upd_obj = session.query(model).filter_by(id=update_item.id).one_or_none()
    if not upd_obj:
        return False

    for key, value in update_item.dict(exclude=set('id')).items():
        setattr(upd_obj, key, value)

    session.add(upd_obj)
    session.commit()
    session.refresh(upd_obj)
    return True


async def get_trigger_obj_by_schema(async_session: AsyncSession,  trigger_schema: ITrigger[Any]) -> Trigger:
    """Возвращает объект ORM таблицы Trigger по объекту схемы.

    Args:
        async_session (AsyncSession): Асинхронная сессия до БД.
        trigger_schema (ITrigger[Any]): Объект с настройками триггера.

    Raises:
        NotFoundException: Происходит если ни одного объекта не было найдено.

    Returns:
        Trigger: Объект ORM таблицы Trigger.
    """
    get_trigger_query = select(Trigger).where(Trigger.name == trigger_schema.name)
    get_trigger_query_result = await async_session.execute(get_trigger_query)
    try:
        trigger_obj: Trigger = get_trigger_query_result.scalar_one()
    except sqlalchemy.exc.NoResultFound:
        raise NotFoundException(f'The trigger {trigger_schema.name} is not found.')
    return trigger_obj


async def get_action_obj_by_schema(async_session: AsyncSession,  action_schema: IAction[Any]) -> Action:
    """Возвращает объект ORM таблицы Action по объекту схемы.

    Args:
        async_session (AsyncSession): Асинхронная сессия до БД.
        action_schema (IAction[Any]): Объект с настройками action.

    Raises:
        NotFoundException: Происходит если ни одного объекта не было найдено.

    Returns:
        Action: Объект ORM таблицы Action.
    """
    action_obj = await get_action_obj_by_name(async_session, action_schema.name)
    return action_obj


async def get_trigger_to_action_relation_by_schema(
    async_session: AsyncSession,
    trigger_schema: ITrigger[Any],
    action_schema: IAction[Any],
) -> TriggerAction:
    """Возвращает объект ORM таблицы TriggerAction по объекту схемы.

    Args:
        async_session (AsyncSession): Асинхронная сессия до БД.
        trigger_schema (ITrigger[Any]): Объект с настройками триггера.
        action_schema (IAction[Any]): Объект с настройками action.

    Raises:
        NotFoundException: Происходит если ни одного объекта не было найдено.

    Returns:
        TriggerAction: Объект ORM таблицы TriggerAction.
    """
    trigger_obj = await get_trigger_obj_by_schema(async_session, trigger_schema)
    action_obj = await get_action_obj_by_schema(async_session, action_schema)

    get_trigger_to_action_relation = select(TriggerAction).where(
        and_(
            TriggerAction.id_trigger == trigger_obj.id,
            TriggerAction.id_action == action_obj.id,
        ),
    )
    get_trigger_to_action_relation_result = await async_session.execute(get_trigger_to_action_relation)
    try:
        relation_obj: TriggerAction = get_trigger_to_action_relation_result.scalar_one()
    except sqlalchemy.exc.NoResultFound:
        raise NotFoundException(
            f'The relation between the trigger {trigger_schema.name}  and'
            f' the action {action_schema.name} was not found.'
        )
    return relation_obj


async def create_trigger_db_obj(async_session: AsyncSession,  trigger_schema: ITrigger[Any]) -> Trigger:
    """Создает запись в таблице Trigger.

    Args:
        async_session (AsyncSession):   Асинхронная сессия.
        trigger_schema (ITrigger[Any]): Объект с настройками триггера.

    Raises:
        AlreadyExistError: Происходит если объект в бд с таким именем уже существует.

    Returns:
        Trigger: Объект ORM таблицы Trigger.
    """
    query = (
        select(SpTriggerType.id, SpTriggerParam.id, SpParam.name)
        .join(SpTriggerParam, SpTriggerType.id == SpTriggerParam.id_sp_trigger_type)
        .join(SpParam, SpTriggerParam.id_sp_param == SpParam.id)
        .where(SpTriggerType.name == trigger_schema.tr_type)
    )
    query_result = await async_session.execute(query)
    new_trigger_attrs: list[tuple[int, int, str]] = query_result.all()

    # Create new trigger
    sp_tr_type_id: int = new_trigger_attrs[0][0]
    new_trigger: Trigger = Trigger(name=trigger_schema.name, id_sp_tr_type=sp_tr_type_id)
    async_session.add(new_trigger)

    for _, sp_tr_param_id, sp_param_name in new_trigger_attrs:
        trigger_param = TriggerParam(
            id_sp_tr_param=sp_tr_param_id,
            value=getattr(trigger_schema.params, sp_param_name),
        )
        new_trigger.params.append(trigger_param)

    try:
        await async_session.commit()
    except sqlalchemy.exc.IntegrityError:
        await async_session.rollback()
        raise AlreadyExistError(f'The trigger {trigger_schema.name} already exist!')
    await async_session.refresh(new_trigger)

    return new_trigger


async def create_action_db_obj(async_session: AsyncSession,  action_schema: IAction[Any]) -> Action:
    """Создает запись в таблице Action.

    Args:
        async_session (AsyncSession):   Асинхронная сессия.
        action_schema (IAction[Any]): Объект с настройками action.

    Raises:
        AlreadyExistError: Происходит если объект в бд с таким именем уже существует.

    Returns:
        Action: Объект ORM таблицы Action.
    """
    # get expected params
    query = (
        select(SpActionType.id, SpActionParam.id, SpParam.name)
        .join(SpActionParam, SpActionType.id == SpActionParam.id_sp_action_type)
        .join(SpParam, SpActionParam.id_sp_param == SpParam.id)
        .where(SpActionType.name == action_schema.action_type)
    )
    query_result = await async_session.execute(query)
    new_action_attrs: list[tuple[int, int, str]] = query_result.all()

    # Create new action
    id_sp_a_type: int = new_action_attrs[0][0]
    new_action: Action = Action(name=action_schema.name, id_sp_a_type=id_sp_a_type)
    async_session.add(new_action)

    for _, sp_tr_param_id, sp_param_name in new_action_attrs:
        action_param = ActionParam(
            id_sp_a_param=sp_tr_param_id,
            value=getattr(action_schema.params, sp_param_name),
        )
        new_action.params.append(action_param)

    try:
        await async_session.commit()
    except sqlalchemy.exc.IntegrityError:
        await async_session.rollback()
        raise AlreadyExistError(f'The action with name {action_schema.name} is already exist!')
    await async_session.refresh(new_action)

    return new_action


async def delete_trigger_db_obj(async_session: AsyncSession,  trigger_schema: TriggerSchema) -> None:
    """Удаляет запись в базе о триггере по объекту схемы.

    Args:
        async_session:  открытая асинхронная сессия до БД.
        trigger_schema: объект схемы.
    """
    trigger_obj = await get_trigger_obj_by_schema(async_session, trigger_schema)
    await async_session.delete(trigger_obj)
    await async_session.commit()


async def delete_action_db_obj(async_session: AsyncSession,  action_schema: IAction[Any]) -> None:
    """Удаляет запись в базе о событии по объекту схемы.

    Args:
        async_session:  открытая асинхронная сессия до БД.
        action_schema: объект схемы.
    """
    action_obj = await get_action_obj_by_schema(async_session, action_schema)
    await async_session.delete(action_obj)
    await async_session.commit()


async def set_trigger_to_action_relation(
    async_session: AsyncSession,
    trigger_schema: ITrigger[Any],
    action_schema: IAction[Any],
) -> None:
    """Создает привязку между триггером и action.

    Args:
        async_session (AsyncSession):   Асинхронная сессия.
        trigger_schema (ITrigger[Any]): Объект с настройками триггера.
        action_schema (IAction[Any]):   Объект с настройками action.
    """
    trigger_obj = await get_trigger_obj_by_schema(async_session, trigger_schema)
    action_obj = await get_action_obj_by_schema(async_session, action_schema)

    async_session.add(TriggerAction(id_trigger=trigger_obj.id, id_action=action_obj.id))
    await async_session.commit()


async def del_trigger_to_action_relation(
    async_session: AsyncSession,
    trigger_schema: TriggerSchema,
    action_schema: IAction[Any],
) -> None:
    """Удаляет привязку между триггером и action.

    Args:
        async_session (AsyncSession):   Асинхронная сессия.
        trigger_schema (ITrigger[Any]): Объект с настройками триггера.
        action_schema (IAction[Any]):   Объект с настройками action.
    """
    relation_obj = await get_trigger_to_action_relation_by_schema(async_session, trigger_schema, action_schema)
    async_session.delete(relation_obj)
    await async_session.commit()


async def serialize_triggers_to_actions_relations_by_schema(
    async_session: AsyncSession,
) -> list[SetTriggerToActionSchema]:
    """Возвращает все связи между триггером и action.

    Args:
        async_session (AsyncSession): Асинхронная сессия.

    Returns:
        list[SetTriggerToActionSchema]: Список связей между триггерами и actions.
    """
    relations: list[SetTriggerToActionSchema] = []

    get_all_relations_query_result = await async_session.execute(select(TriggerAction))
    relation_objs = get_all_relations_query_result.scalars().all()

    for relation_obj in relation_objs:
        trigger_serialized = serialize_trigger_obj_by_schema(relation_obj.trigger)
        action_serialized = serialize_action_obj_by_schema(relation_obj.action)
        relation_serialized = SetTriggerToActionSchema(trigger=trigger_serialized, action=action_serialized)
        relations.append(relation_serialized)

    return relations


def serialize_action_obj_to_dict(action_obj: Action) -> dict[str, Any]:
    """Сериализует объект ORM таблицы Action в dict.

    Args:
        action_obj (Action): объект ORM таблицы Action.

    Returns:
        dict[str, Any]: сериализованные данные.
    """
    action_serialized = {
        'name': action_obj.name,
        'action_type': action_obj.ac_type.name,
        'params': {param.name: param.value for param in action_obj.params},
    }
    return action_serialized


def serialize_action_obj_by_schema(action_obj: Action) -> ActionSchema:
    """Сериализует объект ORM таблицы Action в ActionSchema.

    Args:
        action_obj (Action): объект ORM таблицы Action.

    Returns:
        ActionSchema: сериализованные данные.
    """
    action_dict = serialize_action_obj_to_dict(action_obj)
    action_serialized = ActionSchema.parse_obj(action_dict)
    return action_serialized


async def get_all_serialized_actions_to_dict(async_session: AsyncSession) -> list[dict[str, Any]]:
    """Получает список всех действий из базы данных и сериализует их в словарь.

    Args:
        session (AsyncSession): Асинхронная сессия.

    Returns:
        list[dict[str, Any]]: Список с сериализованными данными actions.
    """
    actions: list[dict[str, Any]] = []

    get_all_triggers_query_result = await async_session.execute(select(Action))
    action_objs = get_all_triggers_query_result.scalars().all()

    for action_obj in action_objs:
        action_serialized: dict[str, Any] = serialize_action_obj_to_dict(action_obj)
        actions.append(action_serialized)

    return actions


async def get_all_serialized_actions(async_session: AsyncSession) -> list[ActionSchema]:
    """Получает список всех действий из базы данных и сериализует их в объекты ActionSchema.

    Args:
        session (AsyncSession): Асинхронная сессия.

    Returns:
        list[ActionSchema]: Список с сериализованными данными actions.
    """
    actions: list[ActionSchema] = []

    get_all_triggers_query_result = await async_session.execute(select(Action))
    action_objs = get_all_triggers_query_result.scalars().all()

    for action_obj in action_objs:
        action_serialized: ActionSchema = serialize_action_obj_by_schema(action_obj)
        actions.append(action_serialized)

    return actions


def serialize_trigger_obj_to_dict(trigger_obj: Trigger) -> dict[str, Any]:
    """Сериализует объект триггера в словарь.

    Args:
        trigger_obj (Trigger): объект триггера из орм.

    Returns:
        dict[str, Any]: Сериализованный триггер.
    """
    trigger_serialized = {
        'name': trigger_obj.name,
        'tr_type': trigger_obj.tr_type.name,
        'params': {param.name: param.value for param in trigger_obj.params},
    }
    return trigger_serialized


def serialize_trigger_obj_by_schema(trigger_obj: Trigger) -> TriggerSchema:
    """Сериализует объект триггера в TriggerSchema.

    Args:
        trigger (Trigger): объект триггера из орм.

    Returns:
        TriggerSchema: Сериализованный триггер.
    """
    trigger_settings = serialize_trigger_obj_to_dict(trigger_obj)
    trigger_schema = TriggerSchema.parse_obj(trigger_settings)
    return trigger_schema


async def get_all_serialized_triggers_to_dict(async_session: AsyncSession) -> list[dict[str, Any]]:
    """Получает список всех триггеров из базы данных и сериализует их в список словарей.

    Args:
        session (AsyncSession): Асинхронная сессия.

    Returns:
        list[dict[str, Any]]: список словарей с параметрами триггеров.
    """
    triggers: list[dict[str, Any]] = []

    query = select(Trigger)
    trigger_db_result = await async_session.execute(query)
    trigger_objs = trigger_db_result.scalars().all()

    for trigger_obj in trigger_objs:
        trigger = serialize_trigger_obj_to_dict(trigger_obj)
        triggers.append(trigger)

    return triggers


async def get_all_serialized_triggers(async_session: AsyncSession) -> list[TriggerSchema]:
    """Получает список всех триггеров из базы данных и сериализует их.

    Args:
        session (AsyncSession): Асинхронная сессия.

    Returns:
        list[TriggerSchema]: _description_
    """
    triggers: list[TriggerSchema] = []

    query = select(Trigger)
    trigger_db_result = await async_session.execute(query)
    trigger_objs = trigger_db_result.scalars().all()

    for trigger_obj in trigger_objs:
        trigger = serialize_trigger_obj_by_schema(trigger_obj)
        triggers.append(trigger)

    return triggers


async def get_serialized_actions_related_trigger(
    async_session: AsyncSession,
    trigger: ITrigger[Any],
) -> list[ActionSchema]:
    """Получает список всех действий привязанных к триггеру.

    Args:
        session (Session): Alchemy session.

    Returns:
        list[ActionSchema]: Сериализованный список actions.
    """
    trigger_obj = await get_trigger_obj_by_schema(async_session, trigger)

    actions_serialized_list: list[ActionSchema] = []
    for action_obj in trigger_obj.actions:
        action_serialized = serialize_action_obj_by_schema(action_obj)
        actions_serialized_list.append(action_serialized)

    return actions_serialized_list


async def get_all_serialized_relations_to_dict(
    async_session: AsyncSession,
) -> list[dict[str, Any]]:
    """Возвращает все связи между триггером и action серивлизованные в словарь.

    Args:
        async_session (AsyncSession): Асинхронная сессия.

    Returns:
        list[dict[str, Any]]: Список связей между триггерами и actions серивлизованные в словарь.
    """
    relations: list[dict[str, Any]] = []

    get_all_relations_query_result = await async_session.execute(select(TriggerAction))
    relation_objs = get_all_relations_query_result.scalars().all()

    for relation_obj in relation_objs:
        trigger_serialized = serialize_trigger_obj_to_dict(relation_obj.trigger)
        action_serialized = serialize_action_obj_to_dict(relation_obj.action)
        relation_serialized = {'trigger': trigger_serialized, 'action': action_serialized}
        relations.append(relation_serialized)

    return relations


async def get_action_obj_by_name(async_session: AsyncSession, action_name: str) -> Action:
    """Возвращает объект ORM таблицы Action по его имени.

    Args:
        async_session (AsyncSession): Асинхронная сессия до БД.
        action_name (str): Имя action.

    Raises:
        NotFoundException: Происходит если ни одного объекта не было найдено.

    Returns:
        Action: Объект ORM таблицы Action.
    """
    query = select(Action).where(Action.name == action_name)
    query_result = await async_session.execute(statement=query)
    try:
        action_obj: Action = query_result.scalar_one()
    except sqlalchemy.exc.NoResultFound:
        raise NotFoundException(f'The action {action_name} is not found.')
    return action_obj


async def get_serialized_action_by_name(async_session: AsyncSession, action_name: str) -> ActionSchema:
    """Возвращает сериализованный объект таблицы Action по его имени.

    Args:
        async_session (AsyncSession): _description_
        action_name (str): _description_

    Returns:
        ActionSchema: сериализованный объект таблицы Action
    """
    action_obj = await get_action_obj_by_name(async_session, action_name)
    action_serialized = serialize_action_obj_by_schema(action_obj)
    return action_serialized
