"""Модуль содержит асинхронные listeners."""
import asyncio
from dataclasses import dataclass
import socket
from typing import Any, Generic, Iterable, Literal, Optional, Protocol, TypeVar
import logging
from triggers import settings

from triggers.core.exceptions import (
    Error,
    NetworkTypeIsNotSupported,
    NotFoundException,
    SocketError,
)
from triggers.listeners.base import (
    ISocketParams,
    ITransportListenerParams,
    ITrigger,
    TPossibleTriggerTypes,
    IListener,
)
from triggers.checkers.checkers import (
    TCheckingQueue,
)

from triggers.listeners.utils import compare_socket_trigger_and_transport_listener


logger = logging.getLogger(__name__)

TPossibleListenerTypes = Literal['transport']
T = TypeVar('T', covariant=True)


class IListenerDestroyer(Protocol):
    """Interface for listener destroyer class."""

    async def __call__(self, listener: IListener[Any]) -> None:
        """Destroy the listener."""
        ...


class IListenerCreator(Protocol):
    """Interface for listener creator."""

    async def __call__(self, trigger: ITrigger[Any], checking_queue: TCheckingQueue) -> IListener[Any]:
        """Add a new listener from trigger."""
        ...


class IProtocolFabric(Protocol):
    """Interface for asyncio transport protocol fabric."""

    def __call__(self) -> asyncio.BaseProtocol:
        """Создает объект подкласса asyncio.BaseProtocol.

        Класс asyncio.BaseProtocol необходим для дальнейшего
        использования при создании асинхронного endpoint'a.

        Returns:
            asyncio.BaseProtocol
        """
        ...


class ITriggerListenerComparator(Protocol):
    """Интерфейс для функции или вызываемого класса который сравнивает trigger и listener."""

    def __call__(self, trigger: ITrigger[Any], listener: IListener[Any]) -> bool:
        """Сравнивает trigger и listener.

        Args:
            trigger (ITrigger[Any]): Объект триггера.
            listener (IListener[Any]): Объект listener.

        Returns:
            bool
        """
        ...


TListenerCreateFabric = dict[TPossibleListenerTypes, IListenerCreator]
TListenerDestroyFabric = dict[TPossibleListenerTypes, IListenerDestroyer]
TypeProtocolFabric = dict[TPossibleListenerTypes, IProtocolFabric]


TR_TYPE_TO_LISTENER_TYPE_SP: dict[TPossibleTriggerTypes, TPossibleListenerTypes] = {
    # Справочник сопоставляющий тип триггера с типом listener.
    'ArtNet': 'transport',
    'RawUDP': 'transport',
}


@dataclass
class TransportListenerParams():
    """Параметры для transport listener.

    network_type (Literal['tcp', 'udp']): Тип сети
    listen_ip (str): IP адрес для прослушивания.
    listen_port (int): Сетевой порт для прослушивания.
    """

    network_type: Literal['tcp', 'udp']
    listen_ip: str
    listen_port: int

    def __hash__(self) -> int:
        """Хэширует параметры."""
        return hash((self.network_type, self.listen_ip, self.listen_port))


@dataclass
class Listener(Generic[T]):
    """Объект listener содержащий все его параметры.

    listener_type (TPossibleListenerTypes): Тип listener'a.
    checking_queue: Очередь куда помещать получаемые данные для дальнейшей проверки
    close_event (asyncio.Event): Ивент срабатывает когда нужно закрыть listener.
    params: Конкретные параметры зависящие от типа listener.
    """

    listener_type: TPossibleListenerTypes
    checking_queue: TCheckingQueue
    close_event: asyncio.Event
    params: T

    def __hash__(self) -> int:
        """Хэширует параметры."""
        return hash(self.params)


TransportListener = Listener[TransportListenerParams]


class DatagramProtocolFabric:
    """Класс создает экземпляр DatagramServerProtocol с предустановленными настройками.

    Класс DatagramServerProtocol нужен для использования при создании асинхронного endpoint'a

    Args:
        listener: Объект listener привязанный к конкретному объекту протокола.
    """

    def __init__(self, listener: IListener[ITransportListenerParams]) -> None:
        self.listener = listener

    def __call__(self) -> Any:
        """Создает экземпляр InetServerProtocol."""
        return DatagramServerProtocol(self.listener)


class DatagramServerProtocol(asyncio.DatagramProtocol):
    """Протокол вызываемый асинхронным транспортом.

    Методы протокола вызываются из асинхронного транспорта.

    Args:
        listener (IListener[ITransportListenerParams]): Объект listener закрепленный за данным протоколом.
    """

    def __init__(self, listener: IListener[ITransportListenerParams]) -> None:
        self.listener = listener
        self.checking_queue = self.listener.checking_queue

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Срабатывает когда транспорт установил соединение."""
        self.transport = transport
        self.soc = self.transport.get_extra_info('socket')

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        """Срабатывает когда транспорт получил данные.

        Помещает полученные данные в очередь для дальнейших проверок.

        Args:
            data (bytes): Полученные данные
            addr (tuple[str, int]): IP адрес и порт удаленной машины от которого были получены данные.
        """
        asyncio.create_task(self.checking_queue.put(((self.soc, data), self.listener)))

    def connection_lost(self, exception: Any) -> None:
        """Срабатывает когда транспорт закрывается."""
        logger.debug('Transport was closed.')


async def transport_closer(transport: asyncio.BaseTransport, event: asyncio.Event) -> None:
    """Закрывает транспорт при срабатывании Ивента.

    Args:
        transport (asyncio.BaseTransport): Объект транспорта который нужно закрыть.
        event (asyncio.Event): Ивент при срабатывании которого транспорт закроется.
    """
    await event.wait()
    if type(transport) is asyncio.DatagramTransport:
        transport.abort()
    else:
        transport.close()


class TransportListenerCreator:
    """Класс для создания Listener."""

    async def __call__(
        self,
        trigger: ITrigger[ISocketParams],
        checking_queue: TCheckingQueue,
    ) -> IListener[ITransportListenerParams]:
        """Создает listener из триггера.

        Args:
            trigger (ITrigger[ISocketParams]): Объект триггера.
            checking_queue (TCheckingQueue): Очередь куда помешать полученные
            listener'ом данные для дальнейших проверок

        Returns:
            IListener[ITransportListenerParams]
        """
        event = asyncio.Event()

        listener: IListener[ITransportListenerParams] = TransportListener(
            listener_type='transport',
            checking_queue=checking_queue,
            close_event=event,
            params=TransportListenerParams(
                network_type=trigger.params.network_type,
                listen_ip=trigger.params.listen_ip,
                listen_port=trigger.params.listen_port if trigger.params.listen_port != 6454 else settings.ARTNET_MASQUERADE_DEFAULT_PORT,
            ),
        )

        transport, _ = await _create_new_transport(listener)
        # Создает задачу которая закроет транспорт при срабатывании Ивента.
        asyncio.create_task(transport_closer(transport, event))

        return listener


async def _create_new_transport(
    listener: IListener[ITransportListenerParams],
) -> tuple[asyncio.BaseTransport, asyncio.BaseProtocol]:
    """Создает новый асинхронный транспорт.

    Args:
        listener: IListener[ITransportListenerParams]: Объект с параметрами для создания транспорта.

    Raises:
        NetworkTypeIsNotSupported: Выбрасывается если тип сети не поддерживается.

    Returns:
        tuple[asyncio.BaseTransport, asyncio.BaseProtocol]
    """
    loop = asyncio.get_running_loop()
    if listener.params.network_type == 'udp':
        try:
            transport, protocol = await loop.create_datagram_endpoint(
                DatagramProtocolFabric(listener),
                local_addr=(listener.params.listen_ip, listener.params.listen_port),
            )
        except socket.error as err:
            raise SocketError(
                err.strerror,
                (listener.params.listen_ip, listener.params.listen_port),
            )
    else:
        raise NetworkTypeIsNotSupported(data=listener.params.network_type)

    return (transport, protocol)


class ListenerCreateLoop:
    """Добавляет новые listener в бесконечном цикле.

    Получает новые триггеры по очереди queue_trigger_add и обрабатывает их.

    Args:
        trigger_to_listener_sp:     Справочник содержащий сопоставления типов триггеров с типами listener.
        create_listener_fabric:     Фабрика по созданию listener'ов.
        queue_with_trigger:         Очередь для получения новых триггеров.
        checking_queue:             Очередь в которую помещаются полученные listener данные для дальнейших проверок.
        listener_to_triggers_dict:  Словарь оперативно заполняющийся данными сопоставление объектов триггеров с
        объектами listener.
    """

    def __init__(
        self,
        trigger_to_listener_sp: dict[TPossibleTriggerTypes, TPossibleListenerTypes],
        create_listener_fabric: TListenerCreateFabric,
        queue_with_trigger: asyncio.Queue[ITrigger[ISocketParams]],
        checking_queue: TCheckingQueue,
        listener_to_triggers_dict: dict[IListener[Any], set[ITrigger[Any]]],
    ) -> None:
        self._trigger_to_listener_sp = trigger_to_listener_sp
        self._listener_to_triggers_dict = listener_to_triggers_dict
        self._create_listener_fabric = create_listener_fabric
        self._queue_with_trigger = queue_with_trigger
        self._checking_queue = checking_queue

    async def start(self) -> None:
        """Асинхронно добавляет новые listener в бесконечном цикле."""
        while True:
            trigger = await self._queue_with_trigger.get()
            logger.debug('Trying to create a new transport from trigger %s', trigger.name)
            if existing_listener := get_listener_by_trigger(trigger, self._listener_to_triggers_dict):
                self._listener_to_triggers_dict[existing_listener].add(trigger)
                logger.debug('Related transport has already exist. Related transport: %s', existing_listener.params)
            else:
                listener_type: TPossibleListenerTypes = self._trigger_to_listener_sp[trigger.tr_type]
                create_listener = self._create_listener_fabric[listener_type]
                try:
                    listener: IListener[Any] = await create_listener(trigger, self._checking_queue)
                except Error as err:
                    logger.error('Transport create error. %s. Data: %s', err.message, err.data)
                else:
                    self._listener_to_triggers_dict[listener] = {trigger}
                    logger.debug('The new transport created. %s', listener.params)


class ListenerDestroyLoop:
    """Закрывает listener при удалении триггера в бесконечном цикле.

    Получает новые триггеры по очереди queue_trigger_del и закрывает listener
    если не осталось связанных с ним триггеров.

    Args:
        queue_with_trigger:         Очередь для получения новых триггеров.
        listener_to_triggers_dict:  Словарь оперативно заполняющийся данными сопоставление объектов триггеров с
        объектами listener.
    """
    def __init__(
        self,
        queue_with_trigger: asyncio.Queue[ITrigger[ISocketParams]],
        listener_to_triggers_dict: dict[IListener[Any], set[ITrigger[Any]]],
    ) -> None:
        self._queue_with_trigger = queue_with_trigger
        self._listener_to_triggers_dict = listener_to_triggers_dict

    async def start(self) -> None:
        """Асинхронно закрывает listener при удалении триггера в бесконечном цикле."""
        while True:
            trigger = await self._queue_with_trigger.get()

            logger.debug('Deleting trigger %s', trigger.name)
            existing_listener = get_listener_by_trigger(trigger, self._listener_to_triggers_dict)
            if not existing_listener:
                raise NotFoundException()
            self._listener_to_triggers_dict[existing_listener].discard(trigger)

            if not self._listener_to_triggers_dict[existing_listener]:
                existing_listener.close_event.set()
                self._listener_to_triggers_dict.pop(existing_listener)
                logger.debug('Transport destroyed.')
            logger.debug('Trigger deleted.')


def get_listener_by_trigger(trigger: ITrigger[Any], listeners: Iterable[IListener[Any]]) -> Optional[IListener[Any]]:
    """Ищет триггер по параметрам в списке listeners.

    Args:
        trigger (ITrigger[Any]): Объект триггера с параметрами для поиска.
        listeners (Iterable[IListener[Any]]): список listeners.

    Returns:
        Optional[IListener[Any]]
    """
    listener_compare_fabric: dict[TPossibleListenerTypes, ITriggerListenerComparator] = {
        'transport': compare_socket_trigger_and_transport_listener,
    }

    for listener in listeners:
        if listener.listener_type == TR_TYPE_TO_LISTENER_TYPE_SP[trigger.tr_type]:
            compare = listener_compare_fabric[listener.listener_type]
            if compare(trigger, listener):
                return listener
    return None


listener_to_triggers_dict: dict[IListener[Any], set[ITrigger[Any]]] = {
    # Словарь оперативно заполняющийся данными сопоставление объектов триггеров с объектами listener.
}

listeners_create_fabric: TListenerCreateFabric = {
    # Фабрика по созданию listeners в зависимости от типа.
    'transport': TransportListenerCreator(),
}
