"""Module contain interfaces and Base classes."""
from dataclasses import InitVar, dataclass, field
import socket
import struct
from typing import Any, Generic, Protocol, TypeVar, Literal, runtime_checkable
import asyncio

from triggers.core.exceptions import NotArtNetException


SOCKET_TYPES: dict[str, socket.SocketKind] = {
    'tcp': socket.SOCK_STREAM,
    'udp': socket.SOCK_DGRAM,
}


TPossibleTriggerTypes = Literal['ArtNet', 'RawUDP']
T = TypeVar('T', covariant=True)

TPossibleListenerTypes = Literal['transport']


@runtime_checkable
class ITransportListenerParams(Protocol):
    """Интерфейс для объекта параметров транспортного listener."""

    network_type: Literal['tcp', 'udp']
    listen_ip: str
    listen_port: int

    def __hash__(self) -> int:
        """Создает хеш."""
        ...


@runtime_checkable
class IListener(Protocol[T]):
    """Интерфейс для объекта с настройками listener."""

    listener_type: TPossibleListenerTypes
    checking_queue: asyncio.Queue[Any]
    close_event: asyncio.Event

    @property
    def params(self) -> T:
        """Параметры."""
        ...

    def __hash__(self) -> int:
        """Создает хеш."""
        ...


@runtime_checkable
class ISocketParams(Protocol):
    """Interface for Socket params class.

    Args:
        network_type (Literal['tcp', 'udp']) - Транспортный протокол.
        listen_ip (str) - прослушиваемый ip адрес.
        listen_port (int) - прослушиваемый порт.
    """

    network_type: Literal['tcp', 'udp']
    listen_ip: str
    listen_port: int


@runtime_checkable
class IRawUdpParams(Protocol):
    """Interface for RawUdp params class.

    Args:
        network_type (Literal['tcp', 'udp']) - Транспортный протокол.
        listen_ip (str) - прослушиваемый ip адрес.
        listen_port (int) - прослушиваемый порт.
        dara (str) - полезная нагрузка.
    """

    network_type: Literal['udp']
    listen_ip: str
    listen_port: int
    data: str

@runtime_checkable
class IArtNetParams(Protocol):
    """Interface for ArtNet params class.

    Args:
        network_type (Literal['tcp', 'udp']) - Транспортный протокол.
        listen_ip (str) - прослушиваемый ip адрес.
        listen_port (int) - прослушиваемый порт.
        universe (int) - вселенная.
        channel (int) - Номер DMX каналла.
        min_level (int) - Максисальное значение в канале.
        max_level (int) - Минимальное значение в канале.
    """

    network_type: Literal['tcp', 'udp']
    listen_ip: str
    listen_port: int
    universe: int
    channel: int
    min_level: int
    max_level: int


class ITrigger(Protocol[T]):
    """Интерфейс для класса с настройками триггера.

    Может принимать различные классы параметров. Класс параметров должен совпадать с типом триггера.
    Класс реализующий это интерфейс должен быть хэшируемым.

    Args:
        name (str) - Имя триггера.
        tr_type (str) - Тип триггера.
        params (T) - Класс параметров. Должен совпадать с типом триггера.
    """

    name: str
    tr_type: TPossibleTriggerTypes

    @property
    def params(self) -> T:
        """Параметры."""
        ...

    def __hash__(self) -> int:
        """Создает хеш."""
        ...


@dataclass(unsafe_hash=True)
class Trigger(Generic[T]):
    """Base Trigger."""

    name: str
    tr_type: TPossibleTriggerTypes
    params: T


@dataclass(unsafe_hash=True)
class SocketParams:
    """Socket trigger params."""

    network_type: Literal['tcp', 'udp']
    listen_ip: str
    listen_port: int
    data: str


@dataclass(unsafe_hash=True)
class ArtNetParams(SocketParams):
    """ArtNet trigger params."""
    network_type: Literal['tcp', 'udp'] = 'udp'
    listen_ip: str = '0.0.0.0'
    listen_port: int = 6454
    universe: int = 0
    data: str = 'any'


@dataclass(unsafe_hash=True)
class SocketTrigger(Trigger[SocketParams]):
    """Socket trigger."""


@dataclass(unsafe_hash=True)
class ArtNetTrigger(Trigger[ArtNetParams]):
    """ArtNet trigger."""


@dataclass
class ArtNetFrame:
    """ArtNet frame definitions."""

    artnet_identification: bytes = field(init=False)
    opcode: int = field(init=False)
    prot_ver_hi: int = field(init=False)
    prot_ver_low: int = field(init=False)
    sequence: int = field(init=False)
    physical: int = field(init=False)
    subuni: int = field(init=False)
    net: int = field(init=False)
    length_hi: int = field(init=False)
    length_lo: int = field(init=False)
    data: bytes = field(init=False)

    raw_bytes: InitVar[bytes]
    fmt: InitVar[bytes] = '8sHBBBBBBBB512s'

    def __post_init__(self, raw_bytes: bytes, fmt: str) -> None:
        """Unpack raw data to packet headers."""
        (
            self.artnet_identification, self.opcode, self.prot_ver_hi, self.prot_ver_low,
            self.sequence, self.physical, self.subuni, self.net,
            self.length_hi, self.length_lo, self.data,
        ) = struct.unpack(fmt, raw_bytes)

        if not self.artnet_identification == b'Art-Net\0':
            raise NotArtNetException
