"""Модуль содержит вспомогательные утилиты."""

import asyncio
import socket
from typing import Any, Iterable

from triggers.core.exceptions import SocketNotFoundError
from triggers.db.database import AsyncSessionLocal
from triggers.db.utils import get_all_serialized_triggers
from triggers.listeners.base import SOCKET_TYPES, ISocketParams, ITrigger, IListener, ITransportListenerParams
from triggers import settings


def get_socket_from_trigger(trigger: ITrigger[ISocketParams]) -> socket.socket:
    """Создает сокет из данных триггера.

    Args:
        trigger (InetTrigger): Триггер из которого можно создать сокет

    Returns:
        socket.socket: сокет

    Raise:
        OSError: raises if socket is already exist
    """
    sock = socket.socket(socket.AF_INET, SOCKET_TYPES[trigger.params.network_type])
    sock.bind(('', trigger.params.listen_port))
    return sock


def search_socket_by_trigger(trigger: ITrigger[ISocketParams], socket_list: Iterable[socket.socket]) -> socket.socket:
    """Ищет соответствие триггеру в списке соккетов.

    Args:
        trigger (InetTrigger)
        socket_list (list[socket.socket])

    Raises:
        SocketNotFoundError: Выбрасывает исключение если сокет не был найден.

    Returns:
        socket.socket
    """
    for soc in socket_list:
        if compare_trigger_and_socket(trigger, soc):
            return soc
    raise SocketNotFoundError(f'Socket for trigger {trigger.name} not found!')


def compare_trigger_and_socket(trigger: ITrigger[ISocketParams], soc: socket.socket) -> bool:
    """Сопоставляет параметры триггера и соккета.

    Args:
        trigger (InetTrigger)
        soc (socket.socket)

    Returns:
        bool
    """
    # FIXME: Переделать в отдельно вызывающиеся проверки.
    checks = [
        soc.type == SOCKET_TYPES[trigger.params.network_type],
        soc.getsockname() == (trigger.params.listen_ip, trigger.params.listen_port),
    ]

    return True if all(checks) else False


def compare_socket_trigger_and_transport_listener(
    trigger: ITrigger[Any],
    listener: IListener[Any],
) -> bool:
    input_params_checks = [
        isinstance(trigger.params, ISocketParams),
        isinstance(listener.params, ITransportListenerParams),
    ]

    if not all(input_params_checks):
        return False

    checks = [
        trigger.params.network_type == listener.params.network_type,
        trigger.params.listen_ip == listener.params.listen_ip,
    ]

    if trigger.params.listen_port != settings.ARTNET_DEFAULT_PORT:
        checks.append(trigger.params.listen_port == listener.params.listen_port)

    return True if all(checks) else False


async def create_listeners_from_triggers(triggers: Iterable[ITrigger[Any]], queue_trigger_add: asyncio.Queue[ITrigger[Any]]) -> None:
    for trigger in triggers:
        await queue_trigger_add.put(trigger)


async def inititate_listeners(
    queue_trigger_add: asyncio.Queue[ITrigger[Any]],
) -> None:
    async with AsyncSessionLocal() as async_session:
        triggers = await get_all_serialized_triggers(async_session)
    await create_listeners_from_triggers(triggers, queue_trigger_add)
