"""Main."""
import asyncio
import logging
import logging.config
from queue import Queue
from typing import Any

from paho.mqtt.client import MQTTMessage, MQTTv5

from triggers import settings
from triggers.actions.async_actions import (
    IRunActionCommand,
    RunMqttSendRawMsgActionCommand,
    RunMqttSendTriggerMsgActionCommand,
    handle_activated_trigger,
)
from triggers.actions.base import TPossibleActionTypes
from triggers.api.mqtt import MqttRPClient, MqttUserdataWithAsyncQueueDict
from triggers.checkers.checkers import TRIGGERS_CHECKS_FABRIC, FlowTriggerCheckerLoop, TCheckingQueue
from triggers.commands.base import IAsyncRouterCommand
from triggers.commands.publisher_commands import (
    PublishActionList,
    PublishRelationList,
    PublishTriggersList,
    send_action_list,
    send_relations_list,
    send_trigger_list,
)
from triggers.commands.router_commands import (
    AddNewActionCommand,
    AddNewTriggerCommand,
    DeleteActionCommand,
    DeleteTriggerCommand,
    CreateTriggerToActionRelationCommand,
    DeleteTriggerToActionRelationCommand,
)
from triggers.core.main_loop import AsyncMainLoop
from triggers.core.msg_handlers import AsyncRouterMsgHandlerWithResponse, RespMsgDict
from triggers.listeners.async_listeners import (
    TR_TYPE_TO_LISTENER_TYPE_SP,
    ListenerCreateLoop,
    ListenerDestroyLoop,
    listeners_create_fabric,
    listener_to_triggers_dict,
)
from triggers.listeners.base import ITrigger
from triggers.listeners.utils import inititate_listeners
from triggers.db.events import create_db_listens_publishes_tables_to_mqtt_after_db_commit


logging.config.dictConfig(settings.LOGGING_CONFIG)
logger = logging.getLogger(__name__)


def main() -> None:
    """Start module."""
    logger.info('Start trigger service')

    receiver_queue: asyncio.Queue[MQTTMessage] = asyncio.Queue()
    publisher_queue: Queue[RespMsgDict] = Queue()

    # FIXME: Обьеденить publisher класс в один принимающий топик и функцию возвращающую payload.
    settings.ServiceLocator = {
        'publish_trigger_list': PublishTriggersList(
            publisher_queue=publisher_queue,
            publisher_topic='lm/trigger_service/trigger/trigger_list',
        ),
        'publish_action_list': PublishActionList(
            publisher_queue=publisher_queue,
            publisher_topic='lm/trigger_service/action/action_list',
        ),
        'publish_relation_list': PublishRelationList(
            publisher_queue=publisher_queue,
            publisher_topic='lm/trigger_service/relation_list',
        ),
    }
    # Publish actions and triggers
    # FIXME: Объединить вызовы в отдельный обработчик берет все записи о публикации, и создает db listens.
    send_trigger_list()
    send_action_list()
    send_relations_list()
    create_db_listens_publishes_tables_to_mqtt_after_db_commit()

    async_queue_trigger_add: asyncio.Queue[ITrigger[Any]] = asyncio.Queue()
    async_queue_trigger_del: asyncio.Queue[ITrigger[Any]] = asyncio.Queue()

    # TODO: вынести router в отдельный файл.
    router: dict[str, IAsyncRouterCommand] = {
        'lm/trigger_service/trigger/add': AddNewTriggerCommand(async_queue_trigger_add),
        'lm/trigger_service/trigger/delete': DeleteTriggerCommand(async_queue_trigger_del),
        'lm/trigger_service/action/add': AddNewActionCommand(),
        'lm/trigger_service/action/delete': DeleteActionCommand(),
        'lm/trigger_service/set_trigger_to_action': CreateTriggerToActionRelationCommand(),
        'lm/trigger_service/erase_trigger_from_action': DeleteTriggerToActionRelationCommand(),
    }

    userdata: MqttUserdataWithAsyncQueueDict = {
        'topics': list(router),
        'msg_queue': receiver_queue,
        'publisher_queue': publisher_queue,
        'publisher_loop_timeout': settings.PUBLISHER_LOOP_TIMEOUT,
    }

    mqtt_client = MqttRPClient(client_id='lm_trigger_service', userdata=userdata, protocol=MQTTv5)
    mqtt_client.connect('localhost')
    mqtt_client.loop_start()
    mqtt_client.publisher_loop_start()
    logger.debug('Mqtt was started')

    msg_handler = AsyncRouterMsgHandlerWithResponse(router, publisher_queue, service_topics=settings.SERVICE_TOPICS)
    router_handle_loop = AsyncMainLoop(msg_queue=receiver_queue, msg_handler=msg_handler)

    loop = asyncio.get_event_loop_policy().get_event_loop()

    loop.create_task(router_handle_loop.start(), name='router_handler_task')

    triggers_active_queue: asyncio.Queue[ITrigger[Any]] = asyncio.Queue()
    trigger_checking_queue: TCheckingQueue = asyncio.Queue()

    action_command_fabric: dict[TPossibleActionTypes, IRunActionCommand] = {
        'send_mqtt_msg_raw': RunMqttSendRawMsgActionCommand(publisher_queue),
        'send_trigger_to_mqtt': RunMqttSendTriggerMsgActionCommand(publisher_queue),
    }

    loop.create_task(
        handle_activated_trigger(
            triggers_active_queue=triggers_active_queue,
            action_command_fabric=action_command_fabric,
        ),
        name='handle_active_trigger_task',
    )

    listener_create_loop = ListenerCreateLoop(
        TR_TYPE_TO_LISTENER_TYPE_SP,
        listeners_create_fabric,
        async_queue_trigger_add,
        trigger_checking_queue,
        listener_to_triggers_dict,
    )
    loop.create_task(
        listener_create_loop.start(),
        name='add_trigger_task',
    )

    listener_destroy_loop = ListenerDestroyLoop(async_queue_trigger_del, listener_to_triggers_dict)
    loop.create_task(
        listener_destroy_loop.start(),
        name='del_trigger_task',
    )

    trigger_checker_loop = FlowTriggerCheckerLoop(
        trigger_checking_queue,
        triggers_active_queue,
        listener_to_triggers_dict,
        TRIGGERS_CHECKS_FABRIC,
    )
    loop.create_task(trigger_checker_loop.start(), name='trigger_checker_loop')

    loop.create_task(inititate_listeners(async_queue_trigger_add))

    logger.debug('Starting main loop.')
    loop.run_forever()


if __name__ == '__main__':
    main()
