"""Бекап."""
from datetime import datetime
import logging
import pathlib
from shutil import copytree, ignore_patterns

from updater import settings
from updater.db.database import SessionLocal
from updater.db.utils import get_current_general_version


logger = logging.getLogger(__name__)


class CreateBackupByCopyFiles():
    """backup all lm files except cue by copy files from lm dir to backup dir.

    Args:
        backup_store: str = settings.BACKUPS_STORE_DIR,
        backup_dir (str):       path to backup folder
        excluded (list[str])    list of excluded files.
    """

    def __init__(self, backup_store: str = settings.BACKUPS_STORE_DIR, lm_base_dir: str = settings.LM_BASE_DIR) -> None:
        self._backup_store = backup_store
        self._lm_base_dir = lm_base_dir

    def create(self, backup_name: str, excluded: list[str]) -> str:
        """Run Command."""
        logger.info('Start full backup %s.', pathlib.Path(backup_name).name)
        backup_dir = self._backup_store + '/' + backup_name
        copytree(self._lm_base_dir, backup_dir, ignore=ignore_patterns(*excluded), dirs_exist_ok=True)
        logger.info('Backup created.')
        return backup_dir


class BackupNameMaker():
    """Backup filename generation logic."""

    # States
    _current_datetime: str = ''
    _version: str = ''
    _backup_filename: str = ''

    def make(self) -> str:
        """Generate backup filename."""
        self._get_current_datetime_stamp()
        self._get_version()
        self._glue_backup_filename()
        return self._backup_filename

    def _get_current_datetime_stamp(self) -> None:
        """Generate current datetime stamp in format %Y%m%d%H%M%S."""
        self._current_datetime = datetime.now().strftime('%Y%m%d%H%M%S')

    def _get_version(self) -> None:
        """Generate current version stamp in format version_subversion."""
        with SessionLocal() as session:
            self._version: str = get_current_general_version(session)

    def _glue_backup_filename(self) -> None:
        """Glue backup filename in format <datetime>_lmp_v<ver>.backup."""
        self._backup_filename = f"{self._current_datetime}_lmp_v{self._version}_full_backup"


class Backup:
    """Резервное копирование всех модулей lm.

    Args:
        backup_ignore (list[str]): Список игнорируемых файлов.

    """

    def __init__(
        self,
        backup_ignore: list[str] = settings.BACKUP_IGNORE,
        backup_filename_maker: BackupNameMaker = BackupNameMaker(),
        backup_creator: CreateBackupByCopyFiles = CreateBackupByCopyFiles(),
    ) -> None:
        self._backup_ignore = backup_ignore
        self._backup_filename_maker = backup_filename_maker
        self._backup_creator = backup_creator

    def create(self) -> str:
        """Создает резервную копию всех модулей lm.

        Returns:
            str: Путь до резервной копии.
        """
        backup_name: str = self._backup_filename_maker.make()
        excluded_files: list[str] = settings.BACKUP_IGNORE
        backup_path = self._backup_creator.create(backup_name, excluded_files)
        return backup_path
