"""Base module for system setting command."""
from abc import ABC, abstractmethod
from asyncio.log import logger
import pathlib
from typing import Any, Protocol, Type, TypedDict


class CommandInfoDict(TypedDict):
    """Command information.

    command: str    - command name
    data: Any       - data for command.
    """

    command: str
    data: Any


class Command(ABC):
    """Abstract class for command pattern."""

    @abstractmethod
    def execute(self) -> Any:
        """Contain all logic.

        The only one public method.
        """
        pass


class ICommandToLogic(Protocol):
    """Interface for command to logic.

    Args:
        payload (Any): msg payload

    Methods:
        execute: Run command. Return Any.
    """

    def __init__(self, payload: Any) -> None:
        ...

    def execute(self) -> Any:
        """Run command."""


class CommandRunner():
    """Execute commands.

    Args:
        commands_fabric (dict[str, Type[ICommandToLogic]]): fabric with executed command.

    Methods:
        run_command: execute property commands
    """

    def __init__(self, commands_fabric: dict[str, Type[ICommandToLogic]]) -> None:
        self.commands_fabric = commands_fabric

    def run_command(self, command: CommandInfoDict) -> None:
        """Execute property commands.

        Args:
            command (CommandInfoDict): property commands
        """
        logger.debug('CommandRunner start command "%s"', command['command'])
        command_runable = self.commands_fabric[command['command']]
        command_runable(command['data']).execute()


def dir_file_handle(src_path, file_handler, dir_handler, recursive=True, exclude=None) -> str:
    if not exclude:
        exclude = set()
    root_path = pathlib.Path(src_path)
    dir_list: list[pathlib.Path] = [root_path]
    for directory in dir_list:
        for path_object in directory.iterdir():
            if path_object.name in exclude:
                continue
            elif path_object.is_file():
                file_handler(path_object.absolute())
            elif path_object.is_dir():
                dir_handler(path_object.absolute())
                if recursive:
                    dir_list.append(path_object)
