"""Модуль содержит различные команды проверок для обновлений."""
import logging
import pathlib
import socket
import tarfile
from typing import Protocol, Type
from packaging import version


from updater import settings
from updater.commands.base import Command
from updater.db.database import SessionLocal
from updater.db.schema import UpdateSchema
from updater.db.models import Updates
from updater.db.utils import get_current_general_version, get_from_db
from updater.core.exceptions import ValidationException


logger = logging.getLogger(__name__)


class IUpdateCheckCommand(Protocol):
    """Interface for update checker class.

    Args:
        data (UpdateSchema): update info

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, data: UpdateSchema) -> None:
        ...

    def execute(self) -> None:
        """Run command."""


class CheckUpdateFileIsExistCommand(IUpdateCheckCommand):
    """Check that update file is exist.

    Args:
        data (UpdateSchema): contain information about update.

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        if not pathlib.Path(self.data.update_path).exists():
            raise ValidationException(message=f'The update file {self.data.filename} is not exist.')


class CheckVersionFromFilenameCommand(IUpdateCheckCommand):
    """Compare current version from database with new version from filename.

    Args:
        data (UpdateSchema): contain information about update.

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        with SessionLocal() as session:
            current_version: str = get_current_general_version(session)

        if version.parse(self.data.version) <= version.parse(current_version):
            raise ValidationException(
                message=f'The update version {self.data.version} is not higher than installed version.',
            )


class CheckUpdateIsNotExistInDbCommand(IUpdateCheckCommand):
    """Check that update is not present in the database.

    Args:
        data (UpdateSchema): dict contain information about update.

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        with SessionLocal() as session:
            version_is_exist = bool(session.query(Updates).filter_by(version=self.data.version).first())

        if version_is_exist:
            raise ValidationException(message=f'The update file {self.data.filename} is already exist in db.')


class CheckUpdateIsExistInDbCommand(IUpdateCheckCommand):
    """Check that update is present in the database.

    Args:
        data (UpdateSchema): dict contain information about update.

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        with SessionLocal() as session:
            is_exist = bool(get_from_db(session, Updates, update_path=self.data.update_path))

        if not is_exist:
            raise ValidationException(message=f'The update file {self.data.filename} is not exist in db.')


class CheckMetadataExistInTarCommand(IUpdateCheckCommand):
    """Check that the update tar file contain metadata file.

    Args:
        data (UpdateSchema): dict contain information about update tar file.

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    metadata_filename = settings.UPDATE_INFO_FILENAME

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        with tarfile.open(self.data.update_path, mode='r:*') as tar_file:
            try:
                tar_file.getmember(f'update/{self.metadata_filename}')
            except KeyError:
                raise ValidationException(message=f'Update {self.data.filename} is not contain metadata.')


def check_connection(host: str, port: int) -> bool:
    """Check connection to destination address by tcp.

    Args:
        host (str): hostname or ip address.
        port (int): network port number

    Returns:
        bool
    """
    soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        soc.connect((host, port))
    except socket.error:
        return False
    finally:
        soc.close()
    return True


class CheckInternetConnectionCommandAdapter(IUpdateCheckCommand):
    """Check internet connection.

    Args:
        data (UpdateSchema): not use.

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    host: str = settings.INET_CON_CHECK_PARAMS['checking_host']  # type: ignore
    port: int = settings.INET_CON_CHECK_PARAMS['checking_port']  # type: ignore

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        if not check_connection(self.host, self.port):
            raise ValidationException(
                message=f'The update {self.data.filename} can not be installed. Internet connection filed',
            )


class CheckIsTarCommand(IUpdateCheckCommand):
    """Check that the update file is a tar file.

    Args:
        data (UpdateSchema): data

    Rises:
        ValidationException: raises if check is false.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, data: UpdateSchema) -> None:
        self.data = data

    def execute(self) -> None:
        """Run command."""
        if not tarfile.is_tarfile(self.data.update_path):
            raise ValidationException(message=f'The update {self.data.filename} is not a tar file.')


checkers_is_archive_fabric: dict[str, Type[IUpdateCheckCommand]] = {
    'tar': CheckIsTarCommand,
}

checkers_metadata_in_archive_fabric: dict[str, Type[IUpdateCheckCommand]] = {
    'tar': CheckMetadataExistInTarCommand,
}


CHECKS_FABRIC: dict[str, Type[IUpdateCheckCommand]] = {
    'check_update_file_is_exist': CheckUpdateFileIsExistCommand,
    'check_update_dont_exist_in_db': CheckUpdateIsNotExistInDbCommand,
    'check_update_is_exist_in_db': CheckUpdateIsExistInDbCommand,
    'check_inet_connection': CheckInternetConnectionCommandAdapter,
    'check_version': CheckVersionFromFilenameCommand,
    'check_is_archive': checkers_is_archive_fabric[settings.ARCHIVE_TYPE],
    'check_metadata_exist_in_archive': checkers_metadata_in_archive_fabric[settings.ARCHIVE_TYPE],
}


class CheckUpdateSimpleCommand(Command):
    """Run checks in loop.

    Rises:
        ValidationException: raises if any of the checks will falls.

    Args:
        data (UpdateInfoDict):  update info.
        checks (list[str]):     checklist.
    """

    checks_fabric: dict[str, Type[IUpdateCheckCommand]] = CHECKS_FABRIC

    def __init__(self, data: UpdateSchema, checks: list[str]) -> None:
        self.data = data
        self.checks = checks

    def execute(self) -> None:
        """Run command."""
        logger.debug('Start checks for update %s', self.data.filename)
        for check_name in self.checks:
            check_command = self.checks_fabric[check_name]
            check_command(self.data).execute()
            logger.debug('Check %s passed.', pathlib.Path(check_name))
        logger.debug('All checks passed.')
