"""Commands."""
import logging
import pathlib
from queue import Queue
import shutil
import subprocess
from typing import Any, Optional, Protocol, Type, TypedDict

from pydantic import ValidationError

from updater import settings
from updater.commands.backup_command import Backup
from updater.commands.base import Command, ICommandToLogic
from updater.commands.update_checkers import CheckUpdateSimpleCommand
from updater.commands.utils import (
    ExtractUpdateCommand,
    IFilesExtractorCommand,
    RunFullRestoreInExternalModuleCommand,
    UpdateStatus,
    handle_update_result,
    parse_update_result,
    parse_version_from_filename,
)
from updater.core.exceptions import (
    Error,
    InstallationProcessException,
    NotFoundException,
    ValidationException,
)
from updater.core.msg_handlers import RespMsgDict
from updater.db.database import SessionLocal
from updater.db.models import Updates, Version
from updater.db.schema import (
    IdInputSchema,
    ModuleNameSchema,
    UpdateFileSchema,
    UpdateSchema,
    VersionSchema,
)
from updater.db.utils import (
    get_from_db,
    get_stacked_update_from_db,
    save_new_versions_in_db,
    save_update_changes_to_db,
    update_db_obj,
)


logger = logging.getLogger(__name__)


class UpdateInfoDict(TypedDict):
    """Update info.

    file: pathlib.Path - path to file
    """

    file: pathlib.Path


class IUpdateHandlerCommandCommand(Protocol):
    """Interface for set params to network interface.

    Args:
        data (Any): data

    Returns:
        None
    """

    def __init__(self, data: Any) -> None:
        ...

    def execute(self) -> None:
        """Run command."""


class PublisherAllTableRow():
    """Публикует все записи из таблицы бд по mqtt.

    Args:
        publisher_queue (Queue[RespMsgDict]): Очередь для отправки по mqtt.

    FIXME: PublisherCommand, ГРЯЗЬ!!!!
    """

    def __init__(self, publisher_queue: Queue[RespMsgDict]) -> None:
        self.publisher_queue = publisher_queue

    def __call__(self, edited_models: set[Any]) -> None:
        """Публикует все записи из таблицы бд по mqtt.

        Args:
            edited_models (set[Any]): Модель sqlalchemy.
        """
        for edited_model in edited_models:
            if edited_model is Updates:
                topic = 'lm/update_service/update/update_list'
                payload = GetUpdateList(None).execute()
                self._send_to_queue(topic, payload, retain=True)
            elif edited_model is Version:
                topic = 'lm/update_service/version/version_list'
                payload = GetVersionsList(None).execute()
                self._send_to_queue(topic, payload, retain=True)

    def _send_to_queue(self, topic: str, payload: Any, retain: bool) -> None:
        pub_msg: RespMsgDict = {
            'topic': topic,
            'cor_data': None,
            'retain': retain,
            'payload': payload,
        }
        self.publisher_queue.put(pub_msg)


class HandleStackingUpdate(ICommandToLogic):
    """Handle stacking update.

    Args:
        payload: None
    """

    def __init__(self, payload: Any) -> None:
        pass

    def execute(self) -> None:
        """Run command."""
        updates_stacked: list[UpdateSchema] = get_stacked_update_from_db()
        for update in updates_stacked:
            if update.status == 'installing':
                update_state: list[dict[str, str]] = parse_update_result(update.extracted_path)
                update_status: UpdateStatus = handle_update_result(update_state)
                if update_status.status == 'installed':
                    save_new_versions_in_db(update, update_state)
            elif update.status == 'backup creating':
                update_status = UpdateStatus('downloaded', 'Update was restored.')
            else:
                update_status = UpdateStatus(
                    'error',
                    f"Update {update.filename} was stacked in {update.status} state.",
                )
            update.status = update_status.status
            update.description = update_status.msg
            save_update_changes_to_db(update)


class GetVersionsList(ICommandToLogic):
    """Get version list of all modules.

    Args:
        payload: None
    """

    def __init__(self, payload: Any) -> None:
        pass

    def execute(self) -> list[dict[str, str]]:
        """Run command."""
        # get from db
        with SessionLocal() as session:
            versions_query = session.query(Version).all()

        # serialization
        versions_list: list[dict[str, str]] = []
        for version_obj in versions_query:
            version: VersionSchema = VersionSchema.from_orm(version_obj)
            versions_list.append(version.dict())

        return versions_list


class GetModuleVersion(ICommandToLogic):
    """Get version list of all modules.

    Args:
        payload (dict[str, str]):   dict in format {'module': <module_name>}
    """

    initial_checks: list[str] = settings.INITIAL_UPDATE_CHECKS

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> Optional[dict[str, str]]:
        """Run command."""
        # validation input
        try:
            module_name: str = ModuleNameSchema.parse_raw(self.payload).module
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())

        # get from db
        with SessionLocal() as session:
            versions_obj = get_from_db(session, Version, module=module_name)
        if not versions_obj:
            raise NotFoundException(f'Module "{module_name}" not found')
        # serialization
        version: VersionSchema = VersionSchema.from_orm(versions_obj)

        return dict(version)


class GetUpdateList(ICommandToLogic):
    """Get updates list.

    Args:
        payload: None
    """

    def __init__(self, payload: Optional[bytes]) -> None:
        pass

    def execute(self) -> list[dict[str, str]]:
        """Run command."""
        # get from db
        with SessionLocal() as session:
            updates_query = session.query(Updates).all()

        # serialization
        updates_list: list[dict[str, str]] = []
        for update_obj in updates_query:
            update: UpdateSchema = UpdateSchema.from_orm(update_obj)
            updates_list.append(update.dict())

        return updates_list


class RestoreUpdateCommand(ICommandToLogic):
    """Restore update command.

    Args:
        payload (bytes): dict in format {'id': <id>}.
    """

    restore_command = RunFullRestoreInExternalModuleCommand

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> None:
        """Run command."""
        self._parse_input()
        self._get_update()
        self._restore_update()
        self._delete_related_backup()

    def _parse_input(self) -> None:
        try:
            self._input_data: IdInputSchema = IdInputSchema.parse_raw(self.payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())

    def _get_update(self) -> None:
        with SessionLocal() as session:
            update_obj: Optional[Updates] = get_from_db(session, Updates, id=self._input_data.id)

            if not update_obj:
                raise ValidationException(message=f"Update with id {self._input_data.id} not found.")

            if not update_obj.backup_path:
                raise ValidationException(message=f"Backup for update {self._input_data.id} not found.")

        self._update = UpdateSchema.from_orm(update_obj)

    def _restore_update(self) -> None:
        self._change_update_status('restoring')

        self.restore_command(self._update.backup_path).execute()

        self._change_update_status('downloaded')

    def _delete_related_backup(self) -> None:
        shutil.rmtree(self._update.backup_path, ignore_errors=True)

    def _update_db_object(self) -> None:
        logger.debug('Saving new update status "%s" for update %s to db', self._update.status, self._update.version)
        with SessionLocal() as session:
            update_db_obj(session, Updates, self._update)
        logger.debug('Update %s status was updated into db.', self._update.version)

    def _change_update_status(self, new_status: str, descr: str = '') -> None:
        self._update.status = new_status
        self._update.description = descr
        self._update_db_object()


class InstallUpdateCommand(ICommandToLogic):
    """Install update command.

    Args:
        ICommandToLogic (_type_): _description_
    """

    preinstall_checklist: list[str] = settings.BEFORE_INSTALL_UPDATE_CHECKS
    postinstall_checklist: list[str] = settings.POST_INSTALL_UPDATE_CHECKS

    backup_store: str = settings.BACKUPS_STORE_DIR
    update_store: str = settings.UPDATES_STORE_DIR

    restore_command: Type[ICommandToLogic] = RunFullRestoreInExternalModuleCommand
    extract_command: Type[IFilesExtractorCommand] = ExtractUpdateCommand

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> None:
        """Run command."""
        self._parse_input()
        try:
            self._pred_install_check()
            self._create_backup()
            self._extract_update_archive()
            self._install_update()
            self._inspect_installation_result()
            self._post_install_checks()
        except Error as err:
            self._change_update_status('error', err.message)
            logger.warning(err.message)
            raise err
        self._clear()

    def _parse_input(self) -> None:
        """Parse and validation input."""
        logger.debug('Validating input data.')
        try:
            input_data: IdInputSchema = IdInputSchema.parse_raw(self.payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

        with SessionLocal() as session:
            update_obj = get_from_db(session, Updates, id=input_data.id)
            if not update_obj:
                raise NotFoundException(f'Update with id "{input_data.id}" not found')

        self.update = UpdateSchema.from_orm(update_obj)

    def _pred_install_check(self) -> None:
        self._change_update_status('pre-installation checking')

        check_command = CheckUpdateSimpleCommand(self.update, self.preinstall_checklist)
        check_command.execute()

        self._change_update_status('pre-installation checks passed')

    def _extract_update_archive(self) -> None:
        self._change_update_status('extracting')

        extracted_path: str = self.update_store + '/' + self.update.filename.split('.')[0]
        pathlib.Path(extracted_path).mkdir(exist_ok=True)

        self.extract_command(self.update.update_path, extracted_path).execute()
        self.update.extracted_path = extracted_path

        self._change_update_status('extracted')

    def _create_backup(self) -> None:
        self._change_update_status('backup creating')
        self.update.backup_path = Backup().create()
        self._change_update_status('backup created')

    def _install_update(self) -> None:
        logger.debug('Start installing update %s', self.update.filename)
        self._change_update_status('installing')

        cmd: list[str] = [
            settings.PYTHON_EXECUTABLE_PATH,
            '-m',
            'update',
            '-b',
            self.update.backup_path,
        ]

        try:
            subprocess.run(cmd, cwd=self.update.extracted_path, check=True, start_new_session=True)
        except subprocess.CalledProcessError as err:
            err_msg = f'A error occurred during installation update. Installation filed. {err.stdout}'
            self.restore_command(self.update.backup_path).execute()
            raise InstallationProcessException(message=err_msg)

    def _inspect_installation_result(self) -> None:
        update_result: list[dict[str, str]] = parse_update_result(self.update.extracted_path)
        upd_status: UpdateStatus = handle_update_result(update_result)
        self._change_update_status(upd_status.status, upd_status.msg)
        if self.update.status == 'error':
            raise InstallationProcessException(message=self.update.description)
        elif self.update.status == 'installed':
            logger.info('Update %s has been installed.', self.update.filename)
            save_new_versions_in_db(self.update, update_result)

    def _post_install_checks(self) -> None:
        check_command = CheckUpdateSimpleCommand(self.update, self.postinstall_checklist)
        try:
            check_command.execute()
        except Error as err:
            self.restore_command(self.update.backup_path).execute()
            raise err

    def _clear(self) -> None:
        pass

    def _change_update_status(self, new_status: str, descr: str = '') -> None:
        self.update.status = new_status
        self.update.description = descr
        save_update_changes_to_db(self.update)


class DeleteUpdateCommand(ICommandToLogic):
    """Delete update.

    Args:
        ICommandToLogic (_type_): _description_

    Raises:
        ValidationException: _description_

    Returns:
        _type_: _description_

    TODO: AddUpdateCommand Написать докстринг
    """

    checks: list[str] = settings.BEFORE_ADDED_UPDATE_CHECKS

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> None:
        """Run command."""
        self._parse_input()
        logger.debug('Start initial for update %s', self.update.filename)
        self._delete_update_files()
        self._delete_db_object()

    def _parse_input(self) -> None:
        """Parse and validation input."""
        logger.debug('Validating input data.')
        try:
            input_data: IdInputSchema = IdInputSchema.parse_raw(self.payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

        with SessionLocal() as session:
            update_obj = get_from_db(session, Updates, id=input_data.id)
            if not update_obj:
                raise NotFoundException(f'Update with id "{input_data.id}" not found')
            update_obj.status = 'deleting'
            session.add(update_obj)
            session.commit()
            session.refresh(update_obj)

        self.update = UpdateSchema.from_orm(update_obj)

    def _delete_update_files(self) -> None:
        logger.debug('Deleting files for update %s', self.update.version)
        if self.update.backup_path:
            shutil.rmtree(self.update.backup_path, ignore_errors=True)
        if self.update.extracted_path:
            shutil.rmtree(self.update.extracted_path, ignore_errors=True)
        if self.update.update_path:
            pathlib.Path(self.update.update_path).unlink(missing_ok=True)
        logger.debug('All files of the update %s was deleted', self.update.version)

    def _delete_db_object(self) -> None:
        logger.debug('Deleting the update version %s from the db', self.update.version)
        with SessionLocal() as session:
            deleted_obj: Updates = session.query(Updates).filter(Updates.id == self.update.id).one()
            session.delete(deleted_obj)
            session.commit()
        logger.debug('Update version %s was deleted from db.', self.update.version)


class AddUpdateCommand(ICommandToLogic):
    """Add new update to database.

    Args:
        ICommandToLogic (_type_): _description_

    Raises:
        ValidationException: _description_

    Returns:
        _type_: _description_
    TODO: AddUpdateCommand Написать докстринг
    """

    checks: list[str] = settings.BEFORE_ADDED_UPDATE_CHECKS

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> None:
        """Run command."""
        self._parse_input()
        logger.debug('Start initial for update %s', self.update.filename)
        self._check_update_file()
        self._create_db_object()

    def _parse_input(self) -> None:
        """Parse and validation input."""
        try:
            input_data: UpdateFileSchema = UpdateFileSchema.parse_raw(self.payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())

        filename = input_data.file.name
        version = parse_version_from_filename(filename)
        if not version:
            raise ValidationException(message='Payload validation error.')

        self.update = UpdateSchema(
            version=version,
            update_path=str(input_data.file),
            filename=filename,
            status='downloaded',
        )

    def _check_update_file(self) -> None:
        CheckUpdateSimpleCommand(self.update, self.checks).execute()

    def _create_db_object(self) -> None:
        logger.debug('Saving update version %s to db', self.update.version)
        with SessionLocal() as session:
            update_obj: Updates = Updates(**self.update.dict())
            session.add(update_obj)
            session.commit()
            session.refresh(update_obj)
        logger.debug('Update version %s was written to db.', self.update.version)


class CheckUpdateCommand(ICommandToLogic):
    """Check update.

    Args:
        ICommandToLogic (_type_): _description_

    Raises:
        ValidationException: _description_

    Returns:
        _type_: _description_
    TODO: CheckUpdateCommand Написать докстринг
    """

    checks: list[str] = settings.INITIAL_UPDATE_CHECKS

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> None:
        """Run command."""
        self._parse_input()
        self._check_update_file()

    def _parse_input(self) -> None:
        """Parse and validation input."""
        logger.debug('Validating input data.')
        try:
            input_data: IdInputSchema = IdInputSchema.parse_raw(self.payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())
        logger.debug('Input data validated successful.')

        with SessionLocal() as session:
            update_obj = get_from_db(session, Updates, id=input_data.id)
            if not update_obj:
                raise NotFoundException(f'Update with id "{input_data.id}" not found')

        self.update = UpdateSchema.from_orm(update_obj)

    def _check_update_file(self) -> None:
        logger.debug('Start checks for update %s', self.update.filename)
        self.update.status = 'checking'
        save_update_changes_to_db(self.update)
        try:
            CheckUpdateSimpleCommand(self.update, self.checks).execute()
        except ValidationException as err:
            self.update.status = 'error'
            raise err
        else:
            self.update.status = 'checked'
        finally:
            save_update_changes_to_db(self.update)


class InitialUpdateCommand(ICommandToLogic):
    """Initialize update.

    Args:
        ICommandToLogic (_type_): _description_
    """

    initial_checks: list[str] = settings.INITIAL_UPDATE_CHECKS

    def __init__(self, payload: bytes) -> None:
        self.payload = payload

    def execute(self) -> None:
        """Run command."""
        self._parse_input()
        logger.debug('Start initial for update %s', self.update.filename)
        self._check_update_file()
        self._create_db_object()

    def _parse_input(self) -> None:
        """Parse and validation input."""
        try:
            input_data: UpdateFileSchema = UpdateFileSchema.parse_raw(self.payload)
        except ValidationError as err:
            raise ValidationException(message='Payload validation error.', data=err.errors())

        filename = input_data.file.name
        version = parse_version_from_filename(filename)

        self.update = UpdateSchema(
            version=version,
            update_path=str(input_data.file),
            filename=filename,
            status='downloaded',
        )

    def _check_update_file(self) -> None:
        CheckUpdateSimpleCommand(self.update, self.initial_checks).execute()
        self.update.status = 'checked'

    def _create_db_object(self) -> None:
        logger.debug('Saving update version %s to db', self.update.version)
        with SessionLocal() as session:
            update_obj: Updates = Updates(**self.update.dict())
            session.add(update_obj)
            session.commit()
            session.refresh(update_obj)
        logger.debug('Update version %s was written to db.', self.update.version)
