"""Вспомогательные модули и классы для update commands."""
from dataclasses import dataclass
import json
import logging
import os
import pathlib
import re
from shutil import copytree
import subprocess
import tarfile
from typing import Protocol, Type
import zipfile

from updater import settings
from updater.commands.base import Command


logger = logging.getLogger(__name__)


@dataclass
class UpdateStatus:
    """Статус обновления.

    status (str):  Статус обновления.
    msg (str):  Пояснительное сообщение.
    """

    status: str
    msg: str


class IFilesExtractorCommand(Protocol):
    """Interface for update checker class.

    Args:
        file (str):     archived file path.
        dst_dir (str):  destination path.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, file: str, dst_dir: str) -> None:
        ...

    def execute(self) -> None:
        """Run command."""


class ExtractZipFilesCommand(IFilesExtractorCommand):
    """Command for executing zip archive.

    Args:
        file (str):     archived file path.
        dst_dir (str):  destination path.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, file: str, dst_dir: str) -> None:
        self.file = file
        self.dst_dir = dst_dir

    def execute(self) -> None:
        """Run Command."""
        with zipfile.ZipFile(self.file, mode="r") as archive:
            archive.extractall(self.dst_dir)


class ExtractTarGzCommand(IFilesExtractorCommand):
    """Command for executing tar gz archive.

    Args:
        file (str):     archived file path.
        dst_dir (str):  destination path.

    Methods:
        execute: Run command. Return None.
    """

    def __init__(self, file: str, dst_dir: str) -> None:
        self.file = file
        self.dst_dir = dst_dir

    def execute(self) -> None:
        """Run Command."""
        with tarfile.open(self.file, mode='r:*') as tar_file:
            tar_file.extractall(self.dst_dir)


extractors_fabric: dict[str, Type[IFilesExtractorCommand]] = {
    'zip': ExtractZipFilesCommand,
    'tar': ExtractTarGzCommand,
}


class ExtractUpdateCommand(IFilesExtractorCommand):
    """Extract archive dependency of type.

    Args:
        file (str):     archived file path.
        dst_dir (str):  destination path.

    Methods:
        execute: Run command. Return None.
    """

    archive_type: str = settings.ARCHIVE_TYPE
    extractors_fabric = extractors_fabric

    def __init__(self, file: str, dst_dir: str) -> None:
        self.file = file
        self.dst_dir = dst_dir

    def execute(self) -> None:
        """Run command."""
        logger.debug('Start extract update')
        extract_command = extractors_fabric[self.archive_type]
        extract_command(self.file, self.dst_dir).execute()
        logger.debug('Update extracted.')


class RunFullRestoreCommand(Command):
    """Restore lm by copy files from backup to lm dir.

    Args:
        backup_dir (str): path to backup folder
    """

    lm_base_dir = settings.LM_BASE_DIR

    def __init__(self, backup_dir: str) -> None:
        self.backup_dir = backup_dir

    def execute(self) -> None:
        """Run Command."""
        logger.info('Start full restore from backup %s.', pathlib.Path(self.backup_dir).name)
        copytree(self.backup_dir, self.lm_base_dir, dirs_exist_ok=True)
        logger.info('Restore finished.')


class RunFullRestoreInExternalModuleCommand(Command):
    """Restore lm by external module.

    Args:
        backup_dir (str): path to backup folder
    """
    # TODO: RunFullRestoreInExternalModuleCommand Переписать на запуск в отдельном сервисе
    # что бы не было проблем с перезапуском lm_update

    lm_base_dir = settings.LM_BASE_DIR

    def __init__(self, backup_dir: str) -> None:
        self.backup_dir = backup_dir

    def execute(self) -> None:
        """Run Command."""
        logger.info('Start full restore from backup %s.', pathlib.Path(self.backup_dir).name)

        running_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent

        cmd: list[str] = [
            settings.PYTHON_EXECUTABLE_PATH,
            '-m',
            'restore',
            '-b',
            self.backup_dir,
            '-p',
            str(os.getpid()),
        ]

        proc = subprocess.Popen(cmd, cwd=running_dir, close_fds=True, start_new_session=True)
        proc.communicate()
        logger.info('Restore finished.')


def handle_update_result(update_state: list[dict[str, str]]) -> UpdateStatus:
    """Проверяет результат обновления.

    Args:
        update_state (list[dict[str, str]]):    Update state.

    Returns:
        UpdateStatus: Статус обновления.
    """
    if check_installed_ok(update_state):
        status = 'installed'
        description = ''
    else:
        if check_update_restored_to_original_state(update_state):
            err_msg = 'A error occurred during installation update. Installation filed. Old version was restored.'
        elif check_update_didnt_start_installation(update_state):
            err_msg = 'Update was started but nothing was happened.'
        else:
            err_msg = 'A error occurred during installation update. Installation filed.'
        status = 'error'
        description = err_msg
    return UpdateStatus(status, description)


def parse_update_result(update_path: str) -> list[dict[str, str]]:
    """Parse updating state.

    Args:
        update_path (str): path to update dir.

    Returns:
        list[dict[str, str]]: parsed state.
    """
    upd_status_filename: str = settings.UPDATE_RESULT_FILENAME
    with open(f"{update_path}/update/{upd_status_filename}") as state_file:
        state: list[dict[str, str]] = json.load(state_file)
    return state


def check_installed_ok(status: list[dict[str, str]]) -> bool:
    """Check that all modules have been updated.

    All modules must have 'installed' status.

    Args:
        status (list[dict[str, str]):  update result.

    Returns:
        bool
    """
    statuses: set[str] = set()

    right_statuses = {frozenset(['installed'])}

    for module in status:
        statuses.add(module['status'])

    return statuses in right_statuses


def check_update_restored_to_original_state(status: list[dict[str, str]]) -> bool:
    """Check that modules not installed but restored.

    Some or all modules have 'restored' status.
    Some or none modules have 'prepared' status.

    Args:
        status (list[dict[str, str]):  update result.

    Returns:
        bool
    """
    statuses: set[str] = set()

    right_statuses = {frozenset(['prepared', 'restored']), frozenset(['restored'])}

    for module in status:
        statuses.add(module['status'])

    return statuses in right_statuses


def check_update_didnt_start_installation(status: list[dict[str, str]]) -> bool:
    """Check that modules did not start installation.

    All modules must have 'prepared' status.

    Args:
        status (list[dict[str, str]):  update result.

    Returns:
        bool
    """
    statuses: set[str] = set()

    right_statuses = {frozenset(['prepared'])}

    for module in status:
        statuses.add(module['status'])

    return statuses in right_statuses


def check_update_finish_with_error(status: list[dict[str, str]]) -> bool:
    """Check that error present in state.

    One or more modules have 'error' status.

    Args:
        status (list[dict[str, str]):  update result.

    Returns:
        bool
    """
    statuses: set[str] = set()

    for module in status:
        statuses.add(module['status'])

    return 'error' in statuses


def parse_version_from_filename(filename: str) -> str:
    """Parse version number form update filename.

    Args:
        filename (str): update filename

    Returns:
        str: version
    """
    version: str = ''
    # TODO: parse_version_from_filename. Вынести паттерн в настройки
    pattern = re.compile(r'^lmp_(?P<raw_ver>\d{9}).update$')
    if matched_obj := pattern.match(filename):
        raw_ver = matched_obj.groupdict()['raw_ver']
        version = '.'.join([str(int(value)) for value in re.findall(r'\d{3}', raw_ver)])
    return version
