"""Msg handlers."""
import logging
from queue import Queue
from typing import Any, Callable, Optional, Protocol, Type, TypedDict

from paho.mqtt.client import MQTTMessage

from updater.commands.base import CommandInfoDict, ICommandToLogic
from updater.core.exceptions import Error


logger = logging.getLogger(__name__)


class RespMsgDict(TypedDict):
    """Response msg dict.

    topic: str                  - response topic.
    cor_data: Optional[str]     - uniq request id.
    payload: Any                - json serializable payload.
    """

    topic: str
    cor_data: Optional[str]
    payload: Any
    retain: bool


class IMsgHandler(Protocol):
    """Interface for message handlers.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(self, topic_fabric: dict[str, Callable[[CommandInfoDict], None]]) -> None:
        ...

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        ...


class NaiveMsgHandler():
    """Naive message handlers.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(self, topic_fabric: dict[str, Callable[[CommandInfoDict], None]]) -> None:
        self.topic_fabric = topic_fabric

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        if command := self.topic_fabric.get(msg['topic']):
            command(msg.get('payload'))


class NaiveMsgHandlerWithErrorTopic():
    """Naive message handlers with error topics.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.
    Handle error. if error occurred then error message puts to the error topic.

    Handles only custom Error exception.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(self, topic_fabric: dict[str, Callable[[CommandInfoDict], None]], error_queue: Queue[Any]) -> None:
        self.topic_fabric = topic_fabric
        self.error_queue = error_queue

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        logger.debug('Start handle msg %s', msg)
        if command := self.topic_fabric.get(msg['topic']):
            try:
                command(msg.get('payload'))
            except Error as err:
                logger.warning(err.message)

                payload: dict[str, Any] = {
                    'topic': 'error',
                    'msg': err.message,
                    'data': msg.get('payload'),
                }
                self.error_queue.put(payload)


class MsgHandlerWithResponse():
    """Naive message handlers with error topics.

    Receives a command from topic_fabric, passes mqtt msg payload to it, then execute it.
    Handle error. if error occurred then error message puts to the error topic.

    Handles only custom Error exception.

    Args:
        topic_fabric: fabric comprises topic names as keys and func that input msg payload as value.
    """

    def __init__(
        self, topic_fabric: dict[str, Type[ICommandToLogic]],
        resp_queue: Queue[RespMsgDict], service_topics: dict[str, str],
    ) -> None:
        self.topic_fabric = topic_fabric
        self.resp_queue = resp_queue
        self.error_topic = service_topics['error']

    def handle_msg(self, msg: MQTTMessage) -> None:
        """Proceed mqtt message.

        Args:
            msg (MQTTMessage): Mqtt message
        """
        logger.debug('Start handle msg %s', msg)
        cor_data: Optional[str] = getattr(msg.properties, 'CorrelationData', None)
        resp_topic: str = getattr(msg.properties, 'ResponseTopic', None) or f"{msg.topic}/response"
        if command := self.topic_fabric.get(msg.topic):
            try:
                resp_payload: Any = command(msg.payload).execute()
            except Error as err:
                logger.warning(err.message)
                resp_topic = self.error_topic
                resp_payload = {
                    'msg': err.message,
                    'data': err.data,
                }

            if resp_payload:
                resp_msg: RespMsgDict = {
                    'topic': resp_topic,
                    'cor_data': cor_data,
                    'payload': resp_payload,
                    'retain': False,
                }
                self.resp_queue.put(resp_msg)
