"""Models."""
from itertools import chain
from typing import Any
from sqlalchemy import Column, Integer, String
from sqlalchemy import event
from sqlalchemy.orm import Session

from updater import settings
from updater.db.database import Base, SessionLocal


class Version(Base):
    """Contain information about current installed versions."""

    __tablename__ = 'current_version'

    id = Column(Integer, primary_key=True, unique=True)
    version = Column(String(50), nullable=False)
    subversion = Column(String(50), nullable=True)
    module = Column(String(50), unique=True)
    description = Column(String, nullable=True)


class Updates(Base):
    """Table contain info about updates."""

    __tablename__ = 'updates'

    id = Column(Integer, primary_key=True)
    version = Column(String(50), unique=True)
    status = Column(String)
    filename = Column(String)
    update_path = Column(String)
    extracted_path = Column(String)
    backup_path = Column(String)
    description = Column(String, nullable=True)


@event.listens_for(SessionLocal, 'before_commit')
def before_commit(session: Session, *args: Any, **kwargs: Any) -> None:
    """Hook executed before commit to the db.

    Add created, edited or deleted object to _edited_objects variable for future processing.

    Args:
        session (Session): Alchemy session.
    """
    session._edited_objects = session.new | session.deleted | session.dirty


@event.listens_for(SessionLocal, 'after_commit')
def after_commit(session: Session, *args: Any, **kwargs: Any) -> None:
    """Hook executed after commit to the db.

    Adds data to mqtt send queue if Updates or Version tables was changing.

    Args:
        Session (Session): Alchemy session.
    """
    edited_models: set = set()
    for edited_obj in session._edited_objects:
        edited_models.add(type(edited_obj))
    if edited_models:
        command = settings.ServiceLocator.get('publish_table')
        if command:
            command(edited_models)
