"""Utilities for working with db."""
import logging
from typing import Optional, Any, Type, TypeVar

from pydantic import BaseModel
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query
from updater.db.database import SessionLocal

from updater.db.models import Version, Updates
from updater.db.schema import UpdateSchema


logger = logging.getLogger(__name__)


T = TypeVar("T", Version, Updates)


def get_or_create(session: Session, model: Type[T], **kwargs: Any) -> T:
    """Get or create db instance.

    Args:
        session (Session): Alchemy session.
        model (Base): declarative base model.

    Raises:
        sqlalchemy.orm.exc.MultipleResultsFound: raises if multiple object identities are returned,
            or if multiple rows are returned for a query that returns only scalar values as opposed
            to full identity-mapped entities.

    Returns:
        Base: model instance.
    """
    instance: Optional[T] = session.query(model).filter_by(**kwargs).one_or_none()
    if not instance:
        instance = model(**kwargs)
        session.add(instance)
        session.commit()
    return instance


def get_from_db(session: Session, model: Type[T], **kwargs: Any) -> Optional[T]:
    """Get db instance.

    Args:
        session (Session): Alchemy session.
        model (Base): declarative base model.

    Raises:
        sqlalchemy.orm.exc.MultipleResultsFound: raises if multiple object identities are returned,
            or if multiple rows are returned for a query that returns only scalar values as opposed
            to full identity-mapped entities.

    Returns:
        Base: model instance.
    """
    instance: Optional[T] = session.query(model).filter_by(**kwargs).one_or_none()
    return instance


def update_db_obj(session: Session, model: Type[T], update_item: BaseModel) -> bool:
    """Update object into db.

    Args:
        session (Session):          Alchemy session.
        model (Type[T]):            Alchemy model class.
        update_item (BaseModel):    Pydantic schema instance. 

    Returns:
        bool: _description_
    """
    upd_obj = session.query(model).filter_by(id=update_item.id).one_or_none()
    if not upd_obj:
        return False

    for key, value in update_item.dict(exclude=set('id')).items():
        setattr(upd_obj, key, value)

    session.add(upd_obj)
    session.commit()
    session.refresh(upd_obj)
    return True


def get_current_general_version(session: Session) -> str:
    """Get current general version number.

    Args:
        session (Session): Alchemy session.

    Returns:
        str: Current general version or '0' if general version is None
    """
    version_obj: Optional[Version] = session.query(Version).filter(Version.module == 'general').first()
    return version_obj.version if version_obj else '0'


def update_general_version(session: Session, version: str) -> Version:
    """Update general version number.

    Args:
        session (Session):  Alchemy session.
        version (str):      new version number.

    Returns:
        Version: general version obj.
    """
    version_obj: Version = session.query(Version).filter(Version.module == 'general').one()
    version_obj.version = version
    return version_obj


def update_module_version(session: Session, module: str, version: str) -> Version:
    """Update module version number.

    Args:
        session (Session):  Alchemy session.
        module (str):       module name.
        version (str):      new version number.

    Returns:
        Version: version obj.
    """
    version_obj = get_from_db(session, Version, module=module)
    if version_obj:
        version_obj.version = version
    else:
        version_obj = Version(
            version=version,
            module=module,
        )
    return version_obj


def get_db_state_stacking_updates(session: Session) -> "Query[Updates]":
    """Создает выборку всех объектов Updates имеющих статус с окончанием ing.

    Args:
        session (Session): Alchemy session.
    """
    result = session.query(Updates).filter(Updates.status.endswith('ing'))
    return result


def get_stacked_update_from_db() -> list[UpdateSchema]:
    with SessionLocal() as session:
        update_db_objects = get_db_state_stacking_updates(session)
    updates_stacked: list[UpdateSchema] = [UpdateSchema.from_orm(upd_obj) for upd_obj in update_db_objects]
    return updates_stacked


def save_new_versions_in_db(update: UpdateSchema, update_state: list[dict[str, str]]) -> None:
    with SessionLocal() as session:
        for module in update_state:
            update_obj: Version = update_module_version(session, module['module'], module['version'])
            session.add(update_obj)
        update_obj = update_module_version(session, 'general', update.version)
        session.add(update_obj)
        session.commit()


def save_update_changes_to_db(update: UpdateSchema) -> None:
    logger.debug('Saving new update status "%s" for update %s to db', update.status, update.filename)
    with SessionLocal() as session:
        update_db_obj(session, Updates, update)
    logger.debug('Update %s status was updated into db.', update.filename)
