"""Entry point."""
import logging
import logging.config
import argparse
import os
import signal
import subprocess
import json
import pathlib
import shutil
import time
from typing import Any, Protocol, Type, TypedDict

from restore.log_config import LOGGING_CONFIG


RUNNING_PATH = pathlib.Path(os.path.realpath(__file__)).parent
PARENT_PID = None


logging.config.dictConfig(LOGGING_CONFIG)
logger = logging.getLogger(__name__)


class ReplaceFilesSettingsDict(TypedDict):
    """Replace file settings dict.

    src: str - source path.
    dst: str - destination path.
    """

    src: str
    dst: str


class SystemdSettings(TypedDict):
    """Dict with settings for manipulating systemd service.

    service_name: str   - service name.
    check: bool         - check or not manipulating result.
    sudo: bool          - run as sudo
    """
    service_name: str
    check: bool
    sudo: bool


class RestoreStageHandleInterface(Protocol):
    """Interface for restore stage handle command.

    Args:
        settings (Any): any settings for handling stage.
        backup_dir (str): path to backup.
    """

    def __init__(self, settings: Any, backup_dir: str) -> None:
        ...

    def execute(self) -> None:
        """Run command."""
        ...


def custom_copy_fnc(src: str, dst: str, *args: Any, **kwargs: Any) -> str:
    """Копирует файл и в случае возникновения ошибки повторяет копирование с правами sudo.

    Функция предназначена для работы с shutil.copytree.
    """
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    try:
        shutil.copy2(src, dst, *args, **kwargs)
    except BaseException as err:
        if isinstance(err, OSError) and err.errno == 26 and err.filename.endswith('.venv/bin/python'):
            pass
        else:
            sudo_cp_command: list[str] = [
                'sudo',
                'cp',
                '-p',
                src,
                dst,
            ]
            subprocess.run(sudo_cp_command, check=True)
    return dst


class RestoreFilesFromBackupCommand():
    """Replace all files in destination directory from backup.

    Args:
        settings (ReplaceFilesSettingsDict):    settings for replacing files.
        backup_dir (str):                       path to backup.
    """

    def __init__(self, settings: ReplaceFilesSettingsDict, backup_dir: str) -> None:
        self.src = backup_dir + '/' + settings['src']
        self.dst = settings['dst']

    def execute(self) -> None:
        """Replace all files in destination directory from backup."""
        logger.debug('Preparing to copy files from %s to %s.', self.src, self.dst)
        # copy files
        try:
            self._replace_files()
        except FileNotFoundError:
            logger.debug('Directory %s not found in the backup.', self.src)
            return
        logger.debug('Files has been copied from %s to %s.', self.src, self.dst)

    def _replace_files(self) -> None:
        shutil.copytree(self.src, self.dst, dirs_exist_ok=True, copy_function=custom_copy_fnc)


class RestoreOneFileFromBackupCommand():
    """Copy one files from backup ro destination.

    Args:
        settings (ReplaceFilesSettingsDict):    settings for replacing file.
        backup_dir (str):                       path to backup.
    """

    def __init__(self, settings: ReplaceFilesSettingsDict, backup_dir: str) -> None:
        self.src = backup_dir + '/' + settings['src']
        self.dst = settings['dst']

    def execute(self) -> None:
        """Run command."""
        # copy file
        logger.debug('Preparing to copy file from %s to %s.', self.src, self.dst)
        try:
            shutil.copyfile(self.src, self.dst)
        except FileNotFoundError:
            logger.debug('File %s not found in the backup.', self.src)
            return
        logger.debug('File has been copied from %s to %s.', self.src, self.dst)


class ReloadSystemdCommand():
    """Run command "systemctl daemon-reload".

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to enabled service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'daemon-reload',
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been enabled', self.service_name)


class EnableSystemdServiceCommand():
    """Enable systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to enabled service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'enable',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been enabled', self.service_name)


class StopSystemdServiceCommand():
    """Stop systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to stop service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'stop',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)

        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been stopped', self.service_name)


class StartSystemdServiceCommand():
    """Start systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to start service %s', self.service_name)

        command: list[str] = [
            'systemctl',
            'start',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command, check=False)
        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been started', self.service_name)


class StartUpdaterServiceCommand():
    """Start update service."""
    # FIXME: StartUpdaterServiceCommand нужен пока не реализован запуск всего модуля отдельным сервисом.

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to start service %s', self.service_name)

        command: list[str] = [
            'systemctl',
            'start',
            self.service_name,
        ]

        if self.sudo:
            command.insert(0, 'sudo')

        subprocess.Popen(command, close_fds=True, preexec_fn=os.setpgrp)

        logger.debug('%s service has been started', self.service_name)


class StopUpdaterProcessCommand():
    """Stop updater process without stopping service."""
    # FIXME: StopUpdaterServiceCommand нужен пока не реализован запуск всего модуля отдельным сервисом.

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to stop process %s', self.service_name)
        # FIXME: Грязь. Решится после запуска в отдельном сервисе.
        os.kill(PARENT_PID, signal.SIGINT)

        logger.debug('%s process has been stopped', self.service_name)


restore_stage_command_fabric: dict[str, Type[RestoreStageHandleInterface]] = {
    'restore_files': RestoreFilesFromBackupCommand,
    'restore_file': RestoreOneFileFromBackupCommand,
    'reload_services': ReloadSystemdCommand,
    'start_service': StartSystemdServiceCommand,
    'stop_service': StopSystemdServiceCommand,
    'start_updater': StartUpdaterServiceCommand,
    'stop_updater': StopUpdaterProcessCommand,
}


def existing_path_type(value: str) -> str:
    """Check that path is exist.

    Implement type for argparse. Checks that value is existing path.

    Args:
        value (str): path to dir

    Raises:
        argparse.ArgumentTypeError: Raises if path is not exist.

    Returns:
        str: path to dir
    """
    if not pathlib.Path(value).exists():
        raise argparse.ArgumentTypeError('Path is not exist')
    return value


def check_cmd_process(cmd_process: subprocess.CompletedProcess[bytes]) -> None:
    """Check that cmd process return without problem.

    Check that return code is zero or logging stderr and raises error.

    Args:
        cmd_process (subprocess.CompletedProcess): Complete command process.

    Raises:
        subprocess.CalledProcessError: raise if return code is not zero.
    """
    if cmd_process.returncode != 0:
        if cmd_process.stderr:
            for row in cmd_process.stderr.decode().splitlines():
                logger.error(row)
        raise subprocess.CalledProcessError(
            cmd_process.returncode,
            cmd_process.args,
            output=cmd_process.stdout,
            stderr=cmd_process.stderr,
        )


def main() -> None:
    """MAIN."""
    logger.info('Preparing to full restore.')

    with open(RUNNING_PATH/'metadata.json') as file:
        metadata = json.load(file)
    logger.debug('Metadata has been loaded.')

    parser = argparse.ArgumentParser()
    parser.add_argument('-b', '--backup', type=existing_path_type, required=True)
    # FIXME: Аргумент --parents_pid будет ненужен если запускать модуль в отдельном сервисе.
    parser.add_argument('-p', '--parents_pid', type=int, required=True)
    args = parser.parse_args()
    backup_dir: str = args.backup
    global PARENT_PID
    PARENT_PID = args.parents_pid
    if backup_dir:
        logger.debug('Backup dir has been received.')

    logger.info('Restoring...')
    for restoring_stage in metadata:
        logger.debug('Start command: %s', restoring_stage['command'])
        stage_command = restore_stage_command_fabric[restoring_stage['command']]
        stage_command(restoring_stage['settings'], backup_dir).execute()
        logger.debug('Command %s has been finished successful', restoring_stage['command'])
        time.sleep(1)
    logger.debug('Finish restoring.')


# FIXME: Сигналы уберутся после запуска в отдельном сервисе.
signal.signal(signal.SIGINT, signal.SIG_IGN)
signal.signal(signal.SIGTERM, signal.SIG_IGN)
if __name__ == '__main__':
    main()
