"""Main."""
import logging
from queue import Queue
from typing import Type

from paho.mqtt.client import MQTTMessage, MQTTv5

from updater import settings
from updater.api.mqtt import MqttRPClient, MqttUserdataDict
from updater.core.main_loop import MainLoop
from updater.core.msg_handlers import MsgHandlerWithResponse, RespMsgDict
from updater.commands.base import ICommandToLogic
from updater.commands.update_commands import (
    AddUpdateCommand,
    CheckUpdateCommand,
    DeleteUpdateCommand,
    GetModuleVersion,
    GetUpdateList,
    GetVersionsList,
    HandleStackingUpdate,
    InitialUpdateCommand,
    InstallUpdateCommand,
    PublisherAllTableRow,
    RestoreUpdateCommand,
)
from updater.db.models import Updates, Version


def main() -> None:
    """Start module."""
    # TODO: Переделать настройку logging на отдельный json файл
    logging.basicConfig(
        filename=settings.LOG_CONFIG['FILE'],
        format=settings.LOG_CONFIG['FORMAT'],
        level=settings.LOG_CONFIG['LEVEL'],
        datefmt=settings.LOG_CONFIG['DATE_FORMAT'],
    )

    receiver_queue: Queue[MQTTMessage] = Queue()
    publisher_queue: Queue[RespMsgDict] = Queue()

    publish_model_list = PublisherAllTableRow(publisher_queue)
    publish_model_list(set([Updates, Version]))
    settings.ServiceLocator = {'publish_table': publish_model_list}

    # Если при старте сервиса какое-то из обновлений находится
    # в не завершенном статусе то оно должно быть обработано
    HandleStackingUpdate(b'{}').execute()

    router: dict[str, Type[ICommandToLogic]] = {
        'lm/update_service/update/initial_update': InitialUpdateCommand,
        'lm/update_service/update/add_update': AddUpdateCommand,
        'lm/update_service/update/check_update': CheckUpdateCommand,
        'lm/update_service/update/delete_update': DeleteUpdateCommand,
        'lm/update_service/update/install_update': InstallUpdateCommand,
        'lm/update_service/update/restore_update': RestoreUpdateCommand,
        'lm/update_service/version/get_versions_list': GetVersionsList,
        'lm/update_service/version/get_module_version': GetModuleVersion,
        'lm/update_service/update/get_updates_list': GetUpdateList,
    }

    userdata: MqttUserdataDict = {
        'topics': list(router),
        'msg_queue': receiver_queue,
        'publisher_queue': publisher_queue,
        'publisher_loop_timeout': 0.5,
    }
    mqtt_client = MqttRPClient(client_id='lm_update_service', userdata=userdata, protocol=MQTTv5)
    mqtt_client.connect('localhost')
    mqtt_client.loop_start()
    mqtt_client.publisher_loop_start()

    msg_handler = MsgHandlerWithResponse(router, publisher_queue, service_topics=settings.SERVICE_TOPICS)
    main_loop = MainLoop(msg_queue=receiver_queue, msg_handler=msg_handler)
    main_loop.start()


if __name__ == '__main__':
    main()
