"""Утилиты для работы с сетевой конфигурацией."""
import json
import subprocess
from typing import Callable, cast


NETPLAN_TEMPLATE = """network:
  version: 2
  ethernets:
    eth0:
      addresses:
      - {ip_with_netmask}
      routes:
      - to: 0.0.0.0/0
        via: {gateway}
      nameservers:
        addresses:
        - 8.8.8.8
"""


def get_runtime_iface_ip_address(ifname: str) -> str:
    """Получить текущий ip адрес интерфейса.

    Args:
        ifname (str): Имя интерфейса.

    Raises:
        ValueError: Возвращается если не найдено не одного адреса у интерфейса.

    Returns:
        str: IP адрес интерфейса с длиной маски. Пример 192.168.0.205/24.
    """
    command: list[str] = [
        'ip',
        '--json',
        'address',
        'show',
        ifname,
    ]
    res = subprocess.run(command, capture_output=True)
    iface_settings = json.loads(res.stdout)[0]
    for family in iface_settings['addr_info']:
        if family['family'] == 'inet':
            return f"{family['local']}/{family['prefixlen']}"
    raise ValueError(f'No ip address configured on the {ifname} interface.')


def get_runtime_iface_gateway(ifname: str) -> str:
    """Получить текущий шлюз для интерфейса.

    Args:
        ifname (str): Имя интерфейса.

    Raises:
        ValueError: Возвращается если шлюз для интерфейса не сконфигурирован.

    Returns:
        str: IP адрес шлюза.
    """
    command: list[str] = [
        'ip',
        '--json',
        'route',
        'show',
        'dev',
        ifname,
        'match',
        '0/0',
    ]
    res = subprocess.run(command, capture_output=True)
    routes = json.loads(res.stdout)
    for route in routes:
        if gateway := route.get('gateway'):
            return cast(str, gateway)
    raise ValueError(f'No gateway configured on the {ifname} interface.')


def make_netplan_config_from_runtime_settings(
    ifname: str = 'eth0',
    get_iface_ip: Callable[[str], str] = get_runtime_iface_ip_address,
    get_iface_gateway: Callable[[str], str] = get_runtime_iface_gateway,
) -> str:
    """Подготавливает netplan конфигурацию интерфейса.

    Args:
        ifname (str, optional): Имя интерфейса. Defaults to 'eth0'.

    Returns:
        str: Netplan конфигурация интерфейса.
    """
    iface_ip_address = get_iface_ip(ifname)
    gateway = get_iface_gateway(ifname)
    return NETPLAN_TEMPLATE.format(ip_with_netmask=iface_ip_address, gateway=gateway)


def write_netplan_config(config: str, path: str = '10-ethernets.yaml') -> None:
    """Сохраняет netplan конфигурацию интерфейса на диск.

    Args:
        config (str): Netplan конфигурация интерфейса.
        path (str, optional): Имя файла конфигурации. Defaults to '10-ethernets.yaml'.
    """
    with open(path, 'w') as file:
        file.write(config)
